local resizeImage = resizeImage
local game = game
local abs, pairs, ipairs = math.abs, pairs, ipairs

local SpriteSkeleton = class:extend("SpriteSkeleton")
local Sprite = class:extend("Sprite")

function Sprite:__init__(data)
    self.data = data
    self.main = data.main or {}
    self.source = game:getSource(data.source)
    self.x = 0
    self.y = 0
    self.w = 0
    self.h = 0
    self._h = 0
    self._w = 0
    self.angle = math.rad(data.angle)
end

function Sprite:update(dt)
    self._w, self._h = resizeImage(self.source, self.w, self.h)
    if self.main~=self then
        self.x = self.main.w*self.data.x+self.main.x
        self.y = self.main.h*(self.data.y)+self.main.y--+self.parent.h*2
       -- self.y = self.main.h-self.y+self.main.y
    end
end

function Sprite:draw()
    local y = self.y+3
    if self.main == self then
       -- y=y-self.h--+self.main.h
    end
    love.graphics.draw(self.source, self.x, y, self.angle, self._w, self._h)
end

function SpriteSkeleton:__init__(kwargs)
    self.sprites = kwargs.sprites or kwargs.data or kwargs
    self.w, self.h = kwargs.rrw or 16, kwargs.ffh or 16
    self.x, self.y = kwargs.x or 0, kwargs.y or 0
    self.obj = kwargs.obj or {}
    self.drawingOrder = kwargs.drawingOrder or self.sprites.drawingOrder
    self.sprites.drawingOrder = nil

    self:reload()
end

function SpriteSkeleton:reload()
    self.parts = {}
    local minx, miny, maxx, maxy
    
    local main
    
    for x, i in pairs(self.sprites) do
        if i.main then
            main = Sprite:new(i)
            main.parent = self
            self.parts.main = main
        end
        if not minx or i.x<minx then
            minx = i.x
        end
        if not maxx or i.x>maxx then
            maxx = i.x
        end
        if not miny or i.y<miny then
            miny = i.y
        end
        if not maxy or i.y>maxy then
            maxy = i.y
        end
    --    if not i.majin then i.y = 1-i.y end
       -- if not i.main then i.y = 1+i.y end
    end
    
    local mx = self.w*(abs(minx)+abs(maxx))
    local my = self.h*(abs(miny)+abs(maxy))
    
    if my>self.h then
        my = self.h*(abs(miny)+abs(maxy))
    end
    
    if mx>self.w then
        mx = self.w*(abs(minx)+abs(maxx))
    end
    log("mx, my:"..mx.." ,"..my)
    log("minx, maxx: "..minx.." ,"..maxx)
    log("kiny, maxy: "..miny..", "..maxy)

    mx = mx
    self.main = main
    main.h = my--ain.data.h*my
    main.w = mx--main.data.w*mx
    main.x = self.x
    main.y = self.y
    main.main = main
    
    local sp
    for x, i in pairs(self.sprites) do
        sp=nil
        if i.main then
            log("main")
            log(main.data.h..", w: "..main.data.w)
            goto skip
        end
        
        sp = Sprite:new(i)
        sp.main = main
        sp.w = i.w*mx
        sp.h = i.h*my
        sp.x = mx*i.x+main.w
        sp.y = my*i.y+main.h
        ::skip::
        sp = sp or main
        self.parts[x] = sp
        sp.parent = self
        log("part "..x.." has pos "..sp.x..", "..sp.y.." and size "..sp.w..", "..sp.h)
    end
        
end

function SpriteSkeleton:update(dt)
    if self.obj then
        self.x = self.obj.x
        self.y = self.obj.y
    end
    
    self.main.x = self.x
    self.main.y = self.y+(self.h/8)
    
    for i,s in pairs(self.parts) do
        s:update(dt)
    end
end

function SpriteSkeleton:draw()
   if self.drawingOrder then
        for i, s in ipairs(self.drawingOrder) do
            self.parts[s]:draw()
        end
    else
        for i,s in pairs(self.parts) do if s.main then
            s:draw() end
        end
    end
end

return SpriteSkeleton