local Tile = class:extend("Tile")--,50010)
local getUpdateCount = getUpdateCount

local lg = love.graphics
local lgdraw = lg.draw
local lggetColor = lg.getColor
local lgsetColor = lg.setColor
local lgrectangle = lg.rectangle

function Tile:__init__(kwargs)
    local game = game
    local map = game.map
    local kwargs = kwargs or {}
    local _tile = kwargs._tile or {}
    self.active = true
    self.life = kwargs.life or 3
    self.x = kwargs.x or 0
    self.y = kwargs.y or 0
    
    self.passive = kwargs.passive
    self.id = _tile.id or string.format("%s, %s",self.x,self.y)
    
    if self.passive then
        self.id = self.id or lume.uuid()
    end
    
    self._w, self._h = 1,1
    self.source = _tile.source or "dirt.png"
    
    self.isBackground = false
    
    self.minis = {}
    self.totalMinis = 0
    
    self.w = map.tile.w
    self.h = map.tile.h
    isTile = true
    self.uuid = self.id
    self.parent = kwargs.parent
    self._x = self.w*self.x
    self._y = self.h*self.y
    if not self.passive then
        map.world:add(self,self._x,self._y, self.w, self.h)
        map.entities[self.id] = self
    end
    self.streaks = {}
    self.totalLife = self.life
end


function Tile:getData()
    local parent = self.parent and self.parent.uuid
    local r = {
        self.uuid,
        self.__class,
        self.x,
        self.y,
        self.w,
        self.h,
        self.angle,
        self.source,
        self.life,
        parent
    }
    return r
end

Tile.changeTo = function(self, tile)
    self.life = tile.life or self.life or 3
    if not tile.ignoreSource then
        self.source = tile.source
        if tile.source == "whiteDirt.png" and tile.parent ~= game.map.player then
            self.source = "dirt.png"
        elseif self.source == "dirt.png" and tile.parent == game.map.player then
            self.source = "whiteDirt.png"
        end
    end
    
    self.isBackground = tile.isBackground
    self.active = tile.active
    self.parent = tile.parent
    self.totalLife = self.life
    
    --self.timeMade = 1000000000000000000000--
    self._tm = love.timer.getTime()
    self.timeMade = self._tm
end

function Tile:getAttacked(attacker,attack)
    if attack or type(attacker) =="number" then
        attack = attacker
        attacker = nil
    elseif not attack then
        attack = attacker.attack
    end
    
    if attacker and attacker.parent and attacker.parent.team==self.team then
        --fix--??
        attack = self.totalLife/2
    end
    
    self.life = self.life - attack
    
    if self.life<=0 and self.active then
        if attacker and attacker.parent and attacker.parent.isPlayer then
            attacker.parent.stats.blocksDestroyed = attacker.parent.stats.blocksDestroyed+1
        end
        local noSpawn
        if attacker and attacker.isExplosion then
            noSpawn = true
        end
        self:destroy(noSpawn)
    end
end

Tile.isVisible = function(self)
    local p = (self._inView) or self.pride
    return p and p >= RECORDS.updateCount
end

Tile.unitr = function(self, unit)
    if unit.tile == self then
        return unit
    elseif unit.tile then
        unit.tile:unUnit(unit)
    end
    unit.tile = self
    --table.insert(self.units, unit)
    --self.unitCount = self.unitCount+1
end

Tile.unUnit = function(self, unit)
    assert(unit.tile==self, string.format("Unit %s does not belong to this tile",unit.__class))

    unit.tile = nil
    self.unit = nil
    
    return unit
end

Tile.getX = function(self)
    return self.x
end

Tile.getY = function(self)
    return self.y
end

Tile.getPos = function(self)
    return self.x,self.y
end
    
Tile.updating = function(self)
    local c = map.allCurrentChunks.x
    local cy = map.allCurrentChunks.y
    local iff = lume.find(c,self.chunkx) and
        lume.find(cy,self.chunky)
    return iff
end
        
Tile.destroy = function(self, noSpawn, sent)
    if self.invincible then
        --"Woah! I got destroyed but I'm invincible"
        self.life = 10000000
        return
    end
    local map = game.map
    
    if not self.active then
        return
    end
    
    self.parent = nil
    self.active = false
    local map = game.map
    
    if not noSpawn then
        local mini = Mini.spawnMini(self,nil,nil,1)
        self.minis[mini.uuid] = mini
        self.totalMinis = self.totalMinis+1
    end
    
    if not self.passive then
        if map.host then
            map.host:sendToAll("destroyTile",{self.x,self.y,noSpawn})
        elseif map.client and not self.isBackground and not sent then
            map.client:send("hostDestroyTile",{self.x,self.y,noSpawn})
        end
    end
    
    self.source = "background_dirt.png"
    self.collider = nil
    self.isBackground = true

end
    
Tile.update = function(self,dt,c)
end

Tile.setInvincible = function(self,inv)
    if inv and not self.invincible then
        self.invincible = true
        self.oldlife = self.life
        self.life = 100000000000
        self.oldsource = self.source
        self.source = "stone.png"
        self.oldback = self.isBackground
        self.isBackground = false
    elseif not inv then
        self.invincible = false
        self.life = self.oldlife
        self.source = self.oldsource
        self.isBackground = self.oldback
    end
end
    
Tile.draw = function(self)

    --really increases fps
    local map = game.map
    local r,g,b,a
    local parent = self.parent
    if self.source == "whiteDirt.png" then
        self.source = "dirt.png"
    end
    if self.isBackground then
        self._tm = self._tm or self.timeMade
        self.timeMade = 1000000000000000
        self.streaks = {}
        --self.color = getColor("orange")
    elseif self._tm then
        self.timeMade = self._tm
        self._tm = nil
    end
    
    for x, i in ipairs(self.streaks) do
        r,g,b,a = love.graphics.getColor()
        love.graphics.setColor(getColor(i.color))
         pn = love.graphics.getLineWidth()
        love.graphics.setLineWidth(ppn or i.w)
        love.graphics.line(i.ox,i.oy,i.x,i.y)--,i.w,i.h)
        love.graphics.setLineWidth(pn)
        love.graphics.setColor(r,g,b,a)
    end
    
    if game.settings.raw then
        if not self.isBackground and map.comic then
            highlightTile(self._x, self._y, self.w, self.h)
        end
        local col = self.isBackground and getColor("dirtbrown",self.alpha) or
            getColor(
            (parent and (parent==game.player and parent.team~=parent.uuid and parent.team or parent.color) 
            or 
            map.tileColors[self.source]),self.alpha)
        local r,g,b,a = love.graphics.getColor()
        love.graphics.setColor(col)
        love.graphics.rectangle("fill",self._x,self._y,self.w,self.h)
        love.graphics.setColor(r,g,b,a)
        return
    end
    if parent and parent==game.player then
        r,g,b,a = love.graphics.getColor()
        love.graphics.setColor(
            getColor(parent.team~=parent.uuid and parent.team or parent.color)
        )
    end
        
    lgdraw(map.tileSources[self.source],
        self._x,self._y,
        0,
        map._tw, map._th
    )
    
    if map.ftile==self then
        local r, g, b, a = lggetColor()
        lgsetColor(getColor("blue"))
        lgrectangle("line",self._x,self._y,self.w,self.h)--,w,h)
        lgsetColor(r,g,b,a)
    end
    
    if self.color then
        local r, g, b, a = lggetColor()
        lgsetColor(self.color)
        lgrectangle("fill",self._x,self._y,self.w,self.h)--,w,h)
        lgsetColor(r,g,b,a)
    end
    
    if self.parent and parent==game.player then
        love.graphics.setColor(r,g,b,a)
    end
    if not self.isBackground and map.comic then
        highlightTile(self._x, self._y, self.w, self.h)
    end
    
end

function Tile:getCenter()
    return self._x + self.w / 2,
         self._y + self.h / 2
end

function Tile:dir(n)
    if n > 0 then return 1 elseif n < 0 then return -1
    else return 0 end
end

function Tile:getDir(...)
    return self:dir(...)
end

function Tile._i_lt(self,other)
    return self.timeMade<other.timeMade
end

return Tile