local bullets = {}
local ease = 1
local abs = math.abs--function() return 0 end
local atan2 = math.atan2

bullets.Bullet = Bullet
bullets.Bullet.myName = "Normal"
EASE = 1
bullets.MissileBullet = Bullet:extend("MissileBullet",m1000)
bullets.MissileBullet.myName = "Heat Seeker"
bullets.MissileBullet.amount = {5,10}

function bullets.MissileBullet:__init__(kwargs)
    kwargs.delay = 1
    kwargs.life = 230
    kwargs.hits = 130
    kwargs.h = 5
    kwargs.w = 10
    
    Bullet.__init__(self,kwargs)
    self.reallyShot = kwargs.reallyShot--shot by a real player?
    self.target = kwargs.target or {x=0,y=0}

    if self.parent.isPlayer then
        local i = (self.parent.index or 1)+1
        if i>#game.players then i = 1 end
        self.target = game.players[i] or {x=0,y=0}
    end
    if self.target == self.parent then
        for x=1,#self.map.turrets do
            local t = self.map.turrets[x]
            if t.team~=self.parent.team then
                self.target = t
            end
        end
    end
    
    self.ease = kwargs.ease or ease
end

function bullets.MissileBullet:update(dt,c)-- ProFi:start()
    Bullet.update(self,dt,c)
    local ease = EASE
    local speed = SP or 2
    local target = self.target
    local targetX = target.x-self.x
    local targetY = target.y-self.y
    local rotation = atan2(targetY, targetX)
    
    if abs(rotation-self.angle)>180 then
        if rotation>0 and self.angle<0 then
            self.angle = self.angle-(360-rotation+self.angle)/ease
        elseif self.angle>0 and rotation<0 then
            self.angle = self.angle+(360-rotation+self.angle)/ease
        end
    elseif rotation<self.angle then
        self.angle = self.angle - abs(self.angle-rotation)/ease
    else
        self.angle = self.angle + abs(rotation-self.angle)/ease
    end
    
    local vx = speed*(90-abs(self.angle))*-1
    local vy = 0
    if self.angle < 0 then
        vy = -speed+abs(vx)
    else
        vy = speed-abs(vx)
    end --self.angle = math.rad(self.angle)
    self.vx,self.vy = vx,-vy
    if target.x>self.x then
        self.vx = self.vx*-1
    end
    
end

function bullets.MissileBullet:draw()

    local v = game:getSource(self.source)
    local xx,yy = v:getWidth()/2, v:getHeight()/2
    if self.isBouncyBullet then
        xx,yy=8,8
    end
    local ca = math.pi^2
    
    love.graphics.draw(game:getSource(self.source),self.x+xx,self.y+yy,math.pi+self.angle+ca,
    self._w,self._h,xx,yy)
    
end

function bullets.MissileBullet:move(x,y,dt)
    self.die = false
    self.life = self.life - dt
    if self.life<=0 then
        self.die = true
    end
    
    
    if not self.active then
        return 0,0
    end
    local actualX, actualY, collisions, len = self.world:move(self,x,y,
    self.checkCollision)
    local cc
    
    for u = 1,len do
        cc = collisions[u]
        Entity.resolveCollision(self,cc)
        if cc.other.life>0 and not self.die then
            for i = 1, self.hits do
                cc.other:getAttacked(self)
                self.hits = self.hits-1
                if not cc.other.isTile then
                    cc.other:applyImpulse(self.vx/2, self.vy/2)
                end
                if cc.other == self.target then
                    self.hits = 0
                    self.life = 0
                    break
                end
                if cc.other.life <=0 then
                    if cc.other.isTile then
                        cc.other:destroy()
                    end
                    break
                end
            end
        end
        if cc.other.isTile  and not cc.other:updating() then--??
            cc.other:update(dt)
        end
        
        --die if hit too many things
        if self.hits<=0 then
            self.die = true
        end

    end
    if self.die then
        self:destroy()
    end
    return actualX, actualY
end

function bullets.MissileBullet:destroy()
    Bullet.destroy(self)
    Explosion:new({map=self.map,x=self.x, y=self.y,life=2,parent=self})
end

---Smokescreen Bullet---
bullets.SmokescreenBullet = Bullet:extend("SmokescreenBullet",j1000)
bullets.SmokescreenBullet.myName = "Smoker"
bullets.SmokescreenBullet.amount = {20,50}

local smSource = "bullets/smokescreen_bullet.png"
--bullets.SmokescreenBullet.super = Bullet
function bullets.SmokescreenBullet:__init__(kwargs)
    local kwargs = kwargs or {}
    kwargs.weight = 10
    self.name = "Smoker"
    kwargs.w = 10
    kwargs.h = 5
    kwargs.source = smSource
    kwargs.elasticity=0
    self.super.__init__(self,kwargs)
    self.opacity = kwargs.opacity or 0.1
    self.opacity2 = kwargs.opacity2 or 0.1
    self.hits = 2
end

function bullets.SmokescreenBullet:destroy()
    self.super.destroy(self)

    Explosion:new({
        source = "smoke.png",
        map = self.map,
        x=self.x,
        y=self.y,
        life=2,
        w=1,h=1,
        opacity = 0,--self.opacity,
        opacity2 = self.opacity2,
        passive=true,
        parent=self.parent
    })
end

---Bouncy Bullet--
bullets.BouncyBullet = Bullet:extend("BouncyBullet")
bullets.BouncyBullet.myName = "Bounce"
bullets.BouncyBullet.amount = {10,30}

function bullets.BouncyBullet:__init__(kwargs)
    self.name = "Bouncer"
    local kwargs = kwargs or {}
    kwargs.elasticity = 1.2
    kwargs.w = math.sqrt(2 * 5 * 5)
    kwargs.h = kwargs.w
    kwargs.hits = math.random(15,20)
    kwargs.source = "bullets/bouncy_bullet.png"
    Bullet.__init__(self,kwargs)
    self.collisionType = "bounce"
    self.isBouncyBullet = true
end
    
function bullets.BouncyBullet:changeVelocityByCollisionNormal(nx, ny, bounciness)
  local bounciness = bounciness or self.elasticity
  if self.vy == 0 then
      self.vy = -100
  end
  
  local vx, vy = self.vx, self.vy
  if (nx < 0 and vx > 0) or (nx > 0 and vx < 0) then
    self.bounced.x = true
  end

  if (ny < 0 and vy > 0) or (vy < 0 and ny > 0 ) then
  
    self.bounced.y = true
  end
  if self.bounced.x or self.bounced.y then
    vy = randomchoice({-vy,vy})
    vx = randomchoice({-vx,vx})
  end
  
  
  if math.abs(vy)<90 then
      vy = getDir(vy)*100
  end
  if math.abs(vx)<90 then
      vx = getDir(vx)*100
  end
  
  self.vy, self.vx = vy, vx
  
end

function bullets.BouncyBullet.checkCollisiuon(item, other)
    if other.isTile and not item.parent.breakBlocks and other.parent and other.parent.team==item.parent.team then
        return nil
    end
    
    if other.isTile and not other.isBackground then
        return "bounce"--item.collisionType
    end
    
    if other == item.parent and item.ignoresParent then
        return nil
    elseif other.isBullet or other.isFire then
        return nil
    elseif not other.isBackground and not item:ignore(other) then
        return item.collisionType
    end
end


---Explosive Bullet---
bullets.ExplosiveBullet = Bullet:extend("ExplosiveBullet")
bullets.ExplosiveBullet.myName = "Boom"
bullets.ExplosiveBullet.amount = {15, 30}

function bullets.ExplosiveBullet:__init__(kwargs)
    kwargs.source = "bullets/explosive_bullet.png"
    Bullet.__init__(self, kwargs)
end

function bullets.ExplosiveBullet:destroy()
    Entity.destroy(self)
    Explosion:new({map=self.map,x=self.x, y=self.y,life=2,parent=self})
end


---Explosive Bouncy Bullet--
bullets.ExplosiveBouncyBullet = bullets.BouncyBullet:extend("ExplosiveBouncyBullet")
bullets.ExplosiveBouncyBullet.myName = "Sugar Rush"
bullets.ExplosiveBouncyBullet.amount = {10,20}

function bullets.ExplosiveBouncyBullet:__init__(kwargs)
    bullets.BouncyBullet.__init__(self, kwargs)
    self.color = lume.randomchoice(colors)
end

function bullets.ExplosiveBouncyBullet:move(...)
    local ax, ay = bullets.BouncyBullet.move(self, ...)
    
    if self.bounced.x or self.bounced.y then
        Explosion:new({map=self.map,x=self.x, y=self.y,life=2,w=10,h=10,parent=self})
    end
    
    return ax, ay
end

function bullets.ExplosiveBouncyBullet:update(...)
    bullets.BouncyBullet.update(self,...)
    self.color[1]=self.color[1]+0.01
end

function bullets.ExplosiveBouncyBullet:draw()
    r, g, b, a = love.graphics.getColor()
    love.graphics.setColor(self.color)
    bullets.BouncyBullet.draw(self)
    love.graphics.setColor(r,g,b,a)
end

---Strong Bullet--
bullets.StrongBullet = Bullet:extend("StrongBullet")
bullets.StrongBullet.myName = "Hercules"
bullets.StrongBullet.amount = {10,20}

function bullets.StrongBullet:__init__(kwargs)
    self.name = "Strong"
    local kwargs = kwargs or {}
    kwargs.attack = 2
    self.delay = ddc or 1
    kwargs.ac = aac or 1500
    kwargs.hits = math.random(10,20)
    kwargs.hits = math.random(5,20)
    Bullet.__init__(self,kwargs)
    
end


---Grenade Bullets---
--Work in progress--
bullets.GrenadeBullet = Bullet:extend("GrenadeBullet")
bullets.GrenadeBullet.myName = "Grenade"
bullets.GrenadeBullet.amount = {5,10}

local Gradius = 8
local MAX_GRENADES = 14050

function bullets.GrenadeBullet:__init__(kwargs)
    
    kwargs.elasticity = .4
    kwargs.life = 4
    kwargs.w = math.sqrt(2 * Gradius * Gradius)
    kwargs.h = kwargs.w
    kwargs.delay = 1
   -- kwargs.ac=100
    kwargs.attack = 0--kwargs.attack or 10
    local notContinue = Bullet.__init__(self, kwargs)
    if notContinue then return end
    self.parent._grenadesActive = self.parent._grenadesActive or 0
    if self.parent._grenadesActive>=MAX_GRENADES then
        self:passiveDestroy()
        return
    end
    self.parent._grenadesActive = self.parent._grenadesActive + 1
  --[[  
    local joyX, joyY = 1,1
    if self.parent.isPlayer then
    local rt = self.parent.Rjoystick
        joyX = kwargs.joyX or self.parent.Rjoystick:getX()*20
        joyY = kwargs.joyY or self.parent.Rjoystick:getY()*20
        joyX = math.cos(rt.angle) * (rt.d-rt.deadzone)/(1-rt.deadzone)*20
        joyY = -math.sin(rt.angle) * (rt.d-rt.deadzone)/(1-rt.deadzone)*20
        
    end
    
    self.ac = kwargs.speed or kwargs.ac or 200
    self.vx = self.vx*joyX
    self.vy = self.vy*joyY
    self.vox = self.ac*self.angleX*joyX*-1
    self.vyi = self.ac*-1*self.angleY*joyY*-1
    
    self.bounceSpeed = 30
    self.ignoresParent = true
    self.attack = 0
  --  self.vx = 90*(self.angleX)
  --  self.vy = -100*(self.angleY)
    self.collisionType = "bounce"]]--
    self.source = "bullets/grenade_bullet.png"
    self.attack = kwargs.attack

    
end

function bullets.GrenadeBullet:getBounceSpeed(nx, ny)
  if nx == 0 then return math.abs(self.vy) else return math.abs(self.vx) end
end

function bullets.GrenadeBullet:emitCollisionSound(nx, ny)
  local speed = self:getBounceSpeed(nx, ny)
  if speed >= self.bounceSoundSpeed then
    media.sfx.grenade_wall_hit:play()
  end
end
IT="slide"
function bullets.GrenadeBullet.checkCollision(item, other)
    if other.isTile and not other.isBackground then
        return "bounce"--item.collisionType
    end
    
    
    if other == item.parent and item.ignoresParent then
        return nil
    elseif other.isPlayer then
       return "slide"
    elseif other.isBullet or other.isMini or other.isFire then
        return nil
    elseif not other.isBackground then
        return IT--item.collisionType
    end


end

function bullets.GrenadeBullet:move(x,y,dt)
    self.die = false
    self.life = self.life - dt
    if self.life<=0 then
        self.die = true
    end
    
    
    if not self.active then
        return 0,0
    end
    local actualX, actualY, collisions, len = self.world:move(self,x,y,
    self.checkCollision)
    local cc, col
    
    for u = 1,len do
        cc = collisions[u]
        col = cc
        if col.other ~= self.parent or not self.ignoresParent then
            local nx, ny = col.normal.x, col.normal.y
            Entity.resolveCollision(self,cc)
            --self:emitCollisionSound(nx, ny)

            if cc.other.isTile and cc.other.life>0 and not self.die then
                for i = 1, self.hits do
                    cc.other:getAttacked(self)
                    self.hits = self.hits-1
                    if cc.other.life <=0 then
                        break
                    end
                end
            end
            if cc.other.isTile  and not cc.other:updating() then--??
                cc.other:update(dt)
            end
        end

    end
    if self.die then
        self:destroy()
    end
    return actualX, actualY
end

function bullets.GrenadeBullet:draw()
    local v = game:getSource(self.source)
    local xx,yy = v:getWidth()/2, v:getHeight()/2
    if self.isBouncyBullet then
        xx,yy=8,8
    end
    local ca = math.pi^2
    --local ang = math.deg(self.angle)
    
    love.graphics.draw(game:getSource(self.source),self.x,self.y,0,
    self._w,self._h)

    --love.graphics.print(self.angleX..", "..self.angleY,self.x,self.y)
    
end

function bullets.GrenadeBullet:update(dt,c)

    self.updateCount = self.updateCount + 1

    self._w, self._h = resizeImage(game:getSource(self.source),self.w,self.h)

    self:updateGravity(dt)
    

    self.bounced.x = false
    self.bounced.y = false
    
    local vx, vy = self.x+self.vx*dt,self.y+self.vy*dt
    self.vx = self.vx/2--self:dir(self.vx)*dt
        
  
    if 1 then --??
        local mx, my = self:move(vx,vy,dt)
        self.x, self.y = mx,my --assert(self.vx<=50,self.vx)
    end
    if self.active then
        self:detectExitedParent()
    end
    self:getTile()
end

function bullets.GrenadeBullet:destroy()
    Entity.destroy(self)
    if not self.parent then return end
    Explosion:new({map=self.map,x=self.x, y=self.y,life=2,attack=self.attack,parent=self})
    self.parent._grenadesActive = self.parent._grenadesActive-1
end

function bullets.GrenadeBullet:passiveDestroy()
    Entity.destroy(self)
end

function bullets.GrenadeBullet:changeVelocityByCollisionNormal(nx, ny, bounciness)
  bounciness = .4
  local vx, vy = self.vx, self.vy

  if (nx < 0 and vx > 0) or (nx > 0 and vx < 0) then
    vx = -vx * bounciness
  end

  if (ny < 0 and vy > 0) or (ny > 0 and vy < 0) then
    vy = -vy * bounciness
  end

  self.vx, self.vy = vx, vy
end

--bullets.GrenadeBullet.changeVelocityByCollisionNormal = bullets.BouncyBullet.changeVelocityByCollisionNormal

---Fly Bullet--
bullets.FlyBullet = Bullet:extend("FlyBullet")
bullets.FlyBullet.myName = "Flyer"
bullets.FlyBullet.amount = {30,70}

function bullets.FlyBullet:__init__(kwargs)
    Bullet.__init__(self,kwargs)
    self.name = "Flyer"
    --self.attack = 3
end








--[[-Flame Bullet--
---???Not Currently Supported because of:...???---
--flames spread weirdly
--flames collide weirdly and may 
----move through wall if shot to wall directly

bullets.FlameBullet = Bullet:extend("FlameBullet")

function bullets.FlameBullet:__init__(kwargs)
    Bullet.__init__(self,kwargs)
    self.source = "flame.png"
    self.attack = 0
end

function bullets.FlameBullet:destroy()
    Bullet.destroy(self)
    Fire:new({map=self.map,x=self.x,y=self.y})
end]]


return bullets

