local Crate = Entity:extend("Crate")
local abs = math.abs
local getUpdateCount = getUpdateCount

local arrow = game:getSource("pointer.png")
local arrowHalfW = arrow:getWidth()/2
local arrowHalfH = arrow:getHeight()/2

function Crate.spawnCrate(parent, amount, tile, source, life)
    return Crate:new({
        parent = parent,
        source = source,
        life = life,--last 1 minute
        map = parent.map or game.map,
        x = tile._x or parent.x,
        y = tile._y or parent.y,
        w = 15,
        item = parent,
        amount = amount
    })
end
    
    
function Crate:__init__(kwargs)
    self.isCrate = true
    
    kwargs.h = kwargs.w
    Entity.__init__(self,kwargs)
    self.source = kwargs.source
    self.parent = kwargs.parent or warn("no parent")
    self.color = "woodbrown"
    self.life = kwargs.life or 3
    if self.parent then
        if self.parent.isTile then
            self.source = "crates/blockCrate.png"
        elseif self.parent.isBullet then
            self.source = "crates/bulletCrate.png"
        elseif self.parent.heal then
            self.source = "crates/healthCrate.png"
            self.heal = self.parent.heal
            self.color = "white"
            self.life = self.heal
        end
    end
    self.w = kwargs.w or kwargs.h or 14
    self.h = self.w
    
    self.totalHealth = self.life
    if not self.passive then
        self.world:remove(self)

        --so crate can fit through tunnels of 1 tile space
        self.world:add(self,self.x,self.y,self.w-2,self.h-2)
        
        --hitchhiking
        self.map:storeBullet(self)
    else
        --self.real = false
    end
    
    self.isExplosive = false
    self.elasticity = kwargs.elasticity or .4
    
    self.collisionType = "slide"
    self._inView = 0
    
   -- self:destroy()
end

function Crate:inView()
    return --not self.map.fogOfWar or
    self._inView >= getUpdateCount()
end

function Crate:drawPointer()
    self.map.camera:detach()
    local screenW, screenH = love.graphics.getDimensions()
    local t, minT
    minT = math.huge
    
    local cam = self.map.camera
    local playerX,playerY = cam:toCameraCoords(game.map.player.x,game.map.player.y)
    local myX, myY = cam:toCameraCoords(self.x, self.y)
    
    local vx, vy = myX - playerX, myY - playerY
    local arrowPosX,arrowPosY
    
    t = (arrowHalfW - playerX) / vx
    if t == t and t > 0 and t < minT then
      minT = t
      arrowPosX = arrowHalfW
      arrowPosY = vy * t + playerY
    end
    t = (arrowHalfH - playerY) / vy
    if t == t and t > 0 and t < minT then
      minT = t
      arrowPosX = vx * t + playerX
      arrowPosY = arrowHalfH
    end
    t = (screenW - arrowHalfW - playerX) / vx
    if t == t and t > 0 and t < minT then
      minT = t
      arrowPosX = screenW - arrowHalfW
      arrowPosY = vy * t + playerY
    end
    t = (screenH - arrowHalfH - playerY) / vy
    if t > 0 and t < minT then
      -- minT = t -- doesn't matter anymore
      arrowPosX = vx * t + playerX
      arrowPosY = screenH - arrowHalfH
    end
    if self:inView() and not self.tile:isVisible() then
        local p = 1
        if vx<0 then
            p = -1
        end
        arrowPosX = myX-(self.w*cam.scale*p)
        arrowPosY = myY+(self.h/2)*cam.scale
    end
    local color
    if self.parent.isBullet then
        color = getColor("red")
    elseif self.parent.isTile then
        color = getColor("chocolate")
    elseif self.heal then
        color = getColor("white")
    elseif self.isPlayer then
        color = getColor("blue")
    end
    
    
    local r,g,b,a = love.graphics.getColor()
    love.graphics.setColor(color)
    love.graphics.draw(arrow, arrowPosX, arrowPosY,
      math.atan2(vy, vx),
      1, 1, arrowHalfW, arrowHalfH)
    love.graphics.setColor(r,g,b,a)
    self.map.camera:attach()
end

function Crate:getSource()
    return self.source
end

function Crate:destroy(sent)
    if self.map.host then
        self.map.host:sendToAll("destroyCrate",self.crateID)
    elseif self.map.isClient and not sent then
        return
    end
    
    Entity.destroy(self)
    self.map.crates[self.crateID] = nil

    if not self.heal and not self.map.isOld then
        self.parent.miniID = self.miniID
        local m = Mini.spawnMini(self.parent)
        m.glow = true
        self.minis[#self.minis+1] = m
        self.totalMinis = self.totalMinis+1
        m:setPos(self.x,self.y)
    end
end

function Crate:update(dt, c) --ProFi:start()

    self.updateCount = self.updateCount + 1
    
    --self._w, self._h = resizeImage(game:getSource(self.source),self.w,self.h)
    self:updateGravity(dt)
    
    local vx, vy = self.x+self.vx*dt, self.y+self.vy*dt
    --friction
    if math.floor(math.abs(self.vx))>0 then
        self.vx = self.vx - self:getDir(self.vx)*10*dt
    end
    if 1 then
        local mx, my = self:move(vx,vy,dt)
        self.x, self.y = mx,my
    end
    self:getTile()-- ProFi:stop() ProFi:writeReport("crate.txt")
end


function Crate:checkLimits()
    local maxVx,maxVy = 1000,100
    
    if self.vy>0 then
        self.oldOnGround = false
    end
    
    if math.abs(self.vx)>maxVx then
        self.vx = maxVx*self:getDir(self.vx)
    end
    
    if self.vy>0 and self.onGround then
        self.vy = 0
    end
    
    if math.abs(self.vy)>maxVy then
        self.vy = maxVy*self:getDir(self.vy)
    end

end

function Crate:changeVelocityByCollisionNormal(nx, ny, bounciness)
  local bounciness = bounciness or self.elasticity
  if not bounciness or bounciness == 0 then
      return
  end
  local vx, vy = self.vx, self.vy

  if (nx < 0 and vx > 0) or (nx > 0 and vx < 0) then
    vx = -vx * bounciness
    self.bounced.x = true
  end
  
  if ((ny < 0 and vy > 0) or (ny > 0 and vy < 0)) and abs(vy)>10 then
    vy = -vy * bounciness
    self.bounced.y = true
  end

  self.vx, self.vy = vx, vy
end

function Crate.checkCollision(item, other)
  
    if other.isCrate or other.isTile and not other.isBackground then
   --     other.life = other.life - item.attack
     --   item.die = true
        return item.collisionType
    end
    if not other.isMini and not other.isBackground and other.isSolid then
        return item.collisionType
    end
end

function Crate:move(x,y,dt)
    self.die = false
    --self.life = self.life - dt
    
    if self.life<=0 then
        self.die = true
    end
    
    
    if not self.active then error()
        return 0,0
    end
    local actualX, actualY, collisions, len = self.world:move(self,x,y,
    self.checkCollision)
    local cc
    for u = 1,len do
        cc = collisions[u]
        self.super.resolveCollision(self,cc)
    end
    if self.die then
        self:destroy(1)
    end
    
    return actualX, actualY
end

function Crate:getAttacked(att)
    Entity.getAttacked(self,att)
    if not self.heal then
        return
    end
    local real = nil
    if att.parent and not att.parent.isPlayer then
        att.parent = att.parent.parent or att.parent
    end
    if att.parent and att.parent.isPlayer then
        att.parent.life = att.parent.life+att.attack
        if att.parent.life > 100 then
            att.parent.life = 100
        end
        if att.parent.isPlayer and att.parent.real then
            real = 1
        end
    end
    
    if real then log(self.crateID)
        self.map.player:send("updateCrate",{self.crateID, self.life, self.x, self.y})
    end
    
end

--todo: draw trail
function Crate:draw()
    
    local v = game:getSource(self.source)
    local xx,yy = v:getWidth()/2, v:getHeight()/2
    local ca = math.pi^2
    local map = self.map
    
    
    love.graphics.draw(game:getSource(self.source),self.x,self.y,map._cw,map._ch)--+xx,self.y+yy,math.pi-(self.angle+ca),
  --  self._w,self._h,xx,yy)
  if self.what then love.graphics.rectangle("fill",self.x,self.y,self.w,self.h) end
    self:drawLife()
end

return Crate