--[[
-- Debris Class
-- Debris are the little pieces that fall when a Block, a Guardian or the Player are destroyed by an Explosion
-- These little pieces bounce around (using collision:getBounce()) until their internal timer (self.lived)
-- expires. Then they simply disappear.
--]]

local Debris = Entity:extend("Debris")

local minSize = .5
local maxSize = 2.5
local minVel = -50
local maxVel = (-1 * minVel)
local bounciness = 0.1

--something is wrong here, runs slooow, freezes game
function Debris:__init__(kwargs)
  kwargs.w,kwargs.h = 
    math.random(minSize, maxSize), math.random(minSize, maxSize)
  local map = kwargs.map
  if map.gore <= 0 then
      return
  end
  Entity.__init__(self,kwargs)
  
  self.elasticity = bounciness

  self.r, self.g, self.b = kwargs.r,kwargs.g,kwargs.b
  if kwargs.color then
      self.color = kwargs.color
      local col = getColor(kwargs.color)
      self.r, self.g, self.b = col[1],col[2],col[3]
  end
  self.lifeTime = 1 + 3 * math.random()
  self.lived = 0
  local minVel = kwargs.minVel or minVel
  local maxVel = kwargs.maxVel or maxVel
  self.vx = math.random(minVel, maxVel)
  self.vy = math.random(minVel, maxVel) --log("woo")
  self.isBackground = true
    if not self.passive then
        --hitchhiking
        self.map:storeBullet(self)
    end
  self.drawBlood = 0
  self.shadows = {}
  self.gorey = self.map.gore>1
end

function Debris:filter(other)
  if self.gorey and other.isTile then return "cross"
  elseif self.gorey then return end
  if other.isDeubris or not other.isBackground then return "bounce" end
end

function Debris:moveColliding(dt)
  local world = self.world

  local future_l = self.x + self.vx * dt
  local future_t = self.y + self.vy * dt

  local next_l, next_t, cols, len = world:move(self, future_l, future_t, self.filter)
  local col
  local nxtile = self.map:getTileP(next_l,next_t) or nil
  if self.tile and not self.tile.isBackground then
      self._oldx, self._oldy = self.x,self.y
     elseif self.tile then
     self._ooldx,self._ooldy = self.x,self.y
    end
  for i=1, len do
        col = cols[i]
        if self.gorey then
            if not col.other.isBackground then
                self.killme = true
                if self.color == "red" and nxtile and not nxtile.isBackground
               --and self.drawBlood>-2 and self.tile and not self.tile.isBackground and self._oldx and self.myy 
                then
                    self:drawStreak(self.tile,{
                        ox=(self._oldx) or col.other._x,oy=self._oldy or col.other._y,x=next_l,y=next_t,w=self.w,h=self.h,color={self.r,self.g,self.b}
                    })
                    self.drawBlood = 0
                end
            elseif self.killme and self.gorey then
                self.lived = self.lifeTime+1
            end
            if col.other.isBackground then
        
                if self.color == "red" and nxtile and  nxtile.isBackground

                then
                    local ww = self.w/2
                    self:drawStreak(self.tile,{
                        ox=(self._ooldx or self._oldx)+ww,oy=(self._ooldy or self._oldy)+ww,x=next_l+ww,y=next_t+ww,w=self.w,h=self.h,color={self.r,self.g,self.b}
                    })
                    self.drawBlood = 0
                end
             end
    
         else
             self:resolveCollision(col)
         end
  end

  self.x, self.y = next_l, next_t
  if self.tile and not self.tile.isBackground then
    self.myx, self.myy = self.x,self.y
  end
end

function Debris:drawStreak(tile,streak)
    if not tile then
        return
    end
    
    if self.map.gore<2 then
        return
    end
    tile.streaks[#tile.streaks+1] = streak
end

function Debris:update(dt)
  self.lived = self.lived + dt
  self.drawBlood = self.drawBlood+1

  if self.lived >= self.lifeTime  then
    self:destroy()
  else
    self:updateGravity(dt)
    self:moveColliding(dt)
  end
  self:getTile()
  if self.co8lor == "red" and self.drawBlood>-2 and self.tile and not self.tile.isBackground and self._oldx and self.myy then
      self.tile.streaks[#self.tile.streaks+1] = {
          ox=self._oldx,oy=self._oldy,x=self.myx,y=self.myy,w=self.w,h=self.h,color={self.r,self.g,self.b}
      }
      self.drawBlood = 0
  end
end

function Debris:draw()
  --for x,i in ipairs(self.shadows) do
      --util.drawFilledRectangle(i.x,i.y,self.w,self.h, self.r,self.g,self.b)
     --end
  util.drawFilledRectangle(self.x, self.y, self.w, self.h, self.r, self.g, self.b)
end

return Debris
