DungeonDigger = class:extend("DungeonDigger")

function DungeonDigger:__init__(kwargs)
    self.timer = Timer.new()
    
    self._toBuild = {}
    self._builtList = {}
    self._toBuildList = {}
    
    self.map = kwargs.map
    self.seed = kwargs.seed or self.map.seed
    self.attack = kwargs.attack or 1
    self.rate = kwargs.rate or 10
    self.attackRate = (kwargs.attackRate or .01)/self.rate
    self.tile = kwargs.tile
    if self.tile then
        self.tile:destroy(true) self.tile.color = getColor("blue")
    end
    self.grow = kwargs.grow
    self.instant = kwargs.instant
    if self.instant == nil then 
        self.instant = true
    end
    
    --contains all rooms by id
    self.rooms = {}
    --contains all tiles in room by id of the room
    self.tilesInRooms = {}
    --contains all turrets
    self.turrets = self.map.turrets
    
    self.digH = kwargs.digH or kwargs.h or self.map.maxY-self.map.minY-1--50
    self.digW = kwargs.digW or kwargs.w or self.map.maxX-self.map.minX--50
    self.w, self.h = self.digW, self.digH
    self.digOptions = kwargs.digOptions or {}
    --If typenis cellular then this is probability that a cell will be a floor
    self.floorProp = kwargs.floorProb
    
    self.digOptions.corridorHeight = {2,4}
    self.digOffset = kwargs.digOffset or {x=self.tile.x-1,y=self.tile.y-1}
    
    self.turretCount = kwargs.turretCount or kwargs.turrets or {2,3}
    if type(self.turretCount)=="table" then
        self.turretCount = math.random(self.turretCount[1], self.turretCount[2])
    end
    
    local types = {"Uniform","BrogueCave","Digger","Arena"}
    local typesTo = {
        dungeon = "Uniform",
        cave = "BrogueCave",
        mine = "Digger",
        maze = "Maze",
        arena = "Arena",
        none = "none"
    }
    local mazes = {"IceyMaze","DividedMaze"}
    self._tunnelType = kwargs.tunnelType or self.map.mapType
    if self._tunnelType == "random" then
        self._tunnelType = lume.randomchoice(types)
    else
        self._tunnelType = typesTo[self._tunnelType]
    end
    if self._tunnelType == "Maze" then
        self._tunnelType = lume.randomchoice(mazes)
    end
    
    self.digOptions.caveChance=0
    self.purpose = false
    self.target = nil
    if not self.map.isClient then
        self:activate()
    end
end

function DungeonDigger:activate()
    self.target = nil
    
    local none
    if self._tunnelType == "none" then
        none = 1
        self._tunnelType = "Uniform"
    end
    
    local function saveTunnel(...) return self:saveTunnel(...) end
    if self._tunnelType == "BrogueCave" then
        self._tunnelType = "Brogue"
        self.digOptions.caveChance = 1
    end
    self._tunnel = rot.Map[self._tunnelType](self.digW, self.digH, self.digOptions)
    if self._tunnelType == "Cellular" then
        self._tunnel:randomize(self._floorProp or self.digOptions.floorProp or .6)
    end
    
    if none then
        --return
    end
    
    self.purpose = true
    self._tunnel:create(none and null or saveTunnel)
    
    if self.instant then
        if self._tunnel.isCave then
            self:fixIfCave()
        end
        if not self._tunnel.uisCave then

            self:assignRoomIDs()-- log(inspect(self:getRandomRoom(),1).."!")
            
            for turc = 1,0 do--??self.turretCount do
                local i = self:getRandomTile(1)-- assert(not i.isBackground,i.type)
                self.turrets[#self.turrets+1] = Turret:new({map=self.map,x=i.x*i.w,y=i.y*i.h})
                self[string.format("turret%s",turc)] = self.turrets[#self.turrets]
            end
            
        end
        self.purpose = false
    end
    self.tile = self._toBuildList[1]
end

function DungeonDigger:saveTunnel(_x, _y, val)
    if val == 1 then return end--don't include walls
    local x = _x + self.digOffset.x
    local y = _y + self.digOffset.y
    self._toBuild[x] = self._toBuild[x] or {}
    self._toBuild[x][y] = val
    local t = self.map:getTile(x,y) if not t then return end
    if self.instant then t:destroy(true) t.tcolor = getColor("cyan") end
    local pos = {x=x,y=y,val=val}
    table.insert(self._toBuildList, pos)
end

function DungeonDigger:getTargetTile()
    if #self._toBuildList<0 then
        self.purpose = false
        return
    end
    
    local t
    local targ
    local sight = 30
    local fov = self.map.fov--.terrainFov
    local tile = self.target or self.tile
    local checkForTarget = function(...) return self:checkForTarget(...) end
    
    fov:compute(tile.x, tile.y, sight, checkForTarget)
    if self.target then
        return self.target
    elseif self.grow then-- error("No arget")
        for x = 1, #self._builtList do
            t = self._builtList[x]
            fov:compute(t.x, t.y, sight, checkForTarget)
            if self.target then
             --   warn("calculated")
                self.goal = {x=t.x, y=t.y}
                break
            end
        end
        if self.target then
            return self.target
    elseif not self.grow then
            targ = lume.randomchoice(self._toBuildList)
            self.target = self.map:getTile(targ.x,targ.y)
            self.goal = {x=self.target.x,y=self.target.y}
            self.shooting = true warn("!")
            return self.target
        end
    end
    if not self.target then error() end
end  

function DungeonDigger:checkForTarget(x, y, range, vis)
    local target = self.map:getTile(x,y)
    
    if self._toBuild[x] and self._toBuild[x][y] then
        self.target = self.map:getTile(x,y)
        if self.target.isBackground then
            self.target = nil
            return
        end
        return true
    end
    
end

function DungeonDigger:scheduleTile(tile) self.sch=tile if 5 then return 5 end
    local function attackTile()
        tile.life = tile.life - self.attack
        --stop action if tile is destroyed
        if tile.isBackground then
            return false
        end
    end
    self.timer:every(self.attackRate,attackTile)
end

function DungeonDigger:update(dt)
    self.timer:update(dt)
    if self.sch then
        self.sch.life = self.sch.life - self.attack
    end
    if self.purpose then
    
        --target is destroyed
        if self.target and self.target.isBackground then
            self._toBuild[self.target.x][self.target.y] = nil
            for x = 1, #self._toBuildList do
                dd = self._toBuildList[x]
                if dd.x == self.target.x and dd.y == self.target.y then
                    table.remove(self._toBuildList,x)
                    table.insert(self._builtList,dd)
                    break
                end
            end
            self.target = nil
        end
        
        if not self.target then 
            ProFi:start()
            self:getTargetTile()
            ProFi:stop()
            ProFi:writeReport("ProFi.txt")
            ProFi:start()
            self:scheduleTile(self.target)
            ProFi:stop()
            ProFi:writeReport("ProFi2.txt")
        end
        
    end
end

function DungeonDigger:getRoom(id)
    --tile is a corridor--????
    if not id then
        return
    end
    if type(id) == "table" and id.isRoom then return id end
    if type(id) == "table" then
        id = id.roomID
    end
    return self.rooms[id]
end

function DungeonDigger:addDoorToMap(x,y)
    tile = self.map.grid[x][y]
    tile.source = "door.png"
    tile.type  = "door"
    tile.isPassable = true
end

function DungeonDigger:lockRoom(dungeon)
    local corridors = self._tunnel:getCorridors()
end

function DungeonDigger:addDoors(room,n)
    local n = n or 1
    for g = 1,n do
        room:addDoor(x,y)
    end
    room:getDoors(function(...) return self:addDoorToMap(...) end )
end

function DungeonDigger:addAllDoorsToMap()
    for _,t in pairs(self._tunnel:getRooms()) do
        t:getDoors(function(...) return self:addDoorToMap(...) end )
    end
    --room:addDoors(isWallCallback)--(x,y,return true?)
end

function DungeonDigger:addTileToRoom(tile,roomID)
    if tile._inserted_in_room then
        return
    end
    local room = (type(roomID) == "table" and roomID) or self:getRoom(roomID)
    local roomID = (type(roomID) == "table" and roomID.id) or roomID
    tile.roomID = roomID
    table.insert(room.tiles,tile)
    if not tile.isBackground then
        tile.type = "wall"
    else
        tile.type = "floor"
    end
    if tile.type == "wall" then assert( not tile.isBackground)
        table.insert(room.walls,tile)
    else
        table.insert(room.spaces,tile)
    end
    tile._inserted_in_room = true
end

function DungeonDigger:_getTile(x,y,room,noUnits)
    local dung = self.map:getTile(x,y)
    if not dung then return dung end
    if dung.units and noUnits then
        dung = nil
    else
        self:addTileToRoom(dung,room)
    end
    return dung
end

local TmpRoom = class:extend("TmpRoom")
function TmpRoom:__init__(map)
    self.w,self.h = map.w,map.h
end
function TmpRoom:getTop()
    return 0
end
function TmpRoom:getBottom()
    return self.h
end
function TmpRoom:getLeft()
    return 0
end
function TmpRoom:getRight()
    return self.w
end

function DungeonDigger:fixIfCave()
    local tmp = TmpRoom:new(self) 
    self._tunnel.getRooms = function() return {tmp} end
end

function DungeonDigger:getRoomTiles(room, noUnits)
    if #room.walls>0 then return {floors = room.spaces, walls = room.walls} end
    
    --check Bottom should actually be top(in kivy coords) else switch
    local spaces = {}
    local dungeon = self.map.grid
    local walls = {}
    
    local top = room:getTop()+self.digOffset.y
    local bottom = room:getBottom()+self.digOffset.y
    local left = room:getLeft()+self.digOffset.x
    local right = room:getRight()+self.digOffset.x
    for t = top-1,bottom+1 do
        for r = left-1,right+1 do
            local dung = self:_getTile(r,t,room,noUnits)
            
            --self:colorTile(dung,"red")
            if dung then
                if dung.type == "wall" then
                    table.insert(walls,{x=r,y=t,[1]=r,[2]=t})
                else
                    table.insert(spaces,{x=r,y=t,[1]=r,[2]=t})
                end
            end
        end
    end
 

    return {floors=spaces,walls=walls}
end

function DungeonDigger:getEmptyRoom()
    local count = 0
    local _repeat = true
    while _repeat do
        if count>100 then
            break
        end
        count = count + 1
        _repeat = false
        local r = self:getRandomRoom()
        check = self:getRoomTiles(r)
        for _,x in ipairs(check.floors) do
            if self.map.grid[x.x][x.y].unit then
                _repeat = true
                break
            end
        end
    end
    return check
end

function DungeonDigger:assignRoomIDs()
    local r = self._tunnel:getRooms()
    for x,room in ipairs(r) do
        room.id = lume.uuid()
        local rand = math.random()*10
        
        self.rooms[room.id] = room
        room.tiles = {}
        room.walls = {}
        room.spaces = {}
        room.isRoom = true
        self:getRoomTiles(room)
    end
end

function DungeonDigger:getRooms()
    return self._tunnel:getRooms()
end 

function DungeonDigger:getRandomRoom(rooms)
    local r
    if not rooms then
        r = self._tunnel:getRooms()
    else
        r = rooms
    end
    local rr = lume.randomchoice(r)
    return rr
end

function DungeonDigger:getRandomSpaceInRoom(room)
    local room = room or self:getRandomRoom()
   -- spaces = self:getRoomTiles(room)
    return self:getEmptyTileInRoom(room)--lume.randomchoice(spaces.floors)
end

function DungeonDigger:getRandomWallInRoom(room)
    local room = room or self:getRandomRoom()
   -- spaces = self:getRoomTiles(room)
    return self:getEmptyTileInRoom(room,true)--lume.randomchoice(spaces.floors)
end

function DungeonDigger:getEmptyTilesInRoom(room,wall)
    local room = room or self:getRandomRoom()
    local spaces = self:getRoomTiles(room)
    if wall then
        return spaces.walls
    else
        return spaces.floors
    end
end

function DungeonDigger:getEmptyTileInRoom(room,wall)
    local tiles = self:getEmptyTilesInRoom(room,wall)
    if #tiles < 1 then
        warn("Tiles to low")
        return self:getRandomTile(true)
    end
    local r = lume.randomchoice(tiles)
    return self.map.grid[r.x][r.y]
end

function DungeonDigger:getRandomTile(allowWalls,anywhere)
    if anywhere then
        local x = math.random(0+self.digOffset.x,self.w)
        local y = math.random(0+self.digOffset.y,self.h)
        return self.map.grid[x][y]
    elseif  not allowWalls then
        return self:getRandomSpaceInRoom()
    elseif type(allowWalls)=="number" then
        return self:getRandomWallInRoom()
    elseif allowWalls then
        local perc = math.random(0,1)
        if perc==0 then
            return self:getRandomSpaceInRoom()
        else
            return self:getRandomWallInRoom()
        end
    end
end

function DungeonDigger:draw()
end

return DungeonDigger