local Mini = Entity:extend("Mini",15000)
local abs = math.abs
local getUpdateCount = getUpdateCount

function Mini.spawnMini(parent, life, data, amount)
    if type(parent)~="table" then
        parent = r_G[parent]:new({parent=self,map=self,passive=true})
    end
    
    return Mini:new({
        parent = data or parent,
        source = parent.source,
        life = life or 5,--last 5 seconds
        map = parent.map or game.map,
        x = parent._x or parent.x,
        y = parent._y or parent.y,
        w = (not parent.isTile and parent.w) or 5,
        h = not parent.isTile and parent.h,
        amount = amount
    })
end
    
    
function Mini:__init__(kwargs)
    self.isMini = true
    
    kwargs.h = kwargs.w
    self.super.__init__(self,kwargs)
    self.source = kwargs.source or "none.png"
    self.parent = kwargs.parent or warn("no parent")
    
    local md
    if self.parent.miniID then md = self.parent.miniID
    elseif self.parent.isTile then md = self.parent.uuid end
    assert(md)
    self.miniID = self.parent.miniID
    self.map._minis = self.map._minis or {}
    --self.map._minis[md] = self.uuid
    self.md = md
    
    self.w = kwargs.w or kwargs.h or 5
    self.h = kwargs.h or self.w
    self.isExplosive = false
    self.elasticity = kwargs.elasticity or .4
    
    self.life = kwargs.life or 30
    self.totalLife = self.life
    self.collisionType = "touch"
    self.amount = kwargs.amount or math.random(10,25)
    if not self.passive then
        --hitchhiking
        self.map:storeBullet(self)
    end
    --self.map.entities[self.uuid] = nil
    if self.parent and self.parent.isTile and not self.parent.minis then
        error(inspect(self.parent,2))
    end
end

function Mini:getSource()
    return self.source
end

function Mini:update(dt, c)
    
    self._w, self._h = resizeImage(game:getSource(self.source),self.w,self.h)
    assert(self.x,self.up)
    self:updateGravity(dt)
    
    local vx, vy = self.x+self.vx*dt, self.y+self.vy*dt
    
    if 1 then
        local mx, my = self:move(vx,vy,dt)
        self.x, self.y = mx,my
    end
    self.up = self.up and self.up+1 or 0
    self:getTile() if self.die then assert(self.die and not self.tile) end
    --self:setAll()
end

function Mini:checkLimits()
    local maxVx,maxVy = 1000,100
    
    if self.vy>0 then
        self.oldOnGround = false
    end
    
    if math.abs(self.vx)>maxVx then
        self.vx = maxVx*self:getDir(self.vx)
    end
    
    if self.vy>0 and self.onGround then
        self.vy = 0
    end
    
    if math.abs(self.vy)>maxVy then
        self.vy = maxVy*self:getDir(self.vy)
    end

end

function Mini:destroy(sent)
    Entity.destroy(self)
    
    --self.map._minis[self.md] = nil
    
    if self.parent.isTile then
        if self.parent.minis then
            self.parent.minis[self.uuid] = nil
            self.parent.totalMinis = self.parent.totalMinis-1
        else
            game.map.player.name = "!!!! minis"
        end
    end
    
    self.active = false
    
    if not sent then
        --self.map.player:send("do",{"destroyMini",self.md,1,nil,nil,"map"})
    end
end

function Mini:changeVelocityByCollisionNormal(nx, ny, bounciness)
  local bounciness = bounciness or self.elasticity
  if not bounciness or bounciness == 0 then
      return
  end
  local vx, vy = self.vx, self.vy

  if (nx < 0 and vx > 0) or (nx > 0 and vx < 0) then
    vx = -vx * bounciness
    self.bounced.x = true
  end
  
  if ((ny < 0 and vy > 0) or (ny > 0 and vy < 0)) and abs(vy)>10 then
    vy = -vy * bounciness
    self.bounced.y = true
  end

  self.vx, self.vy = vx, vy
end

function Mini.checkCollision(item, other)
    if other.isPlayer then
        return nil--"cross"
    elseif other.isTile and not other.isBackground then
   --     other.life = other.life - item.attack
     --   item.die = true
        return item.collisionType
    end
    if other.isMini then
        return "slide"--???
    end
    if not other.isMini and not other.isBackground and other.isSolid then
        return item.collisionType
    end
end

function Mini:move(x,y,dt)
    self.die = false
    self.life = self.life - dt
    
    if self.life<=0 then
        self.die = true
    end
    
    
    if not self.active then error()
        return 0,0
    end
    local actualX, actualY, collisions, len = self.world:move(self,x,y,
    self.checkCollision)
    local cc, other
    for u = 1,len do
        cc = collisions[u]
        other = cc.other
        self.super.resolveCollision(self,cc)
            
    end
    if self.die then
        self:destroy()
    end
    
    return actualX, actualY
end

--todo: draw trail
function Mini:draw()
    local v = game:getSource(self.source)
    local xx,yy = v:getWidth()/2, v:getHeight()/2
    local ca = math.pi^2
    
    love.graphics.draw(game:getSource(self.source),self.x,self.y,0,self._w,self._h)--+xx,self.y+yy,math.pi-(self.angle+ca),
  --  self._w,self._h,xx,yy)
    
end

return Mini