local Nest = GridMap:extend("Nest")
local Tike = Tile
local RECORDS = RECORDS
local choice = lume.randomchoice
local _Path = class:extend("_Path")
function _Path:__init__(x,y)
    self.x = x
    self.y = y
end

function _Path:getX()
    return self.x
end

function _Path:getY()
    return self.y
end

function _Path:getPos()
    return self.x, self.y
end

local sizes = {
    small = 50,
    medium = 200,
    large = 400,
}


           EIGHT={
                { 0,-1},
                { 1,-1},
                { 1, 0},
                { 1, 1},
                { 0, 1},
                {-1, 1},
                {-1, 0},
                {-1,-1}
               }
_colors = {}
addColor = function(x,y)
    if type(x) == "table" then return addColor(x._x or x.x, x._y or x.y) end
    table.insert(_colors,{x=x,y=y})
end

function Nest:destroy(destroyPeer,noSwitch)

    game.map = nil
    game.currentMap = nil noil=4
    game.player = nil
    self.isOld = true
    self.active = false
    self:unpause()
    
    for x,i in ipairs(game.players) do
        i:crush()
    end
    
    --destroy peer?
    if destroyPeer and self:hasPeer() then
        self:hasPeer():destroy()
    end
    
    game.players = {}
    self.player = {}
    gooi.removeComponent(self.pauseMenu)
    gooi.removeComponent(self.message)

    done = {}

    Gs.__stack = {}
    GS.old = nil
    
    if not noSwitch then
        GSS(mainMenu)
        mainMenu:create()
    end
    
end

function Nest:unUnit(tile, ent)
    if tile then
        tile.block = nil
        ent.tile = nil
        lume.remove(tile.units, ent)
        if tile.unit==ent then
            if tile.units[1] then
                tile.unit = tile.units[1]
            else
                tile.unit = nil
            end
        end
    end
end

function Nest:addUnit(tile, ent)
    if tile then
        tile.units[#tile.units+1] = ent
        ent.tile = tile
        tile.unit = ent
    end
end

function Nest:log(l)
    log(l)
end

function Nest:dopause()
    if self.score then
        self.score:destroy()
    end
    
    self.score = ScoreDisplay:new({
        map = self,
        data = {3,4},
        h = H()-(H()/4),
        w = W()-(W()/4)
    })
    self.score:create()
end

function Nest:unpause(ignore)
    
    if self.player and self.player.dead and not ignore then
        return
    end
    
    if self.score then
        self.score:destroy()
        self.score = nil
    end
end

function Nest:setMatchTime(time)
    self.matchTime = self.timer:after(time or self.time, function()
        if not self.isClient then
            if self.host then
                self.host:sendToAll("do",{"sendAwayStats"})
                self.tryToEnd = true
            else
                self.canEndGame = true
            end
            self:endGame()
        end
    end)
end

function Nest:getScores()
        if 1 then return {} end
        --returns players in order of their scores
        
        local pl,ts
        local score
        local scores = {}
        local tscores = {}
        local randomchoice = lume.randomchoice
        
        for x, i in ipairs(game.players) do
            score = i:getScore()
            table.insert(scores, score)
            tscores[score] = tscores[score] or {}
            table.insert(tscores[score],i)
        end
        
        table.sort(scores)
        
        local players = {}
        for x = 1,#scores do
            ts = tscores[scores[x]]
            
            --so tied players randomly come higher
            pl = randomchoice(ts)
            lume.remove(ts,pl)
            
            table.insert(players,pl)
        end
        return players

end

function Nest:endGame()
    --wait to get achievements
    if self.host and not self.canEndGame then
        return
    end
    
    if self.score then
        self.score:destroy()
    end
    
    local function openAchievements()
        if not self.am then
            self.am = AchievementsMenu:new({
                map=self,
                h = H()-(H()/4),
                w = W()-(W()/4)
             })
            self.am:load()
        end
        self.am:create()
        self.score = self.am
    end
    
    if self.match == "deathmatch" then
        local player = self:getScores()[1]
        gooi.alert({
            text=string.format("%s is the winner!",player.name),
            ok=openAchievements
        })
        --endMenu
    
    elseif self.match == "team deathmatch" then
        
    elseif self.match == "king pin" then
    
    elseif self.match == "capture the flag" then
    
    end
    
    if self.host then
        self.host:sendToAll("do",{"endGame",nil,nil,nil,nil,"map"})
        openAchievements()
        self.am:destroy()
    else
        openAchievements()
    end
end

function Nest:start()
    assert(not self.isOld)
    if self.size then
        self.w = sizes[self.size]
        self.h = self.w
    end
    game.progress.total = self.w*self.h+(((self.w+self.h)*10)+100)
    mainMenu:progress()
    
    self.countG = 0
    

    self.entities[game.player.uuid] = game.player

    if not self.isClient then
        self:spawnPlayers()
    end
    
    self.player = game.player or {uuid=1} --????
    
    local pw = W(10)
    local B_W = W()/5.5
    local ww  = B_W/1
    local h   = H()/7
    local three = .1
    local two = .2
    local four = 1
    local five = 2
    
    self.cameraButton = gooi.newButton({
        group=self.player.uuid,
        text="",
        icon=game:getSource("camera.png")
    }):onRelease(function()
        local sc = self.camera.scale
        self.camera.scale = (sc==three and four) or (sc==four and five)  or (sc==five and two)
            or (sc==two and three) or .2
        self.player:fixCamera(true)
        end)
    
    self.pauseButton = gooi.newButton({
        group=self.player.uuid,
        text="",
        icon=game:getSource("pause.png"),
        --x=W()/2-pw/2,
        --h=H(50),
        --w=pw
    }):onRelease(function()
        self:dopause()
    end)
    self.pauseMenu = gooi.newPanel({
        group=self.player.uuid,
        layout="grid 1x2",
        x=W()/2-ww/2,
        w=ww,
        h=h
    }):add(
        self.pauseButton,
        self.cameraButton
    ):setStyle({bgColor=getColor(self.player.color,.5)})

    self.message = gooi.newLabel({x=W()*0,y=H()/7,w=W(),group=self.player.uuid}):setText(""):center()

    self:load()
    if mainMenu then--.drawProgress then
        mainMenu.drawProgress = false
        if self.host then
            self.host:sendToAll("activate")
        end
        GSS(game.currentMap)

    
    end
    --[[bys = {}
    for x = 1,4 do
        local t = randomchoice(self.allDungeons)
        local tt = t:getRandomTile()
        local b = Crawler:new({
            map = self,
            x=tt._x,
            y=tt._y,
            h=tt.h,
            w=tt.w,
            vx = 60,
            vy = 60
        })
        bys[x] = b
        self.player:getTile()
        --b.boid.target = {x=self.player.tile._x,y=self.player.tile._y}
        bj = b
    end]]
end

function Nest:__init__(kwargs)
    local kwargs = kwargs or {}
    GridMap.__init__(self,kwargs)
    observeGlobals()
    self.seed = kwargs.seed or math.random()
    rot.setSeed(self.seed)
    self.timer = Chrono()
    
    self.world = bump.newWorld()
    
    self.dungeonWidth = kwargs.dungeonWidth or kwargs.dW or 50
    self.dungeonHeight = kwargs.dungeonHeight or kwargs.dH or self.dungeonWidth
    
    --extra size to add to dungeons
    self.extra = kwargs.extra or math.floor(self.dungeonWidth/2)
    
    --defaults to 3 by 3 dungeons (9 dungeons)
    self.size = kwargs.size or 2
    
    --stores all dungeons in a grid
    self.dungeons = {}
    self.diggers = {}
    
    --stores all dungeons
    self.allDungeons = {}
    
    self.tileWidth = 16
    self.tileW = self.tileWidth
    
    self.tileHeight = self.tileW
    self.tileH = self.tileHeight
    
    self.tW, self.tH = self.tileW, self.tileH
    tW, tH = self.tW, self.tH
    
    self.totalW = (self.dungeonWidth+self.extra)*self.size
    self.totalH = (self.dungeonHeight+self.extra)*self.size
    
    --stores all entities
    self.entities = {}
    self.creatures = {}
    
    --stores entities that aren't blocks
    self.notBlocks = {}
    
    
    self.camera.scale=2.3
    self.cameraMan = nil
    self.vintage = true--cuts some costs and makes game run faster
    
    self.storeBullet = self.store
    self.active = 1
    
    self.fogOfWar = 1
    
    self.stateMachine = StateMachine:new({
        map = self
    })
    
    self.spawner = Spawner:new({map=self})
    self.noise = NoiseManager:new({map=self})
    self.soundGen = SoundGenerator:new({map=self})
    
    self.ingoreWallsForFOV = false
    
    self:build()
    
    self.camera.scale = 2
    --self.fov = rot.FOV.Recursive(function(...) return self:lightCallback(...) end)
    --function fo()
    --self.fov = rot.FOV.Precise(function(...) return self:lightCallback(...) end)
    --end
end

function Nest:addNoise(...)
    return self.noise:addNoise(...)
end

function Nest:setSeed(seed)
    self.seed = seed or math.random()
    rot.setSeed(self.seed)
end

function Nest:getItem(uuid)
    return self.entities[uuid]
end

function Nest:updateEntity(data)
    local obj,val = self.entities[data.uuid]
    for x = 1,#self.toSet do
        val = self.toSet[x]
        obj[val] = data[val]
    end
    return obj
end

function Nest:getUpdateData(entity)
    if type(entity) == "string" then
        entity = self.entities[entity]
    end
    
    if entity.getData then
        return entity:getData()
    else
        error("Entity "..tostring(entity.__class).." can not get data")
    end
end

--cuts all connections to entity
function Nest:kill(entity)
    if entity.isBot then
        self.spawner:destroy(entity)
    end
    self.entities[entity.uuid] = nil
    self.creatures[entity.uuid] = nil
    self:unUnit(entity.tile,entity)
end

function Nest:store(ent)
    if self.active then
        self.entities[ent.uuid] = ent
        if ent.isBot then
            self.creatures[ent.uuid] = ent
        end
    end
end

function Nest:spawnItem(item, room, y, real)
    local tile
    
    local it
    if type(item) == "string" then
        item = items[item]
    elseif item.__init__ then--item is made
        it = item
    end
    
    if type(room) == "number" and not real then
        tile = self:getTile(room,y)
    elseif type(room) ~= "number" then
        tile = self.player.tile.dungeon:getEmptyTileInRoom(room)
    end
    
    if real then
        tile = {_x=room,_y=y}
    end
    
    if not it then
        it = Item:new({
            map = self,
            data = item,
            x = tile._x,
            y = tile._y,
        })
    else
        it.x = tile._x
        it.y = tile._y
    end
    
    
    it:activate()
    it:onSpawn(tile)
    return it
end

function Nest:spawnTorch()--remove--????
    return self:spawnItem("torch",self.player.tile.room)
end

spt = function() return self:spawnTorch() end
sspt = function(k) for x = 1,(k or 10) do spt() end end
function Nest:sendMessage(mes,color,isPrivate)
     if 1 then return end
    local _color,_mes = color or game.settings.color,mes
    if type(_color) == "string" then
        color = getColor(_color)
    elseif _color then
        error("Color should be text")
    end
    
    mes = lume.wordcut(mes)--self._mw/(self.font:getWidth(mes)*2))
    local n,x, messes
    for x = 1,#mes do
        n = x+#self.messages
        if n>self.messageLen.h then
            messes = lume.copy(self.messages)
            for xx = 1,#messes do
                self.messages[xx] = messes[xx+1]
            end
            self.messages[-1] = nil
            n = self.messageLen.h
        end
        
        self.messages[n] = {mes[x],color}--.."\n"
    end
    
    if not isPrivate and self:hasPeer() then
        if self.client then
            self.client:sendMessage(_mes,_color)
        elseif self.host then
            self.host:sendMessage(_mes,_color)
        end
    end
end

function Nest:build()
    self.size = 3--?
    local dung = self.dungeons
    local defaultTile = {
        isWall = true
    }

    local options = {
        rooms = {10,11}
    }
    for x = 1,self.size do
        dung[x] = {}
        for y = 1, self.size do
            if x~=2 or y~=2 then
                dung[x][y] = DungeonDigger:new({
                    w = self.dungeonWidth+self.extra,
                    h = self.dungeonHeight+self.extra,
                    x = (x-1)*self.dungeonWidth,
                    y = (y-1)*self.dungeonHeight,
                    type = randomchoice({"Uniform","Digger"}),
                    options = options,
                    map = self
                })
            --Make middle dungeon maze
            else
                dung[x][y] = DungeonDigger:new({
                    w = self.dungeonWidth+self.extra,
                    h = self.dungeonHeight+self.extra,
                    x = (x-1)*self.dungeonWidth,
                    y = (y-1)*self.dungeonHeight,
                    type = "maze",
                    options = options,
                    map = self
                })
                dung["middle"] = DungeonDigger:new({
                    w = self.dungeonWidth/2,
                    h = self.dungeonHeight/2,
                    x = (x-1)*self.dungeonWidth+math.floor(self.dungeonWidth/4),
                    y = (y-1)*self.dungeonHeight+math.floor(self.dungeonHeight/4),
                    type = "BrogueCave",
                    options = options,
                    map = self
                })
                dung[x][self.size+1] = dung["middle"]--for looping purposes
            end
            self.diggers[#self.diggers+1] = dung[x][y]
            self.allDungeons[#self.allDungeons+1] = dung[x][y]
            
        end
    end
    
    
    --breaking unnecessary walls to free space
    local list = {{0,1},{1,0},{0,-1},{-1,0},{-1,-1},{1,1},{-1,1},{1,-1}}
    for x= 1,self.totalW do
        for y = 1,self.totalH do
            local t = self.grid[x] and self.grid[x][y]
            if t then
                --check if tile is surrounded by other tiles,
                --if so then destroy it to limit size
                t.isBackground = not t.block
                local surrounded = true
                for _,i in ipairs(list) do
                    local tt = self.grid[x+i[1]] and self.grid[x+i[1]][y+i[2]]
                    if tt and not tt.digWall then
                        surrounded = false
                        break
                    end
                end
                if surrounded and t.block then
                    t.block:destroy()
                    t.block = nil
                end
            end
        end
    end
    
    --tiles can't be made anymore
    self.locked = true
end

function Nest:destroyItem(uuid,...)
    self.entities[uuid]:destroy(...)
end

function Nest:seeThrough(obj)
    if obj.isBackground then return true
    elseif obj.isPlayer then return true
    end
end

function Nest:isWalkable(x,y)
    
    local val = x--self.grid[x.x or "none"] and self.grid[x.x][x.y]
    --log(inspect(val,1))
    if not val then return false end
    if val.block then return end
    if val.unit and not (self:seeThrough(val.unit)) then return end
    if val.unit and val.unit.isBackground then return true
    end
    return true
end
Do=4

ggxx=4
Ve = 2
local forPlayer
--returns false if light can't pass through tile
function Nest:lightCallback(fov, x,y)
    local tile = self.grid[x] and self.grid[x][y]
    if not tile then return end
    
    if self.ignoreWallsForFOV then
        return true
    end
    
    local ptile = self.player.tile

    local abs = math.abs
    
    if forPlayer
    and self.player:ignoreTile(tile) 
    and not tile.light then
        assert(not tile.light)
        return
    end
    
    if tile.parent == self.player and tile.isBackground then return true end
    if tile._0unit then
        if tile.unit~=self.player and not tile.unit.isBullet
        then
            if not tile.unit.active then
                return error(tostring(tile==tile.unit.tile)..","..tostring(tile.unit.tile)..tile.unit.class)--inspect(tile.unit,1))
            end
            return false
        end
    end
    if not nil then
        if tile.unit and not tile.unit.isBackground and not tile.unit.seeThrough then
            return false
        end
        if tile.unit and not(self:seeThrough(tile.unit)) then return false end
    else
        if tile.units[1] and tile.units[1].isBlock and not tile.units[1].isBackground then
            return false
        end
    end
    
    return true
end

function Nest:reflectCallback(lighting, x, y)
    local tile = self.grid[x] and self.grid[x][y]
    if not tile then return 0 end
    if tile.unit and not tile.unit.isPlayer and not tile.unit.isBackground then return 0 end
    return (lightt or 0.3)
end

function Nest:addLight(x, y, color)
    
    local tile
    local color = color or getColor("red",true)
    if type(color)=="string" then
        color = getColor(color,true)
    end
    
    if type(x) == "number" then
        tile = self.grid[x] and self.grid[x][y]
    else
        tile = x.tile or x
    end
    if not tile then return 0 end
    self.lighting:setLight(tile.x, tile.y, color)
end
lig = function(col) self:addLight(self.player.tile.x, self.player.tile.y, col) 
self:computeLight() end

local map
local lightingFunc = function(x, y, color)

    local tile = map.grid[x] and map.grid[x][y]
    if not tile then return end
    tile.color = getColor(color)
    log("got "..x..","..y..inspect(color)..inspect(tile.color))
end

function Nest:computeLight()
    map = self
    self.lighting:compute(lightingFunc)
end

function Nest:getRandomTileInDungeon(digger)
    local digger = digger or choice(self.diggers)
    return digger:getRandomTile()
end


function Nest:spawnPlayers() --log(inspect(game.players,2))
    local pl
    for x = 1,#game.players do
        pl = game.players[x]
        self:spawnPlayer(pl)
    end
    self.player = game.player
end


function Nest:spawnPlayer(pl,x,y)

    local x = x
    local y = y
    local tile
    
    --local tile = self:getTile(x,y)--,nil,true)
    if not x then
        tile,l,i = nil,0
        while (not tile or not tile.dug) do
            tile = self:getRandomTileInDungeon()--choice(self.allTiles)
            --i,l = self.world:querySegment(tile._x,tile._y,self.w,self.h*2,filterBackground)
        end
    else
        tile = self:getTile(x,y)
    end
    pl=pl or 1
    if type(pl)=="number" then
        log("numb!"..pl)
        pl = game.players[pl]
    else
    
    end
    
    local player = pl or game.player or Player:new({map=self})
    player:setPos(tile._x, tile._y)
    player.life = 100
    player.dead = false
    
    tile.isBackground = true
    x,y = tile.x,tile.y
    self.player = game.player
    --self.player.update=null
    tile.source = "background_dirt.png"
    --tile.collider = nil
  --  for r,o in ipairs(getRound(5)) do 
    --    local t = self:getTile(x+o.x,y+o.y)
    --    if t then t:destroy(9) end
   -- end
    
    self.player:fixCamera(false)
    
    if self.host then
        self.host:sendToAll("spawnPlayer",{pl.index or 1,x,y})
    end
    
end

local function filterBackground(item)
    return not item.isBackground
end

function Nest:playerDied(player,index)
    if player.real then
        self.respawnTime = self.timer:after(10, function()
            self:respawnPlayer(player.index)
        end)
        --player:send("do",{"die",index,nil,nil,nil,player.index})
        self:dopause()
    end
end

local round = lume.round
function Nest:getRespawnTime()
    return 10-round(self.respawnTime.time)
end

function Nest:respawnPlayer(playerIndex)
    local player = game.players[playerIndex]
    self.respawnTime = nil
    if self:hasPeer() then
        if self.client then
            self.client:send("requestRespawn",player.index)
        elseif self.host then
            self:doRespawn(player)
        end
    else
        self:doRespawn(player)
    end
end

function Nest:doRespawn(player)
    self:spawnPlayer(player)
end


local currentDraw = {}
local func = function(x,y)
    local t = game.map:getTile(x,y)
    local self = game.map
    if t then
        t.pride = RECORDS.updateCount+1
        --if self.player:ignoreTile(t) and not t.light then
            --return
        --end
    

        table.insert(currentDraw,t)

    end
    --log(x..","..y)
end

function Nest:update(dt)--log(dt.." is dt")
    --self.tickRate = self.tickRate+dt
    --[[if self.tickRate>=self.updateRate then
        self.tickRate = 0
        if self.slow then
            dt = self.updateRate
        end
    elseif not returnTick then 
        --return 
    end]]
    media.cleanup()
    RECORDS.updateCount = RECORDS.updateCount+1
    self.timer:update(dt)
    
    local trash = self.score and self.score:update(dt)
    local dt = dt or 0--??o
    if track then ProFi:start() end
    
    self.spawner:update(dt)
    self.noise:update(dt)
    self.soundGen:update(dt)
    

    
    if not self.camera.shaking then
        self.camera:setFollowStyle("none")
    end
    if self.cameraMan then
        self.camera:follow(self.cameraMan.x,self.cameraMan.y)
    end
    self.camera:update(dt)
    --self.map:update(dt)
    
    if self.cameraMan and self.cameraMan.animation then
        self.cameraMan.animation:update(dt)
        if self.cameraMan.animation.done then
            self.cameraMan.animation = nil
        end
    end
    
    local trash = self.player.paddy and self.player.paddy:update(dt)
    
    
    --self:updateChunks(dt,RECORDS.updateCount)

    --fix??!!
    if nil1 then
        local visibleThings, len, obj = self.world:queryRect(self.camera:getVisible())
        --table.sort(visibleThings,sortTiles)


        local up = getUpdateCount()
        for i=1, len do
            obj = visibleThings[i]
            obj._inView = up
            obj:update(dt,RECORDS.updateCount)
        end
    end
    
    local visibleThings = currentDraw
    local len = #currentDraw--, len = self.world:queryRect(self.camera:getVisible())
    
    local tile
    for i=1, len do
        tile = visibleThings[i]
        for x, obj in ipairs(tile.units) do
            --local obj = tile.units[x]
            if not obj.isPlayer and not obj.isBot then
                obj:update(dt)
            end--lighting is strange!!
        end
    end
    
    for i, m in pairs(self.creatures) do
        m:update(dt)
    end
    
    currentDraw = {}
    

    local player
    self.player:update(dt)
    for i = 1,#game.players do
        local player = game.players[i]
      --  player:update(dt,RECORDS.updateCount)
    end

    if self.player.kickedOut then
        log("woah, out")
        ff()--??
        gooi.alert({text="Kicked out!"})
    end
    

    
    self.cameraMan = self.cameraMan or {x=self.player.x,y=self.player.y}
    if GF then
  --  self.cameraMan = {x=game.players[1].x,y=game.players[1].y}
    end
    
    if self.player.followX then
        self.cameraMan.x = self.player.x
    end
    if self.player.followY then
        self.cameraMan.y = self.player.y
    end
    local player = self.player 
    if ggy then player = bj end
    if self.player.target and self.zelda then
        player = self.player.target
    end
    self.cameraMan.x = player._x or player.x
    self.cameraMan.y = player._y or player.y
    
    local maxZoom = 10
    local minZoom = 0.05
    if self.camera.scale >  maxZoom then
        self.camera.scale = maxZoom-1
        self.cameraMan.animation= nil
    elseif self.camera.scale < minZoom then
        self.camera.scale = minZoom+0.01
    end
    
    if self.fogOfWar then
        local tile = self.player.tile
        if not self.player.tile then
            warn("No tile!")
        else
        
            if not tile.isBackground then
                tile = self:getTile(tile.x, tile.y-1)
            end
        
            ---locak d =(self,player.faceY or 0{self.player.flipX,
            forPlayer = true
            self.fov:compute(tile.x,tile.y,self.player.sight,func)
        end
        forPlayer = false
        --table.sort(currentDraw)
    end
    
    if self.tryToEnd then
        if self.canEndGame then
            self.tryToEnd = false
            self:endGame()
        end
    end
    
    
    --if peer then peer:update(dt) peer.pause = nil end
    
    if track then ProFi:stop() ProFi:writeReport("update.txt") error() end 
end
function ch()    seekNewGlobals(function(n) error(inspect(n)) end) 
end

function Nest:kickOutPlayer(pl,sent)
    if type(pl)=="number" and pl~=self.player.index then
        pl = game.players[pl]
    end
    if pl.kickedOut then return end
    
    if not self.host and not sent then
        return
    elseif self.host then
        self.host:sendToAll("do",{"kickOutPlayer",pl.index,1,nil,nil,"map"})
        if not pl.index then error(inspect(pl,1)) end
        local cl = self.host:getClientByIndex(pl.index-1)
        cl:disconnectNow()
    end
    log("kicking out!")
    pl:destroy()
    pl.kickedOut = 1
    --game.players[pl.index] = {}
end

function Nest:collectGarbage(allow)
    if 1 then--not self.isClient or allow then
        log("trying collection...")
      --  collectgarbage("collect")
        log("garbage collected.")
        if self.host then
       --     self.host:sendToAll("do",{"collectGarbage",true,0,0,0,"map"})
        end
    end
end

function cg(n) collectgarbage("step",n) end
function sw(b) bullets.SmokescreenBullet = b or Bullet end

local function sortTiles(a,b)
    return a.timeMade<b.timeMade
end

function Nest:drawGlow(obj)
    if not obj.tile then return end
    local glow = game:getSource("glow.png")
    local ow, oh = obj.tile.w*2,obj.tile.h*2--obj.w+obj.w/2, obj.h+obj.h/2
    if obj.isMini then
  --      ow,oh = 16,16
    end
    local w,h = resizeImage(glow, ow, oh)
    local x,y = obj.x,obj.y--obj.x-obj.w/2, obj.y-obj.h/2
    love.graphics.draw(glow, x-obj.tile.w/2,y-obj.tile.h/2,0,w,h)
end

local function filter(item)
    return item.isTile
end

function Nest:drawObj(obj)
        obj._inView = self.up
        local r,g,b,a = self.r,self.g,self.b,self.a
        if not obj.isBullet and not obj.isPlayer and not obj.isPuff and not obj.isMini then
            if obj.glow then
                self:drawGlow(obj)
            end --if obj.isFake then obj:draw() end--?!! end
            if self.fogOfWar then
                if obj.isTile and obj.pride and obj.pride>=RECORDS.updateCount then
                    Tile.draw(obj)--visibleThings[i]:draw()
                elseif obj.isTile then
                    love.graphics.setColor(1,1,1,.5)
                    Tile.draw(obj)--visibleThings[i]:draw()
                    love.graphics.setColor(r,g,b,a)

                elseif obj.isDebris or (obj.tile and obj.tile:isVisible()) then
                    obj:draw()
                end
            else
                obj:draw()
            end
        end
end

local maxdis
function Nest:draw(dt)
    if dtrack then ProFi:start() end
    local x, y, w, h = self.camera:getVisible()--Window()
    --error(x..", "..y..", "..w..", "..h)
    --x, y = self.camera:toCameraCoords(x,y)
    --w,h = w*self.camera.scale, h*self.camera.scale
    

    self.camera:attach()
    --[[for nx,i in pairs(self.colorTiles) do
        self:_colorTile(game:get(nx),i)
    end]]
    --self:drawChunks()
    
    local obj
    local up = getUpdateCount()
    local r,g,b,a = love.graphics.getColor()
    self.r,self.g,self.b,self.a=r,g,b,a
    self.up=up
    --if 1 then table.sort(visibleThings) end
    
    
    if ggxx then
    local visibleThings, len = self.world:queryRect(self.camera:getVisible())
    for i=1, len do
        obj = visibleThings[i]
        obj._inView = up
        if obj.glow then
            self:drawGlow(obj)
        end  
        if 1 then--(obj.tile and self:tileIsVisible(obj.tile)) then
            obj:draw()
        end
    end
    end
    local visibleThings = currentDraw
    local len = #currentDraw--, len = self.world:queryRect(self.camera:getVisible())
    
    maxdis = lume.distance(
        self.player.x-self.player.sight*tW,
        self.player.y-self.player.sight*tH,
        self.player.x,
        self.player.y
    )
    local r,g,b,a = love.graphics.getColor()
        
        
    local tile --log("size at "..len)
    for i=1, len do
        tile = visibleThings[i]
        self:drawTile(tile,true)
    end
    
    love.graphics.setColor(r,g,b,a)
    for x,i in ipairs(_colors) do
    if ggx then love.graphics.rectangle("fill",i.x,i.y,tW,tH) end
    end

    local player
    for i = 1,#game.players do
        local player = game.players[i]
        if player.glow then
            self:drawGlow(player)
        end
        if player.behind then
            local t = player.tile
            local t1 = self:getTile(t.x-1,t.y)
            local t2 = self:getTile(t.x+1,t.y)
            player:draw()
            if not t.isBackground then
                self:drawTile(t,t.dis)
            end
            if not t1.isBackground then
                self:drawTile(t1,t1.dis)
            end
            if not t2.isBackground then
                self:drawTile(t2,t2.dis)
            end
            love.graphics.setColor(r,g,b,a)

        else
            player:draw()
        end
    end


    if doDebug and ggxx and debugBump then
        drawBump(self.world)
    end
    
    self.camera:detach()
    
    self.camera:draw()
    
    trash = self.score and self.score:draw()
    if doDebug then
        love.graphics.print(inspect(love.graphics.getStats()),100,H()-450,0,2,2)
        love.graphics.print(string.format("FPS: %s",love.timer.getFPS()),100, H()-150,0,2,2)
        love.graphics.print(collectgarbage("count")/1024,100, H()-100,0,2,2)
        love.graphics.print(love.graphics.getStats().texturememory/(1024*1024),100,H()-50,0,2,2)
    end
    if self.drawProgress then
        game.progress:draw()
    end
    
    if dtrack then ProFi:stop() ProFi:writeReport("udraw.txt") error(tt) end

end

function Nest:tileIsVisible(tile)
    return (tile.pride and tile.pride >= RECORDS.updateCount) or not self.fogOfWar
end

function Nest:isVisible(...)
    return self:tileIsVisible(...)
end

function Nest:setProgress(n)
    local p = game.progress
    p:setProgress(n)
end

function Nest:progress(n)
    if not self.drawProgress then
        self.drawProgress = true
    end
    local p = game.progress
    p:progress(n)
end

function Nest:stopProgress()
    self.drawProgress = false
end


function Nest:drawTile(tile,dis)
    if tile then
        local dis = lume.distance(tile._x, tile._y, self.player.x, self.player.y)
        tile.dis = dis
        local aa = dis/maxdis
        local light = .7
        local light2 = 0
     --   if ddis then
       --     light2 = 1
    --    end
        
        if lightAll or tile.light then--??
            light = tile.light or light or 1
            light2 = 1
        end
    
        if not tile.color then
            local p = (1-aa)*light
            love.graphics.setColor(p,p,p,p+light2)
        else
            love.graphics.setColor(tile.color)
        end
    
        if not units then
          --  return
        end
        
        for x = 1,#tile.units do
            local obj = tile.units[x]
            if not obj.isPlayer then
                obj:draw()
            end
            --if #tile.units>1 and tile.units[1].isBlock and tile.units[2].isBlock then warn("tile"..tostring(tile)) end
        end
    end
end

function Nest:updateChunks(dt, c)
    local xval, yval
    local _
    for _ = 1, #self.allCurrentChunks.x do
        xval = self.allCurrentChunks.x[_]
        --[[if self.allChunks[xval-5] then
            self.allChunks[xval-5] = nil
        end
        if self.allChunks[xval+5] then
            self.allChunks[xval+5] = nil
        end]] 

        for _y = 1, #self.allCurrentChunks.y do
            yval = self.allCurrentChunks.y[_y]
            if (not self.allChunks[xval] or not self.allChunks[xval][yval]) and 
            not self.loadedAll
            then
                self:loadChunks({start=xval,en=xval},{start=yval,en=yval})
                local chunk = self.allChunks[xval][yval]
                --???self:load()
                chunk:update(dt,c)
            elseif (self.allChunks[xval] and self.allChunks[xval][yval]) then
            
                local chunk = self.allChunks[xval][yval]
                chunk:update(dt,c)
            end
        end
    end
end

function Nest:getTile(_tile,x,y,dontSave,er)
    if type(_tile) == "number" then
        local y = x
        local x = _tile
        local t = (self.grid[x] and self.grid[x][y])
        if not t and not self.locked then
            self.grid[x] = self.grid[x] or {}
            self.grid[x][y] = {x=x,y=y,_x=x*self.tileW,_y=y*self.tileH,units={}}
            table.insert(self.allTiles,self.grid[x][y])
            return self.grid[x][y]
        else
           return t
        end
    end
    if self.locked then
        return
    end
    error("Invalid args: "..tostring(_tile)..", "..tostring(x)..", "..tostring(y))
    
end

function Nest:getTileP(x,y)
    return self:getTileWithPos(x,y)
end



function Nest:getAdjacentTiles(x,y)
    local tiles = {}
    for _,x in lume.iter(getRound(1)) do
        local tile = self.tiles[x+c[1]] and self.tiles[x+c[1]][y+c[2]]
        if tile then
            table.insert(tiles,tile)
        end
    end
    assert(#tiles~=0,"Couldn't get adjacent tiles.")
    return tiles
end

function Nest:drawChunks()
    local xval, yval
    for _ = 1, #self.allCurrentChunks.x do
        xval = self.allCurrentChunks.x[_]
        for _y = 1, #self.allCurrentChunks.y do
            yval = self.allCurrentChunks.y[_y]
            if not self.allChunks[xval] or not self.allChunks[xval][yval] then
            else
                local chunk = self.allChunks[xval][yval]
                chunk:draw()
            end
        end
    end
end


function Nest:findPathTo(obj,x,y)

    local obj = obj
    local x = x
    local y = y
    if type(obj) ~= "table" then
        x = obj
        y = x
        obj = y
    end
    
    local floor = math.floor
    local _x,_y = (obj.getPos and obj:getPos()) or obj.x, obj.y
    
    --start = {x=floor(_x),y=floor(_y)}
    
    if not self.grid[floor(x)] then log("no tile,returning") return end
    if not self.grid[floor(x)][floor(y)] then log("no tile, returning") return end
    if self.grid[floor(x)][floor(y)].collidable == true then
        log("collidable")
        return
    end
    obj._goto = self.grid[floor(x)][floor(y)]
    self.c_unit = obj


    local function check(x,y)
        return self:isWalkable(x,y)
    end
    
    local path = {}
    local function buildPath(x,y)
        table.insert(path,_Path:new(x,y))
    end
    
    goal = {x=x,y=y}
    start = {x=_x,y=_y}
    local pth = rot.Path.AStar(goal.x,goal.y,check)
    pth:compute(start.x,start.y,buildPath)
    return path
end

function Nest:locad()
    local fov = function(fov,x,y) return self:lightCallback(fov,x,y) end
    self.fov = rot.FOV.Precise:new(fov, {topology=4})
    
    local tfov = function(fov,x,y) return self:tLightCallback(fov,x,y) end
    self.terrainFov = rot.FOV.Precise:new(tfov)

end

function Nest:mousereleased(x, y, button)
    local trash = self.score and self.score:mousereleased(x,y,button)
    if trash then
        self.buttonPressed = true
    end
    --self.player:released()
    gooi.released()
end

function Nest:mousepressed(x, y, button)
    local trash = self.score and self.score:mousepressed(x,y,button)
    if trash then
        self.buttonPressed = true
    end
   -- self.player:pressed()
    gooi.pressed()
end

function Nest:touchpressed(id, x, y, dx, dy, pressure)
    self.super.touchpressed(self,id,x,y,dx,dy,pressure)
    
    local trash = self.player.pressed and self.player:pressed(id,x,y,dx,dy,pressure)
    local trash = (self.score and self.score:touchpressed(id,x,y,dx,dy,pressure)) or trash
    if trash then
        self.buttonPressed = true
    end
    local x,y = self.camera:toWorldCoords(x,y)
   
    --local tile = self:getTileWithPos(x,y)
    if tile then
        log(inspect(tile.color)..'\n'..inspect(tile.source))
        --self.player:placeBlock(0,0,nil,tile)
    end
end

function Nest:touchmoved(id,x,y,dx,dy,pressure)
    local t = self.player.touchmoved and self.player:touchmoved(id,x,y,dx,dy,pressure)
    local x,y = self.camera:toWorldCoords(x,y)
   
    --local tile = self:getTileWithPos(x,y)
    if tile then
        log(inspect(tile.color)..'\n'..inspect(tile.source))
        --self.player:placeBlock(0,0,nil,tile)
    end
end

function Nest:touchreleased(id, x, y, dx, dy, pressure)
    self.super.touchreleased(self,id,x,y,dx,dy,pressure)
    
    local trash = self.player.released and self.player:released(id,x,y,dx,dy,pressure)
    
    trash = (self.score and self.score:touchreleased(id,x,y,dx,dy,pressure)) or trash
    if trash then
        self.buttonPressed = true
    end
    if not self.buttonPressed then
    else
        self.buttonPressed = false
        self.ftile = false
        --log("??????")
        return
    end
    
    local x,y = self.camera:toWorldCoords(x,y)
   
    local tile = self:getTileWithPos(x,y)
    if tile and ggx then
        local s = ""
        _colors={}
        for x,i in ipairs(tile.units) do
            s = s..","..i.class..":"..tostring(i.isBackground)
        end
        local s = tile.x..","..tile.y
        addColor(tile)
        log(s)--....color)..'\n'..inspect(tile.source))
        --self.player:placeBlock(0,0,nil,tile)
    end
    if self.stealthInspection and tile then
        self.ftile = tile
        log(tile.life)
    end
    
    local sx,sy = self.map:convertPixelToTile(x,y)
    sx = math.floor(sx) ndd=4
    sy = math.floor(sy)
    --bj.boid.target = {x=x, y=y}
    --bj.boid.path = nil
end

function Nest:textinput(text)
  gooi.textinput(text)
end

function Nest:keypressed(key, scancode, isrepeat)
  gooi.keypressed(key, scancode, isrepeat)
  if key == "escape" then
    p:close()
    love.event.quit()
  end
end

function Nest:keyreleased(key, scancode)
  gooi.keyreleased(key, scancode)
end

return Nest