local Player = Entity:extend("Player")
local getUpdateCount = getUpdateCount

local floor = math.floor

Player.res = res

local selectedStyle = {
    bgColor = colors.green
}

local anim = {
	imageSrc = "players/gopher.png",
	defaultState = "running",
	states = {
		-- 1st line
		running = { -- the name of the state is arbitrary
			frameCount = 4,
			offsetX = 0,
			offsetY = 0,
			frameW = 40,
			frameH = 64,
			nextState = "running", -- we loop the running state
			switchDelay = 0.1
		},
	}
}
function Player:reload(name, color, hat, hat2, hatCol, hat2Col, stomachColor, gunColor)
    local t1 = self.hat and self.hat:destroy()
    local t2 = self.hat2 and self.hat2:destroy()
    
    self.color = color or game.settings.color or "chocolate"
    self.stomachColor = stomachColor or game.settings.stomachColor or "white"
    self.gunColor = gunColor or game.settings.gunColor or "white"
    self.name = name or game.settings.name
    
    self.hat = hats[hat or game.settings.hat]:new({self})
    self.hat2 = hats[hat2 or game.settings.hat2]:new({self})
    self._hat = hat or game.settings.hat
    self._hat2 = hat2 or game.settings.hat2
    
    self._hat2Col = hat2Col or game.settings.hat2Color
    self._hatCol = hatCol or game.settings.hatColor
    self.hat:setColor(hatCol or game.settings.hatColor)
    self.hat2:setColor(hat2Col or game.settings.hat2Color)
    
    self.backShooting = game.settings.backShooting
    self.timeAsKingpin = 0--????
    
    if self.real then
        local n = self.getChild(self.displayBullets,1)
        self._bul(n, true)
        self.bulletStorage[1][1] = mainMenu.bulletType
        self.bulletStorage[1][2] = 1500
        self.currentBullet = self.bulletStorage[1][1]
    end
    if self.display then
        game:saveSettings()
    end
end

function Player:bleeding()
    if self.life <= 15 then
        return true
    end
end

function Player:inMist()
    return nil--self.madePuff
end

function Player:__init__(kwargs)
    Entity.__init__(self,kwargs)
    --table.insert(game.players,self)
    kwargs = kwargs or {}
    --self._kwargs = kwargs
    game.debugLabel.group=self.uuid
    
    self.data = kwargs.data or {}
    self.client = kwargs.client
    self.host = kwargs.host
    self.index = kwargs.index
    
    self.kills = 0
    self.deaths = 0
    
    self.tShoot = false
    
    self.color = kwargs.color or game.settings.color or "chocolate"
    self.stomachColor = kwargs.stomachColor or game.settings.stomachColor or "white"
    self.gunColor = kwargs.gunColor or game.settings.gunColor or "white"

    self.source = kwargs.source or 
        kwargs.spriteImage or kwargs.sprite or game.settings.playerSource
        
        
    self.dropletsImage = game:getAsset("players/droplets.png")
    self.tearsImage = game:getAsset("players/tears_anim.png")
 
    self:feel("normal")
        
        
    self.tearSprite = animx.newActor():
        addAnimation("default",{
            img = self.tearsImage,
            tileWidth = 40*res,
            delay = .4
        }):loopAll():switch("default")
        
    self.drawTears = false
    self.drawDroplets = false
    
    self.handImage = game:getSource(
        kwargs.handSource or game.settings.handSource)
    self.stomachImage = game:getSource(
        kwargs.stomachSource or game.settings.stomachSource)
    self.gunImage = game:getSource(
        kwargs.gunSource or game.settings.gunSource)

    self.angle = 0---2--1+math.pi/2--kwargs.angle or 0
    self.x = kwargs.x or 0
    self.y = kwargs.y or 0
    self.h = kwargs.h or 16
    self.w = kwargs.w or 16
    if not self.passive then
        self.world:remove(self)
        --so player can fit through tunnels of 1 tile space
        self.world:add(self,self.x,self.y,self.w-2,self.h-2)
    else
        --self.real = false
    end
    
    self.sight = kwargs.sight or 10
    self.paddy = Paddy:new({joysticks=true})
    self.isPlayer = true
    self.Ljoystick = self.paddy.Ljoystick
    self.Rjoystick = self.paddy.Rjoystick
    
    --self.collider = self.map.world:newRectangleCollider(self.x,self.y,self.w,self.h)

    self.speed = {x=0,y=0}
    self.weight = 750
    self.gravity = 0
    self.initGravity = 400
    self._initGravity = self.initGravity
    
    self.lifeBar = LifeBar:new({parent=self,w=self.w+(self.w/2)})
    self.life = 100

    self._speed = 100
    self.bullets = {}
    self.bulletLen = 0
    
    self.jumpAcceleration = 300--800
    self.runAccel = 500
    self.brakeAccel = 1000--50--2000
    self.maxVx = 200
    self.maxVy = 200
    self.maxCrawlX = 100
    self.maxCrawlY = 100
    self.isCrawling = false
    self.wallOnSide = false
    self.dead = false
    self.timeAsKingpin = 0
    self.team = self.team or self.uuid
    
    --stats
    self.stats = {}
    self.stats.jumpsPeformed = 0
    self.stats.timeInAir = 0
    self.stats.timeInDangerRange = 0
    self.stats.dustProduced = 0
    self.stats.blocksPlaced = 0
    self.stats.numberOfTimesPaused = 0
    self.stats.blocksDestroyed = 0
    self.stats.minisCollected = 0
    self.stats.killedSelf = 0
    
    self.stats.shotsFired = 0
    local sh = "%ssShot"
    for x, i in pairs(bullets) do
        self.stats[string.format(sh, x)] = 0
    end
    
    self.old = {x=self.x,y=self.y} self._beforeDrawTrash = {}
    self._afterDrawTrash = {}
    self._drawTrash = {}
    self._updateTrash = {}
    
    self.flipX = 1
    
    self.bulletStorage = {
        {mainMenu.bulletType or "Bullet",1500},
        --{"GrenadeBullet",120}
    }
    
    self.blockStorage = {
        {"Tile",15}
    }
    
    self.Tmini4 = 0
    local function bul(n, load,block, mini)
        local storage, display
        if block then
            storage = self.blockStorage
            display = self.displayBlocks
        else
            storage = self.bulletStorage
            display = self.displayBullets
        end
        
        if type(n) == "number" then-- if n==4 then warn(4) end
            n = self.getChild(display,n)
        end
        
        local u = storage[n.data]
   
        if not u then
            return
        end
        
        if self:getStorageIndex(storage,u[1]) == 4 then
            if load and not block then
                n:setText(u[2])
                n:setSource(self.getBulletSource(bullets[u[1]]))
            elseif load and block then
                n:setText(u[2])
                n:setSource(tiles[u[1]]:new({map=self.map}))
            end
            return
        end
        
        
        local d = storage[4]
        if d then
            local uu, tmpBu
            if not block then
                tmpBu = bullets[u[1]]:new({parent=self,map=self.map,passive=true})
                if self.tmpMini then
                    self.tmpMini.parent = tmpBu
                    self.tmpMini.amount = u[2]
                    self.tmpMini.source = tmpBu.firstSource
                end
            
                uu = u
                u = d
                
                if self.currentBullet == uu[1] then
                    self.currentBullet = u[1]
                end
                
                load = true
            else
                tmpBu = tiles[u[1]]:new({map=self.map})
                if self.TtmpMini then
                    self.TtmpMini.parent = tmpBu
                    self.TtmpMini.amount = u[2]
                    self.TtmpMini.source = tmpBu.source
                end
            
                uu = u
                u = d
                
                if self.currentBlock == uu[1] then
                    self.currentBlock = u[1]
                end
                
                load = true
            end
            
            storage[n.data] = d
            storage[4] = uu
            
            local v = self.getChild(display,4)
            v:setText(n.text)
            v:setIcon(game:getSource(tmpBu.source))
        end
        
        if load then
            n:setText(u[2])
            if not block then
                n:setSource(self.getBulletSource(bullets[u[1]]))
            else
                local t = tiles[u[1]]:new({map=self.map})
                --t.source = mini.source
                n:setSource(t.source)
                t:destroy()
            end
        end
        
        self.moveTgo =  null
        if not load then
            self.map:setMessage(string.format("You equipped %s",u[1]))
            if not block then
                self.currentBullet = u[1]
                if self.oldCu then
                    self.oldCu:setStyle(n.style)
                end
                n:setStyle(selectedStyle)
                self.oldCu = n
                --self.currentBlock = nil
                self.Rjoystick:setDigital(false)
            else
                self.currentBlock = u[1]
                if self.oldBu then
                    self.oldBu:setStyle(n.style)
                end
                n:setStyle(selectedStyle)
                self.oldBu = n
               -- self.currentBullet = nil
               -- self.Rjoystick:setDigital(true)
                assert(self.currentBlock)
            end
        end
    end
    
    local red = {bgColor=getColor("red")}
    local green = {bgColor=getColor("green")}
    
    local blo = function(n,load,block, mini) return bul(n,load,true, mini) end
    local function canc(n)
        self.breakBlocks = not self.breakBlocks
        if self.breakBlocks then
            n:setStyle(green)
        else
            n:setStyle(red)
        end
    end
    self._blo = blo
    self._bul = bul
    
    local B_W = _W/5.5
    --self.sprite.switch = null
    
    self.displayBullets = gooi.newPanel({
        x=0,y=0+50,layout="grid 1x4",
        w = _W/5.5,
        h = _H/10,
        group = self.uuid
    }):add(
        gooi.newImageButton({drawtext=true,data=1,source="none.png"}):onRelease(bul),
        gooi.newImageButton({drawtext=true,data=2,source="none.png"}):onRelease(bul),
        gooi.newImageButton({drawtext=true,data=3,source="none.png"}):onRelease(bul),
        gooi.newImageButton({drawtext=true,data=4,source="none.png"}):onRelease(bul)
    ):setStyle({bgColor=getColor(self.color,.5)})
    
    self.displayBlocks = gooi.newPanel({
        x=W()-B_W,y=0+50,layout="grid 1x5",
        w = B_W,
        h = _H/10,
        group = self.uuid
    }):add(
        gooi.newImageButton({drawtext=true,data=1,source="none.png"}):onRelease(blo),
        gooi.newImageButton({drawtext=true,data=2,source="none.png"}):onRelease(blo),
        gooi.newImageButton({drawtext=true,data=3,source="none.png"}):onRelease(blo),
        gooi.newImageButton({drawtext=true,data=4,source="none.png"}):onRelease(blo),
        gooi.newImageButton({drawtext=true,data=5,text="X"}):onRelease(canc)
    ):setStyle({bgColor=getColor(self.color,.5)})
    self.displayBlocks.sons[5].ref:setStyle(red)
    
    self.getChild = function(t,index)
        return t.sons[index].ref
    end
    
    self.getBulletSource = function(bul)
        return bul:new({parent=self,map=self.map,passive=true}).firstSource
    end
    
    local n = self.getChild(self.displayBullets,1)
    local nb = self.getChild(self.displayBlocks,1)
    bul(n,true)
    
    blo(nb,true)
    n:setStyle(selectedStyle)
    self.oldCu = n

    nb:setStyle(selectedStyle)
    self.oldBu = nb
    
    self.currentBullet = self.bulletStorage[1][1]
    self.currentBlock  = self.blockStorage[1][1]
    
    if not self.real then
        gooi.removeComponent(self.displayBlocks)
        gooi.removeComponent(self.displayBullets)
    end
    
    
    self:reload()
end

function Player:feel(emotion)
    
    if emotion == "crying" then
        self.drawTears = true
        emotion = "whimpering"
    elseif emotion == "whimpering" then
        self.drawTears = false
        self.drawDroplets = true
    else
        self.drawTears = false
        self.drawDroplets = false
    end
    
    self.sprite = game:getEmotionSprite(emotion)
    self.spriteImage = game.emotionImages[emotion]
    
end

function Player:feelRandom()
    self:feel(lume.randomchoice(game.emotions))
end

function Player:crush()
    gooi.removeComponent(self.displayBlocks)
    gooi.removeComponent(self.displayBullets)
end

function Player:sendPrivateMessage(...)
    return self.map:sendPrivateMessage(...)
end

function Player:sendMessage(...)
    return self.map:sendMessage(...)
end

function Player:send(name, data)
    local p = self.map:hasPeer()
    if not p then
        return
    end

    data = data or {}
    if type(data)=="table" then
        data[6] = data[6] or self.index
    end
    if p and self.map.client then
        p:send(name, data)
        return true
    elseif p then
        data [6] = 1
        p:sendToAll(name, data)
    end
    return nil
end

function Player:kill(sent,index)
    if not sent and self.map.isClient then
        return
    end
    
    self.kills = self.kills+1
    
    if self.map.host then
        self.map.host:sendToAll("do",{"kill",true,me and index ,nil,nil,self.index})
    end
    
    if index==self.index then
        self.stats.killedSelf = self.stats.killedSelf+1
    end
    
end

function Player:addToWorld()
    self.world:add(self,self.x,self.y,self.w-2,self.h-2)
end

function Player:die(bullet)
    self.deaths = self.deaths+1
    self.dead = true
    local bullet = bullet or {}
    p = bullet.parent or bullet
    if not p.isPlayer then p = p.parent end
    
    if p and p.isPlayer then
        local index = p.index
        p:kill(nil,index)
    end
    
    self:spawnDebris({"red",self.color})
    self.map:playerDied(self)
    self.map.world:remove(self)--disappear
end

function Player:setScore(kills, deaths)
    self.kills = kills
    self.deaths = deaths or self.deaths
end

function Player:getScore()
    local score   = -100
    local match = self.map.match
    if match == "deathmatch" or match == "team deathmatch" then
        score = self.kills-self.deaths
    elseif match == "kingpin" then
        score = self.timeAsKingpin
    end
    return score
end

local deathm_info = "%s    | kills: %s | deaths: %s | score: %s"
local kingp_info  = "%s    | time as king pin: %s"
function Player:printScore()
    local match = self.map.match
    
    if match == "deathmatch" or match == "team deathmatch" then
        return string.format(
            deathm_info,self.name, self.kills, self.deaths, self.kills-self.deaths
        )
    
    elseif match == "kingpin" then
        return string.format(
            kingp_info,self.name, self.timeAsKingpin
        )
    end
end

function Player:getAchievement()
    local allAwards = {}
    local done = {}
    
   --[[ self.stats.jumpsPeformed = 0
    self.stats.timeInAir = 0
    self.stats.timeInDangerRange = 0
    self.stats.dustProduced = 0
    self.stats.blocksPlaced = 0
    self.stats.numberOfTimesPaused = 0
    self.stats.blocksDestroyed = 0
    self.stats.minisCollected = 0
    self.stats.killedSelf = 0
    
    self.stats.shotsFired = 0]]
    local sh = "%ssShot"
    
    local stats = self.stats
    
    local stats = {}
        --[[{player index, val}
        
        mostJumpsPeformed=0,
        leastJumpsPeformed=0,
        
        mostTimeInAir=0,
        leastTimeInAir=0,
        
        mostTimeInDangerRange = 0,
        leastTimeInDangerRange = 0,]]
        
        
        
        
    for xx = 1,#game.players do
        pl = game.players[xx]
        pl.stats.playerKilled = pl.kills
        for x, i in pairs(pl.stats) do
            if type(x) ~= "string" then
                goto continue
            end
            if type(i) == "number" then
                local str = x:sub(1,1):upper()
                str = string.format("%s%s",str,x:sub(2,-1))
                local least = string.format("least%s",str)
                local most  = string.format("most%s",str)
                
                if not stats[least] or i<stats[least][2] then
                    stats[least] = {xx,i}
                end
                
                if not stats[most] or i>stats[most][2] then
                    stats[most] = {xx,i}
                end
            end
            ::continue::
        end
    end
    
    for xx = 1,#game.players do
    
        pl = game.players[xx]
        for x, i in pairs(stats) do
            if i[1] == xx then
                pl.stats[x] = true
            end
        end

        pl.awards = {}
        pl.nawards = {}
        
        if pl:mostAcrobatic() then
            table.insert(pl.awards,"was the most acrobatic.")
        end
        if pl:bravest() then
            table.insert(pl.awards,"was the bravest.")
        end
        if pl:coward() then
            table.insert(pl.nawards,"was a coward.")
        end
        if pl:mostStrategic() then
            table.insert(pl.awards,"was the most strategic.")
        end
        if pl:mostBrutal() then
            table.insert(pl.awards,"was the most brutal.")
        end
        if pl:mostDistracted() then
            table.insert(pl.nawards,"was the most distracted.")
        end
        if pl:mostCommitted() then
            table.insert(pl.awards,"was the most committed.")
        end
        if pl:architect() then
            table.insert(pl.awards,"was an architect.")
        end
        if pl:aNobody() then
            table.insert(pl.nawards,"was a nobody.")
        end
        if pl:reckonedForce() then
            table.insert(pl.awards,"was a force to be reckoned with.")
        end
        if pl:punchingBag() then
            table.insert(pl.awards,"was a punching bag.")
        end
        if pl:suicidal() then
            table.insert(pl.nawards,"was suicidal.")
        end
        if pl:triggerHappy() then
            table.insert(pl.awards,"was the most trigger happy.")
        end
        if pl:mostFrustrated() then
            table.insert(pl.nawards,"was the most frustrated.")
        end
        if pl:mostLazy() then
            table.insert(pl.nawards,"was lazy.")
        end


        local pawards = lume.copy(pl.awards)
        local pnawards = lume.copy(pl.nawards)
        
        for x,i in ipairs(pl.awards) do
            if done[i] then
                lume.remove(pawards, i)
            end
        end
        for x,i in ipairs(pnawards) do
            if done[i] then
                lume.remove(pnawards, i)
            end
        end
            
        for count = 1,2 do
            if #pawards>0 then
                local award = eliminate(pawards)
                if not done[award] then
                    self.map.am.add(string.format("%s %s",pl.name,award))
                    table.insert(allAwards, string.format("%s %s",pl.name,award))
                    done[award] = true
                end
            end
            if #pnawards>0 then
                local naward = eliminate(pnawards)
                if not done[naward] then
                    self.map.am.add(string.format("%s %s",pl.name,naward))
                    table.insert(allAwards, string.format("%s %s",pl.name,naward))
                    done[naward] = true
                end
            end
        end
    end
    
    if self.map.host then
        self.map.host:sendToAll("do",{"setPlayerAwards",allAwards})
    end

end

function Player:sendAwayStats() warn("sending stats")
    self:send("setPlayerStats", self.stats)
end

function Player:setPlayerAwards(awards, nawards)
    --self.awards = awards
    --self.nawards = nawards
    self.map:endGame()
    for x, i in ipairs(awards) do
        self.map.am.add(i)
    end
    self.map.am:load()
end

function Player:mostAcrobatic()
    return self.stats.mostJumpsPeformed and self.stats.mostTimeInAir
end
function Player:bravest()
    return self.stats.mostTimeInDangerRange
end
function Player:coward()
    return self.stats.leastTimeInDangerRange and self.stats.leastShotsFired
end
function Player:mostStrategic()
    return self.stats.SmokescreenBulletsShot>10 and self.stats.dustProduced>10 and
    self.stats.blocksPlaced>15
end
function Player:mostBrutal()
    return self.mostShotsFired and self.stats.mostPlayersKilled
end
function Player:mostDistracted()
    return self.stats.numberOfTimesPaused>0 and self.stats.mostNumberOfTimesPaused
end
function Player:mostCommitted()
    return self.stats.committed
end
function Player:mostDestructive()
    return self.stats.mostBlocksDestroyed
end
function Player:architect()
    return self.stats.mostBlocksPlaced
end
function Player:aNobody()
    return self.stats.leastShotsFired and self.stats.playersKilled==0
end
function Player:reckonedForce()
    return 0--60% of kills later --???
end
function Player:punchingBag()
    return self.stats.playersKilled == 0 and self.stats.mostDeaths
end
function Player:suicidal()
    return self.stats.killedSelf>0
end
function Player:triggerHappy()
    return self.stats.mostShotsFired
end
function Player:mostFrustrated()
    return self.stats.shotsFired>10 and self.deaths>1 and self.stats.playersKilled==0
end
function Player:mostLazy()
    return self.stats.shotsFired<5
end

function Player:changeTeamTo(color)
    self.team = color
    local w = mainMenu.waitButtons[self.index or 1]
    if w.text == self.name then
        w.team = color
        w:setStyle(mainMenu.styles[color])
    end
end


function Player:getStorageIndex(storage, bul)
    local bul = bul
    local b
    for x = 1, #storage do
        b = storage[x]
        if b[1] == bul then
            return x
        end
    end
    error(inspect(storage)..bul)
    warn(self.currentBullet)
    --error(inspect(bul,1))--fix ohhh????--???
end

function Player:getBulletIndex(bul)
    
    return self:getStorageIndex(self.bulletStorage, bul or self.currentBullet)
end

function Player:collectBullet(mini)
    local bl = #self.bulletStorage
    local data = {mini.parent.__class, mini.amount}
    for x, b in ipairs(self.bulletStorage) do
        if x~= 4 and data[1] == b[1] then
            b[2] = b[2]+data[2]
            self.getChild(self.displayBullets,x):setText(b[2])
            return
        end
    end
    
    if bl<3 then
        self.bulletStorage[bl+1] = data
        self._bul(bl+1,true)
        
    else
        self.bulletStorage[4] = data
        self._bul(4,true)
        self.tmpMini = mini
        --used to keep track of collision
        self.mini4 = 0
        return true
    end
    
    return
end


function Player:collectBlock(mini)
    local bl = #self.blockStorage
    local data = {mini.parent.__class, mini.amount}
    for x, b in ipairs(self.blockStorage) do
        if x~= 4 and data[1] == b[1] then
            b[2] = b[2]+data[2]
            self.getChild(self.displayBlocks,x):setText(b[2])
            return
        end
    end
    
    if bl<3 then
        self.blockStorage[bl+1] = data
        self._blo(bl+1,true,ni,mini)
        
    else
        self.blockStorage[4] = data
        self._blo(4,true)
        self.TtmpMini = mini
        --used to keep track of collision
        self.Tmini4 = 0
        return true
    end
    
    return
end

function Player:checkLimits()
    local currentBullet = bullets[self.currentBullet] or {}
    local maxVx,maxVy = self.maxVx, self.maxVy
    if currentBullet.isDelayBullet then
        if self.shot then
            maxVx, maxVy = 10,10
        end
        if self.vy>0 then
                self.oldOnGround = false
        end
    
        if self.isCrawling then
            maxVx, maxVy = self.maxCrawlX, self.maxCrawlY
        end
    
        if math.abs(self.vx)>maxVx then
            self.vx = maxVx*self:getDir(self.vx)
        end
    
        if self.vy>0 and self.onGround then
            self.vy = 0
        end
    
        if math.abs(self.vy)>maxVy then
            self.vy = maxVy*self:getDir(self.vy)
        end
    
    --never do
    elseif nil and currentBullet.isFlyBullet then
        if self.shot then
            maxVx, maxVy = 10,10
        end
        if self.vy>0 then
                self.oldOnGround = false
        end
    
        if math.abs(self.vx)>maxVx then
            self.vx = self.maxVx*self:getDir(self.vx)
        end
    
        if self.vy>0 and self.onGround then
            self.vy = 0
        end
    
        if math.abs(self.vy)>maxVy then
            self.vy = self.maxVy*self:getDir(self.vy)
        end
    
    else
        if self.shot then
            --disabled
            --slows down player while shooting
            --maxVx, maxVy = self.maxVx/4,100
        end
        
        if self.vy>0 then
            self.oldOnGround = false
        end
    
        if self.isCrawling then
            maxVx, maxVy = self.maxCrawlX, self.maxCrawlY
        end
    
        if math.abs(self.vx)>maxVx then
            self.vx = maxVx*self:getDir(self.vx)
        end
    
        if self.vy>0 and self.onGround then
            self.vy = 0
        end
    
        if math.abs(self.vy)>maxVy then
            self.vy = maxVy*self:getDir(self.vy)
        end
    end

end

function Player:shoot(angle, bullet, joyX, joyY)
    local gb, bul

    if self.real then
        gb = self:getBulletIndex()
    
        bul = self.bulletStorage[gb]
        if bul[2]<1 then
            self.shot = false
            return
        end
    else
        self.currentBullet = bullet
        self.Rjoystick.angle = angle
    end
    
    local newBullet = bullets[self.currentBullet]:new({
        map = self.map,
        x = self.x+self.w/2,
        y = self.y,
        weight = 10,
        ddelay=.5,
        parent = self,
        angle = angle,
        joyX = joyX or 1,
        joyY = joyY or 1--????
    })
    
    if not newBullet.active then
        self.shot = false
        return
    end
    
    self.drawSpark = true
    local sh = "%ssShot"
    self.stats.shotsFired = self.stats.shotsFired+1
    local shh = string.format(sh,self.currentBullet)
    self.stats[shh] = self.stats[shh]+1
    if self.currentBullet == "FlyBullet" then
        self.canFly=1
    end
    
    if self.real then
        self.bulletLen = self.bulletLen+1
        bul[2] = bul[2]-1
        self.getChild(self.displayBullets,gb):setText(bul[2])
        self:send("do",{"shoot",angle,self.currentBullet})
    end
    
    return newBullet
end

function Player:placeBlock(x, y, block,_tile)
    local tile = tile or self.tile
    if not self.real then
        if not _tile then--not sent
            warn("Wanted to place a fake block at player "..tostring(self.index))
            return
        end
        
        self.currentBlock = block
        
        local t = self.map:getTile(x,y)
        local xx,yy = x,y
        if t then--and t.isBackground and t~=self.tile then--and #t.units<1 then
            t:changeTo(tiles[self.currentBlock]:new({
                map=self.map,
                tmp=true,
                passive=1,
                _x=xx*self.tile.w,
                _y=yy*self.tile.h,
                parent=self}))
            t.parent = self
        end
        return
    end
    
    local block = self.currentBlock
    assert(self.currentBlock)
    assert(x<=1 and y<=1)
    local gb = self:getStorageIndex(self.blockStorage,self.currentBlock)
    local blo = self.blockStorage[gb]
    
    local xx,yy = tile.x, tile.y
    local t = _tile or self.map:getTile(xx+x,yy+y)
    
    if blo[2]<1 or not t.isBackground then
        return
    end
    
    --pushes player to it's tile position to avoid glitchy placement.
    --resets when block not placed for 2 seconds
    if self.real then--and not self.adjusted then
        self.x = tile._x
        self.y = tile._y
        self.map.world:update(self,self.x,self.y)
    end
    self.adjusted = 1
    
    if t and t.isBackground and not t.unit then--and #t.units<1 then
        t:changeTo(tiles[self.currentBlock]:new({
            x=t.x,
            y=t.y,
            map=self.map,
            tmp=true,
            parent=self,
            passive=1,
            real=1,
        }))
        blo[2] = blo[2]-1
    
        self.getChild(self.displayBlocks,gb):setText(blo[2])
        
    elseif t and doT then
        t:getAttacked(.5)
        self.paddy.ignoreButton = .1
        return
    end
    
    self.stats.blocksPlaced = self.stats.blocksPlaced+1
    if self.real then
        self:send("do",{"placeBlock",t.x,t.y,block,true})
    end
    
end

function Player.checkCollision(item, other)
    if other.isBackground then
        return nil
    elseif other.parent==item and other.ignoresParent then
        return
    elseif other.isGrenadeBullet then
        return "bounce"
    elseif other.isBullet and other.parent == item then
        return
    elseif other.isBullet then
        return "cross"
    elseif other.isFire then
        return --"cross"
    elseif other.isMini then
        return "cross"
    elseif other.isPuff then
        return
    elseif not item:ignore(other) then
        return "slide"
    end
end

function nn()
pl=game.player.sprite
pll=pl.switch
pl.switch = function(self,s) if s=="running" then error() end end
end

function Player:changeVelocityByKeys(dt, dir, dirY)
    self.isJumpingOrFlying = false

    if not self.active then return end

    local vx, vy = self.vx, self.vy

    if dir then
        dir, dirY = dir, dirY
    else
        dir, dirY = self.Ljoystick:getPos()
    end
    
    if not self.isBot and self.real then
        if (IM:isDown("Left")) then
            dir = -1
        elseif (IM:isDown("Right")) then
            dir = 1
        elseif love.keyboard.isDown("right") then error("Something bad happened in controls") end
    end
    
  
    if self.shot then
    
        vx = vx/2
    end
  
    if dir~=0 then
        self.sprite:switch("running")
        ptt = (ptt or 0)+1
        --self.name = "ru"..ptt
    end
    if dirY<0 then
        self.sprite:switch("jumping")
        ptt = (ptt or 0)+1
        --self.name = "ju"..ptt
    elseif dir==0 then
        self.sprite:switch("standing")
        ptt = (ptt or 0)+1
        --self.name = "st"..ptt
    end
  
    if self.real then
        self:send("do",{"changeVelocityByKeys",dt,dir,dirY})
    end
  
    if dir<0 then
        vx = (vx - 1 * (vx > 0 and self.brakeAccel or self.runAccel)) * (dir*-1)
    elseif dir>0 then
        vx = (vx + 1 * (vx < 0 and self.brakeAccel or self.runAccel)) * dir
    else
        local brake = dt * (vx < 0 and self.brakeAccel or -self.brakeAccel)
        if math.abs(brake) > math.abs(vx) then--or self.shot then
            vx = 0
        else
            if (frr or 150)>math.abs(brake) then
                vx = vx + brake
            --friction
            if math.floor(math.abs(vx))>0 then
                vx = vx - self:getDir(vx)*(frr or 150)*dt
            end
            end
        end
    end

    if self.vx>self.maxVx then
        self.vx = self.maxVx
    end
    
    if dirY>0 and self.fly then
        self.gravity = self.gravity-500*dir
    end
    
    if dirY<0 and self.fly then
        self.gravity = self.gravity+500*dirY*-1
    end
    
    if self.isCrawling and math.abs(dir) <= 0.5 then
        dir = 0
        if dirY<0 then
            vy = -100
        elseif dirY>0 then
            vy = 100
        else
            vy = 0
        end
    elseif dir~= 0 then
    end
  
    if dir~=0 and self._new then
        self._new = nil
    end

 
    self.vx, self.vy = vx, vy
end

function Player:getAttacked(attacker)
    Entity.getAttacked(self, attacker)
    if self.life <= 0 and not self.dead then
        self:die(attacker)
    end
    self._isDamaged = .3
end

function Player:move(x,y,dt)
    if self.map.isClient then
   --     return self.x, self.y
    end
    
    local actualX, actualY, collisions, len = self.world:move(self,x,y,
    self.checkCollision)
    
    local currentBullet = self.currentBullet or {}
    local col
    local cross
    --local mini4
    
    --self.oldG = self.onGround
    local ignore = false
    self.onGround = false
    local onground = false
    
    for i = 1, len do
        col = collisions[i]
        cross = col.cross --ignore?
        if not cross then
            self:resolveCollision(col,true)
        end
        if col.other == self.tmpMini then
            self.mini4 = 0
        elseif self.tmpMini then
            self.mini4 = self.mini4 + 1
        end
        --assert(not col.other.isBullet,inspect(col.other,2))
        if col.other.isMini and col.other ~= self.tmpMini then

            local r
            if col.other.parent.isBullet then
                r = self:collectBullet(col.other)
                if not r then
                    assert(col.other~=self.tmpMini,r)
                    col.other:destroy()
                end
            elseif col.other.parent.isTile then
                r = self:collectBlock(col.other)
                if not r then
                    assert(col.other~=self.TtmpMini,r)
                    col.other:destroy()
                end
            end
            if not r then
                self.stats.minisCollected = self.stats.minisCollected+1
            end
            
            goto continue
        end
        
        local og = self:checkIfOnGround(col.normal.y)
        if not og then self.onGround=false end
        if not cross and col.other.isTile then
            if not self.canJump and self.oldnx ~= col.normal.x then
                --wall Jumped
                --self.canJump = true
            end
            self.oldnx = col.normal.x
        end
        
        if col.normal.y == -1 then
            game.gameStarted = true
            onground = true
            if self._new and self.real then
                self:fixCamera(true)
         --       self._new = nil
            end
        end
        
        
        if not cross and col.normal.y == -1 or col.normal.x ~=0 then
            if col.normal.x~=0 and self.trail then
                self:createPuff()
                
            end
            if not self.canJump  then
                
                if col.normal.x~=0 and not self.trail then
                    --puffs when land on wall
                   if not self.wallOnSide then
                        media.sfx.player_landed:play()
                        self:createPuff()
                    end
                elseif col.normal.x==0 then
                    self:createPuff()
                    if self.oldG then
                        --player kicked up dust
                    end
                end
                
            end
            if col.normal.y==-1 then--and col.other.isTile then
                self.canJump = true
                self.jumpCount = 0
                self.oldOnGround = true
                self.vy = 0
            end
            
            if col.normal.x ~= 0 then
                self.wallOnSide = col.normal.x
                if not self.canJump then self.canJump = true end
                
            end
            if col.normal.y == -1 and not self.canJump then --self.vy > 30 then
                ignore = true
                self.gravity = 0
                --causes puff on land
                self:createPuff(3,5)
                media.sfx.player_landed:play()
            end
            if self.shot then
                self._vx = self._vx or self.vx
                self.vx = self._vx/2
            else
                self._vx = nil
            end
            self.jumping = false
            if self.sprite.current == "jumping" then
                self.sprite:switch("standing")
            end
        end
        if not cross and col.normal.x==0 then
            self.wallOnSide = false
        end
        if not cross and col.normal.y == 1 then
            --if col.normal.x==0 then
                --self.canjump = false
            --end
            if not self.shot and currentBullet.isFlyBullet then
                --self.vy = self.initGravity/2
                --self.canJump = false
            end
        end
        if not cross and col.normal.y==0 then
            self.oldOnGround = false
        end

        ::continue::
    end


    if not onground then
        self.stats.timeInAir = self.stats.timeInAir+dt
    end
    
    self.oldOdnGround = self.onGround
    
    --not colliding with tmpMini block
    if self.tmpMini and self.mini4>5 then
        local n = self.getChild(self.displayBullets,4)
        n:setText("")
        n:setIcon(game:getSource("none.png"))
        self.tmpMini = nil
        self.bulletStorage[4] = nil
    end
    --not colliding with TtmpMini block
    if self.TtmpMini and self.Tmini4>5 then
        local n = self.getChild(self.displayBlocks,4)
        n:setText("")
        n:setIcon(game:getSource("none.png"))
        self.TtmpMini = nil
        self.blockStorage[4] = nil
    end
    
    return actualX, actualY
end

function Player:setJump(n)
    self.jumpAcceleration = n
end

function Player:jump()
    if self.real then
        if not self.canJump then
            return
        end
        self:send("do",{"jump"})
    end
    
    
    if 1 and (not sjump) then
        --current player jump is buggy--!!???
        return Bot.jump(self)
    end
    if nil then
        self.stats.jumpsPeformed = self.stats.jumpsPeformed+1
        self.sprite:switch("jumping")
        media.sfx.player_jumped:play()
        self.jumping = true
        self.vy = self.vy-self.jumpAcceleration/4--+jumpDec
        
        --create Puff when jumping off wall
        if self.wallOnSide then
            self:createPuff(5,7)
        end
        
        --stops player from flying prematurely
        if not self.canFly then
        self.jumpCount = self.jumpCount or 0
        if self.jumpCount>4 then
            self.canJump = false
            self.jumpCount = 0
        else
            self.jumpCount = self.jumpCount+1
            self.canJump = true
        end
            --self.canJump =  false
        end
    end
end

local myCMTrash = {}
function Player:update(dt,c)
    --self.sprite:update(dt)
    if self.real then
    if self.map.score and self.displayBullets.enabled then
        gooi.removeComponent(self.displayBullets)
        gooi.removeComponent(self.displayBlocks)
        gooi.removeComponent(self.map.pauseMenu)
        self.dB=4
    elseif not self.map.score and self.dB then
        gooi.addComponent(self.displayBullets)
        gooi.addComponent(self.displayBlocks)
        gooi.addComponent(self.map.pauseMenu)
        self.dB=nil
    end
    end
    
    if game.player == self then
        IM.active = true
    end
    
    self.updateCount = self.updateCount + 1
    
    if self.dead then
        return
    end
    
    self.tmpBurnt = false
    --self:playEffects()
    self:getTile()
    
    --detecting if player kicked up dust
    if self.canJump then
        self.oldG=true
    else
        self.oldG=false
    end
    
    if self.vy>0 then
        self.oldOnGround = false
    end
    
    --now puff is formed when player lands
    --due to jumping or fall
    if self:isFalling() then
        --for wall jumping don't use
        --self.canJump = false
    end
    
    self.sprite.x = self.x
    self.sprite.y = self.y
    
    self.old.x, self.old.y = self.x, self.y
    
    self.speed = {x=0,y=0}
    
    self:updateGravity(dt)
    
    if IM.active then
        for x = 1,3 do
            local slot = string.format("Slot%s",x)
            local func = self._bul
            if IM:isDown("Build") then
                func = self._blo
            end
            if IM:isDown(slot) then
                func(x)
                break
            end
        end
    end
    
    if self.adjusted then
        self.adjusted = self.adjusted-dt
        if self.adjusted<=0 then
            self.adjusted = nil
        end
    end
    
    if self.real then
        local vd,cc= self:getDisplacement()
        if self.vy<0 and not self.canFly then--(self.shot and self.currentBullet.isFlyBullet) then
     --       self.canJump = false
        end
        if self.canFly then
            self.canJump = true
        end
        if self.canFly then
            self.canFly = self.canFly-dt
            self.trail = 1
            if self.canFly<0 then
                self.canFly = nil
                self.trail = nil
            end
        end
        
        local left = IM:isDown("Left")
        local right = IM:isDown("Right")

        if (self.Ljoystick:getYDir()*-1 > 0 or IM:isDown("Jump")) and self.canJump then
            self:jump()
        end
        
        if self.Ljoystick:getXDir() < 0 or left then
            self.flipX = -1
        elseif self.Ljoystick:getXDir() > 0 or right then
            self.flipX = 1
        end
        
        self:changeVelocityByKeys(dt)
    
        local angle, dd = self.Rjoystick:getAngle()
        local distanceFromJoystick = dd or self.Rjoystick.d
        
        self.lastAngle = nil
        if IM.active then
            for x, i in pairs(Paddy.shootAngles) do
                if IM:isDown(x) then
                    angle = math.rad(i)
                    distanceFromJoystick = 1
                    self.lastAngle = angle
                    self.Rjoystick.angle = angle
                    break
                end
            end
        end
        
        
    
        if --??self~=game.player or 
        (self.currentBullet and distanceFromJoystick>.3 and not IM:isDown("Build")) then
            if angle ~= 0 then
                if math.deg(angle)>90 and math.deg(angle)<270 and not self.backShooting then
                    self.flipX = -1
                elseif not self.backShooting then--if math.deg(angle)<180 and math.deg(angle)>0 then
                    self.flipX = 1
                end
                self.shot = true
                self:shoot(angle, nil, self.Rjoystick:getX(),self.Rjoystick:getY())
            else
                self.shot = false
            end
        else
            if angle~=0 then
                if math.deg(angle)>90 and math.deg(angle)<270 and not self.backShooting then
                    self.flipX = -1
                elseif not self.backShooting then--if math.deg(angle)<180 and math.deg(angle)>0 then
                    self.flipX = 1
                end
            end
        end
        
        if self.toShoot then
            self:processShot(self.toShoot.x, self.toShoot.y)
            self.toShoot = false
        end
        
        if self.currentBlock then
            local build = IM:isDown("Build")
            local xDir, yDir
            if self.paddy:isDown("up") or (build and IM:isDown("ShootUp")) then
                xDir,yDir = 0,-1
            elseif self.paddy:isDown("down") or (build and IM:isDown("ShootDown")) then
                xDir,yDir = 0,1
            elseif self.paddy:isDown("left" )or (build and IM:isDown("ShootLeft")) then
                xDir,yDir = -1,0
            elseif self.paddy:isDown("right") or (build and IM:isDown("ShootRight")) then
                xDir,yDir = 1,0
            elseif self.paddy:isDown("upleft") or (build and IM:isDown("ShootTopLeft")) then
                xDir,yDir = -1,-1
            elseif self.paddy:isDown("upright") or (build and IM:isDown("ShootTopRight")) then
                xDir,yDir = 1,-1
            elseif self.paddy:isDown("downright") or (build and IM:isDown("ShootBottomRight")) then
                xDir,yDir = 1,1
            elseif self.paddy:isDown("downleft") or (build and IM:isDown("ShootBottomLeft")) then
                xDir,yDir = -1,1
            end
            
            if xDir~=nil then
                self:placeBlock(xDir,yDir)
            end
        end
        
        gooi.update(dt)
    else
        local angle = self.Rjoystick.angle --haha
    end
    
    self:checkLimits()--???!!!!!fix
    local x,y = 0,0
    
    if angle~=0 then
       -- self.vx = self.vx-(self:getDir(self.vx)*150)
        if self.sprite.current~="walking" then
            self._previousState = self.sprite.current--????
           -- self.sprite:switch("walking")
        end
    else
   --     self.sprite:switch(self._previousState or "running")
    end
    
    local vx=self.vx
    local vx, vy = self.x+vx*dt,self.y+self.vy*dt
    
 
    --fo my own movements--????
    if not self.passive then--not self.map.Client then
        local mx, my = self:move(vx,vy,dt)
        self.x, self.y = mx,my
    end
    
    self:fixCamera()
    
    self.lifeBar.x = self:getCenter()
    self.lifeBar.x = self.lifeBar.x-self.lifeBar.w/2--self.x-self.lifeBar.w/4
    self.lifeBar.y = self.y-self.lifeBar.h-self.h/5
    self.hat:update(dt)
    self.hat2:update(dt)
    
    if self._isDamaged then
        self._isDamaged = self._isDamaged-dt
        if self._isDamaged<=0 then
            self._isDamaged = nil
        end
    end
    
    if (self.LAST_SHOT) then
        self.LAST_SHOT = self.LAST_SHOT - dt
    end
    IM.active = false
end

function Player:processShot(x,y)

    if (not IM:isDown("Build")) then
        local angle = -math.atan2(
        y-self.y,
        x-self.x)
        self.shot = true
        self:shoot(angle)
        self.Rjoystick.angle = angle
        
        if angle~=0 then
            if (math.deg(angle)<-90 or math.deg(angle)>(90)) and math.deg(angle)<(ntt or 360) and not self.backShooting then
                self.flipX = -1
            elseif not self.backShooting then
                self.flipX = 1
            end
        end
    else
        local dirx = self:dir(x-self.x)
        local diry = self:dir(y-self.y)
        self:placeBlock(dirx, diry)
    end
end

function Player:isDamaged()
    return self._isDamaged
end

function Player:setRstickAngle(ang)
    self.Rjoystick.angle = ang
end

function Player:isFalling()
    return self.vy>0 and not self.onGround
end

function Player:draw(x,y,w,h)
    if self.dead then
        return
    end
    
    self:drawSprite(x,y,w,h)

    if 1 then
        self.lifeBar:draw()
    
    
        local fw = love.graphics.getFont():getWidth(self.name)
        love.graphics.print(self.name, self:getCenter()-fw/4, self.lifeBar.y-self.lifeBar.h*2,0,.5,.5)
    end

    if self.real then
        self.map.camera:detach()
        if not DESKTOP then
            self.paddy:draw()
        end
        gooi.draw(self.uuid)
        self.map.camera:attach()
    end
end

function Player:drawSprite(x,y,w,h)
    local _x,_y,__w,__h = x,y,w,h
    
    local x,y,w,h = x or self.x-1,y or self.y,w or 1, h or 1
    w,h = self.w*w, self.h*h
    
    if __w then
        self._oldw,self._oldh = self._w, self._h
    end
    
    self._w, self._h = resizeImage(self.spriteImage,w*4,h)--game

    local pcolor = getColor(self.color)
    pcolor[4] = self.alpha or 1
    
    local r,g,b, a = love.graphics.getColor()
    love.graphics.setColor(pcolor)
    
    
    local v = self.spriteImage
    local xx,yy = v:getWidth()/2, v:getHeight()/2
    
    if self.flipX<0 then
        xx = xx+w
    end

    local tx, ty = x,y
    if tnn then--self.passive then
        tx,ty = x+xx-w*5,y+yy-h*2
    else
        if self.flipX==-1 then
            tx = tx+w
        end
    end
    self.sprite:draw(tx,ty,self.angle,self._w*self.flipX,self._h)
    
    local color = getColor(self.stomachColor)
    color[4] = self.alpha or 1
    love.graphics.setColor(color)
    love.graphics.draw(self.stomachImage,tx,ty,self.angle,self._w*self.flipX,self._h)
    
    local color = getColor(self.gunColor)
    color[4] = self.alpha or 1
    local _w, _h = resizeImage(self.handImage, w, h)
    local ang = (self.Rjoystick.angle)*-1
    local v = self.handImage--game:getSource(self.source)
    local xx,yy = 0,0--v:getWidth()/2, v:getHeight()/2(ang+math.pi)
    
    if self.flipX<0  then
        xx = xx+w-(self.w/4)*2 xx=-w/2
        if ang~=0 and not self.passive then
            ang = math.rad(180-math.deg(ang))*-1
        end
    elseif self.flipX>0 then
  
    end
    
    self.hat:draw(_x,_y,__w,__h)
    self.hat2:draw(_x,_y,__w,__h)
  
    love.graphics.setColor(color)
    love.graphics.draw(self.gunImage, x+(w/1.5)+xx,y+h/2+yy,
        ang,
        _w*self.flipX,_h,v:getWidth()/2,v:getHeight()/2)
        
    love.graphics.setColor(pcolor)
    love.graphics.draw(self.handImage, x+(w/1.5)+xx,y+h/2+yy,
        ang,
        _w*self.flipX,_h,v:getWidth()/2,v:getHeight()/2)
    love.graphics.setColor(r,g,b,a)
    
    if self.drawSpark then
        Spark:new({
        x = x+(w/1.5)+xx,--+(v:getWidth()/()*_w*self.flipX,
        y = y+h/2+yy,--(v:getHeight()/2)*_h,
        map = self.map,
        angle = ang,
        ox = v:getWidth()/2,
        oy = v:getHeight()/2,
        ssource = "players/gunSpark.png",
        _h=_h,
        _w=_w*self.flipX
        })
        self.drawSpark = false
    end
    t_h,t_w = _h,_w
    
    if self.drawTears then
        self.tearSprite:draw(tx,ty,self.angle,self._w*self.flipX,self._h)
    elseif self.drawDroplets then
        love.graphics.draw(self.dropletsImage,tx,ty,self.angle,self._w*self.flipX,self._h)
    end
    if self.line then love.graphics.line(self.x,self.y, self.line.x,self.line.y) end
    --love.graphics.rectangle("line",self.x, self.y, self.h-2, self.w-2)
    if __w then
        self._w, self._h = self._oldw, self._oldh
        self._oldw, self._oldh = nil, nil
    end
end

function Player:createPuff(min, max)
    local min = min or math.random(1,3)
    self.madePuff = 3
    if max then
        min = math.random(min, max)
    end
    local min = 2
    for i = 1,min do
        if math.random(0,1)==1 then
        Puff:new({map=self.map,
            x=self.x + self.w,
            y=self.y + self.h / 2,
            minSize=16 * (1 - math.random()),
            maxSize=39,
            vx=2, vy=3}
        )
        self.stats.dustProduced = self.stats.dustProduced+1
        end
    end
end


function Player:spawnDebris(color)
    if not self.map.active then
        return
    end
    --media.sfx.explosion:play()
    media.sfx.entity_destroyed:play()
    color = color or self.color or "brown"
    
    local area = self.w * self.h
    local debrisNumber = 50--math.floor(math.max(100, area / 100))
    

    for i=1, debrisNumber do
        local color = color
        if type(color) == "table" and type(color[1])~=number then
            color = lume.randomchoice(color)
        end
        
        Debris:new({
            map = self.map,
            x = math.random(self.x, self.x + self.w),
            y = math.random(self.y, self.y + self.h),
            minVel = -(mv or 400),
            maxVel = mv or 400,
            color = color
        }
      )
    end
end

function Player:destroy()
    Entity.destroy(self)
   -- lume.remove(game.players, self)
    self.dead = true
end

function Player:playEffects()
  --create puffs when jumping
  if self:isFalling() and self.trail then--self.onGround and self.vy>0 then
      --add puff here for trail
      self:createPuff()
  end
  if self.vy<0 and self.canJump and not self.wallOnSide then
    if not self.onGround and self.trail then 
       self:createPuff()
    --  media.sfx.player_jump:play()
    else
      Puff:new({map=self.map,
               x=self.x,
               y=self.y + self.h / 2,
               minSize=16 * (1 - math.random()),
               maxSize=39,
               vx=2, vy=3}
               )
      Puff:new({map=self.map,
               x=self.x + self.w,
               y=self.y + self.h / 2,
               minSize=16 * (1 - math.random()),
               maxSize=39,
               vx=2, vy=3}
               )
      if nil then --media.sfx.player_propulsion:countPlayingInstances() == 0 then
       -- media.sfx.player_propulsion:play()
      end
    end
  else
  end

 
end

function Player:fixCamera(stop,stt)
    if self.real then
        if not stt then self:getTile() end
        if not self.tile then
            return
        end

        self.followX = true
        self.followY = true
        
        local up, right
        local cM = self.map.cameraMan or myCMTrash
        local cMx, cMy
        
        local tx,ty = self.tile.x, self.tile.y
        local sc = (self.map.camera.scale/2.3)*(_W/W())
        if self.tile.x>self.map.maxX-(16.5/sc) then
            self.followX = false
            right = 1
        elseif tx<self.map.minX+(17.5/sc) then
             self.followX = false
        end
        if ty>self.map.maxY-(maty or 10/sc) then
            self.followY = false
            up = 1
        elseif ty<self.map.minY+(mity or 11/sc) then
            self.followY = false
        end
        
        if not stop then return end
        
        if not self.followX and right then
            cMx = self.map:getTile(floor(self.map.maxX-(16.5/sc)),ty)
        elseif not followX then
            cMx = self.map:getTile(floor(self.map.minX+(17.5/sc)),ty)
        end
        
        if not self.followY and up then
            cMy = self.map:getTile(tx,floor(self.map.maxY-(10/sc)))
        elseif not self.followY then
            cMy = self.map:getTile(tx,floor(self.map.minY+(11/sc)))
        end
        cM.y = cMy and cMy._y or cM.y
        cM.x = cMx and cMx._x or cM.x
        
        if self.currentBlock then
            self.tile.color = getColor("green")
        else
             self.tile.color = nil
        end
    end
end
function Player:released(...)
    return self.paddy:touchreleased(...)
end

function Player:pressed(...)
    return self.paddy:touchpressed(...)
end

function Player:touchmoved(...)
    return self.paddy:touchmoved(...)
end

--These can be used by close function in attacks
--after their not current
function Player:trashDraw()
    for x,i in lume.iter(self._drawTrash) do
        i()
    end
end

function Player:trashBeforeDraw()
    for x,i in lume.iter(self._beforeDrawTrash) do
        i()
    end
end

function Player:trashAfterDraw()
    for x,i in lume.iter(self._afterDrawTrash) do
        i()
    end
end

function Player:drawTrash(func)
    table.insert(self._drawTrash,func)
end

function Player:afterDrawTrash(func)
    table.insert(self._afterDrawTrash,func)
end

function Player:beforeDrawTrash()
    table.insert(self._beforeDrawTrash,func)
end

function Player:updateTrash(func)
    table.insert(self._updateTrash,func)
end

function Player:removeTrash(func)
    return lume.remove(self._updateTrash,func) or
    lume.remove(self._drawTrash,func) or
    lume.remove(self._afterDrawTrash,func) or
    lume.remove(self._beforeDrawTrash,func)
end


return Player