--[[
-- Puff Class modified from bump.lua's example
-- Represents a Puff of smoke, created either by explosions or by the Player's propulsion
-- Puffs don't interact with anything (they can be displaced by explosions)
-- They gradually change in shape & color until they disappear (lived & lifeTime are used for that)
-- Since puffs continuously change in size, we keep adding and removing them from the world (this is ok,
-- it's the same thing that bump does internally to move things around)
--]]

local lume = lume
local boomPng = "boom.png"
--local log = null
local Puff = class:extend("Puff",5100)

--function that returns current updates loops in game
local getUpdateCount = getUpdateCount

local rad = math.rad
local game = game


local defaultVx      = 0
local defaultVy      = -10
local defaultVa      = 10
local defaultMinSize = 2
local defaultMaxSize = 10
local dustes = {"dust2.png"}--,"dust2.png"}--,"dust2"}

function Puff:__init__(kwargs)
  self.active = true
    
  kwargs = kwargs or {}
  self._kwargs = kwargs
    
  self.uuid = kwargs.uuid or lume.uuid()
  self.updateCount = getUpdateCount()

  self.map = kwargs.map
  self.x = kwargs.x
  self.y = kwargs.y
  
  self.passive = kwargs.passive
  if not self.passive then
    --hitchhiking
    self.map:storeBullet(self)

  end
  
  local vx, vy = kwargs.vix or defaultVx, kwargs.viy or defaultVy
  self.va = kwargs.va or math.random(-defaultVa,defaultVa)
  minSize = kwargs.minSize or defaultMinSize
  maxSize = kwargs.maxSize or defaultMaxSize

  self.w = math.random(minSize, maxSize)
  self.h = self.w--math.random(minSize, maxSize)
  self.angle = 0
  
  self.parent = kwargs.parent or nil
  self.source = kwargs.source or lume.randomchoice(dustes)
  
  --set to 0 if smoke bullet and user
  self.opacity = kwargs.opacity or 100/255
  --set to 0.95 if smoke bullet and user
  --????use 0.08 for smoke
  self.opacity2 = kwargs.opacity2 or kwargs.outline or 1
  if self.source and self.source~="explosion.png" then self.opacity2=100/255 end
  
  --so that the player who caused the puff from a Smokescreen Bullet
  --can still see
  if self.parent == game.map.player then
    self.opacity2 = .95
    self.opacity = 0
    if self.source then
        self.opacity2 = .02
    end
  end
  
 
  if self.opacity>1 then
      self.opacity = self.opacity/255
  end
  
  if self.opacity2>1 then
      self.opacity2 = self.opacity2/255
  end
  
  if kwargs.favor then
      self.opacity = 0
      self.opacity2 = .95
  end
  
  local l = kwargs.lifeTime or kwargs.life or 0.1
  self.lifeTime = l + math.random()
  self.lived = 0
  self.vx, self.vy = vx, vy
  self.dead = false

  self.isBackground = true
  if self.source==boomPng then
      self.opacity  = 1
      self.opacity2 = 1
  end
  if self.source then
      self._w, self._h = resizeImage(game:getSource(self.source),self.w,self.h)
  end
end

function Puff:getCenter()
    return self.x + self.w / 2,
         self.y + self.h / 2
end

function Puff:expand(dt)
  local cx,cy = self:getCenter()
  local percent = self.lived / self.lifeTime
  if percent < 0.2 then
    self.w = self.w + (200 + percent) * dt
    self.h = self.h + (200 + percent) * dt
  else
    self.w = self.w + (20 + percent) * dt
  end
  --vv=vv or 2
  if self.w>2 then
      if self.source == "dust2.png" then
          self.source = "dust.png"
      end
  end
  

  self.x = cx - self.w / 2
  self.y = cy - self.h / 2
end

function Puff:destroy()
    assert(not self.destroyed)
    self.map:kill(self)
    self.dead = true
    self.destroyed = true
end

function Puff.checkCollision(item, other)
    if other then
        return nil
    end
end


function Puff:update(dt,c)
  if self.source then
      self._w, self._h = resizeImage(game:getSource(self.source),self.w,self.h)
  end
  
  self.lived = self.lived + dt

  if self.lived >= self.lifeTime  then
    assert(not self.dead)
    self.dead = true
  else
    self:expand(dt)
    local next_l = self.x + self.vx * dt
    local next_t = self.y + self.vy * dt
    --self.angle = self.angle+self.va

    self.x, self.y = next_l, next_t
  end
  
  if self.dead then
      --problem comes from destroy because after awhile
      --when the game closes last thing seen is dest
      self:destroy()
  end
end

function Puff:getColor()
  local percent = math.min(1, (self.lived / self.lifeTime) * 1.8)
  --return colors.brown[1],colors.brown[2],colors.brown[3]
  return  (math.floor(155*percent))/255,
          (255 - math.floor(155*percent))/255,
          100/255
end

function Puff:draw()
  if not self.source then
        local rr,gg,bb,a= love.graphics.getColor()
        local r,g,b = self:getColor()
        util.drawFilledRectangle(self.x, self.y, self.w, self.h, r,g,b,self.opacity,self.opacity2)
        love.graphics.setColor(rr,gg,bb,a)
        if ty then love.graphics.setColor(colors.brown[1],colors.brown[2],colors.brown[3],25) end
  else
        local rr,gg,bb,a= love.graphics.getColor()
        love.graphics.setColor(255,255,255,self.opacity2)
        local so = game:getSource(self.source)
        local vw,vh = so:getDimensions()
        local ang = rad(self.angle)
        if self.angle~= 0 then
            love.graphics.draw(game:getSource(self.source),self.x,self.y,ang,self._w,self._h,
            vw/2,vh/2)
        else
            love.graphics.draw(game:getSource(self.source),self.x,self.y,0,self._w,self._h)
        end
        love.graphics.setColor(rr,gg,bb,a)
  end
  
end

return Puff



