Turret = Entity:extend("Turret")

local huge, abs = math.huge, math.abs
local getUpdateCount = getUpdateCount


function Turret:__init__(kwargs)
    self.name = kwargs.name
    self.map = kwargs.map or {world={add=null}}
    self.world = kwargs.world or self.map.world or {}
    kwargs.map = self.map
    kwargs.world = self.world
    
    Entity.__init__(self,kwargs)
    self.source = "turretBox.png"
    self.muzzle = "muzzle.png"
    
    --assert(self.angle>0)
    self.angleY = math.sin(self.angle)
    self.angleX = math.cos(self.angle)

    self.delay = kwargs.delay or 0.1

    
    self.ac = kwargs.speed or kwargs.ac or 500
   -- self.initGravity = kwargs.gravity or self.initGravity
    
    self.hits = kwargs.hits or 1
    --how many seconds bullet lasts
    self.life = kwargs.life or 100
    self.attack = kwargs.attack or 1.5
    self.activated = true
    
    self.range = kwargs.range or kwargs.sight or 10000
    self.brutal = kwargs.brutal or false
    
    self.main = kwargs.main or kwargs.isMain or kwargs.class or kwargs.isClass
    if self.main then
        for _,i in pairs(kwargs) do
            self[_] = self[_] or lume.copy(i)
        end
        CLASSES[self.name] = self
    end
    --self.vx = self.ac*self.angleX
    --self.vy = self.ac*-1*self.angleY
    
    --self.angleY, self.angleX=1,1
    self.ignoresParent = true
    self._bullet = kwargs.bullet or"Bullet"
    assert(type(self._bullet)=="string")
    self.bullet = bullets[self._bullet]
    self.bullets = {}
    self.bulletLen = 0
    self.parent = kwargs.parent or {}
    self.parent = {team=self.parent.team}
    self.map.turrets[#self.map.turrets+1] = self
    self.index = #self.map.turrets
    if self.map.host then
        self.map.host:sendToAll("spawnTurret",{
            self.x, self.y, self._bullet,#self.map.turrets,self.parent.team
        })
    end
    self.map:storeBullet(self)
    
end

function Turret:reloadAngles(ang)
    local angle = ang or self.angle
    self.angleY = math.sin(angle)
    self.angleX = math.cos(angle)
end

function Turret.checkCollision(item, other)
    if not other.isBackground and not other.isMini then
        return item.collisionType
    end
end

function Turret:changeVelocityByCollisionNormal(nx, ny, bounciness)
  local bounciness = bounciness or self.elasticity
  if not bounciness or bounciness == 0 then
      return
  end
  local vx, vy = self.vx, self.vy

  if (nx < 0 and vx > 0) or (nx > 0 and vx < 0) then
    vx = -vx * bounciness
    self.bounced.x = true
  end
  
  if ((ny < 0 and vy > 0) or (ny > 0 and vy < 0)) and abs(vy)>10 then
    vy = -vy * bounciness
    self.bounced.y = true
  end

  self.vx, self.vy = vx, vy
end

function Turret:move(x,y)
    local actualX, actualY, collisions, len = self.world:move(self,x,y,
    self.checkCollision)
    local col
    for i = 1, len do
        col = collisions[i]
        self:resolveCollision(col)
    end
    
    return actualX, actualY
end

function Turret:destroy(...)
    Entity.destroy(self,...)
    lume.remove(self.map.turrets,self)
end

local dirs = {{-1,0},{1,0}}
function Turret:update(dt, c)
    c = c or getUpdateCount()
    
    if c-1<self.updateCount then --set to <= if shadow
    --    return
    elseif c-1>self.updateCount then
        self.updateCount = c-1
    end
    self.updateCount = self.updateCount + 1

    self:getTile()
    if not self.tile then self:destroy() end
    --[[local t = self.tile
    local updateG, tU = true
    for x = 1, #dirs do
        tU = dirs[x]
        local tt = self.map:getTile(tU[1]+t.x,tU[2]+t.y)
        if not tt or tt.isBackground then
        --    updateG = false
        end
    end
    
    if updateG then
        Entity.update(self, dt)
        --self:updateGravity(dt)
        --self.x, self.y = self:move()
    end
    ]]
    local target = self:getTarget()
    if target and self.activated then
        self:shootAt(target)
        self.angle = self:getAngle(target)
        self:reloadAngles()
    end
    
    if (self.LAST_SHOT) then
        self.LAST_SHOT = self.LAST_SHOT-dt
    end
    if self.life<=0 then
        self:destroy()
    end
end

function Turret:getTarget()
    return self:getNearest(game.players,self.map.turrets)
end

local function filter(item)
    if item.isBackground or item.isBullet or item.isMini then
        return false
    end
    return true
end

function Turret:getNearest(objects,obj2)
    local distance, nearest, object, tx, ty
    local cx, cy = self.x, self.y--self:getCenter()
    local shortestDistance = self.range---huge
    for i = 1, #objects do
        object = objects[i]
        if object.team~=self.parent.team and object.life>0 and not object.dead then
            if self.map.player==object then
                --log("firing player at team "..object.team.."\n    My team is "..tostring(self.parent.team))
            end
            
            distance = self:getSquaredDistance(object)
        
            if distance < shortestDistance then
                tx, ty = object:getCenter()
                local itemInfo, len = self.world:querySegment(cx,cy,tx,ty,filter)
                -- ignore itemsInfo[1] because that's always self?
                local info = itemInfo[1]
                if info==self then info = itemInfo[2] end
                if info then assert(not info.isBackground) end
                if (info and info==object) or self.brutal then
                    nearest = object
                    shortestDistance = distance
                    --warn("Target")
                end
            end
        end
    end
    if obj2 and #obj2>1 then
        for i = 1,#obj2 do
            object = obj2[i]
            if object.life>0 and object~=self and object.parent.team~=self.parent.team then
                distance = self:getSquaredDistance(object)
        
                if distance < shortestDistance then
                    tx, ty = object:getCenter()
                    local itemInfo, len = self.world:querySegment(cx,cy,tx,ty,filter)
                    -- ignore itemsInfo[1] because that's always self?
                    local info = itemInfo[1]
                    if info==self then info = itemInfo[2] end
                    if info then assert(not info.isBackground) end
                    if (info and info==object) or self.brutal then
                        nearest = object
                        shortestDistance = distance
                        --warn("Target")
                    end
                end
            end
        end
    end
                
    return nearest
end

function Turret:getSquaredDistance(object)
    local dx, dy = self.x - object.x, self.y - object.y
    return dx*dx + dy*dy
end

function Turret:shootAt(target)
    local b = self.bullet:new({
        x = self.x,
        y = self.y,
        angle = self:getAngle(target),
        parent = self,
        map = self.map,
        target = target
    }) --warn()
    b.delay = 0.2
    table.insert(self.bullets,b)
    self.bulletLen = #self.bullets
end

function Turret:getAngle(object)
    return -math.atan2(object.y-self.y, object.x-self.x)
end

function Turret:draw()


    
    local v = game:getSource(self.source)
    local m = game:getSource(self.muzzle)
    local mw, mh = resizeImage(m, self.w, self.h)
    local vw, vh = resizeImage(v, self.w, self.h)
    
    local xx,yy =v:getWidth()/2, v:getHeight()/2
    local mx,my =m:getWidth(), m:getHeight()
    local ca = math.pi^2
    local ang = math.deg(self.angle)
    
    love.graphics.draw(v,self.x,self.y,0,vw,vh)--,xx,yy)
    love.graphics.draw(m,self.x+self.w/2,self.y+self.h/2,-self.angle,
    mw,mh,0,my/2)
    self:drawLife()
end

return Turret

