local Underground = GridMap:extend("Underground")
local Tike = Tile
local RECORDS = RECORDS
local choice = lume.randomchoice


local _Path = class:extend("_Path")
function _Path:__init__(x,y)
    self.x = x
    self.y = y
end

function _Path:getX()
    return self.x
end

function _Path:getY()
    return self.y
end

function _Path:getPos()
    return self.x, self.y
end

local sizes = {
    small = 50,
    medium = 200,
    large = 400,
}

Underground.slowRange = 40
Underground.slowChance = .8

function Underground:destroy(destroyPeer,noSwitch)
    
    game.map = nil
    game.currentMap = nil noil=4
    game.player = nil
    self.isOld = true
    self.active = false
    self:unpause(true)
    
    for x,i in ipairs(game.players) do
        i:crush()
    end
    
    --destroy peer?
    if destroyPeer and self:hasPeer() then
        self:hasPeer():destroy()
    end
    
    game.players = {}
    self.player = {}
    
    gooi.removeComponent(self.pauseMenu)
    gooi.removeComponent(self.message)

    done = {}

    Gs.__stack = {}
    GS.old = nil
    
    
    if not noSwitch then
        GSS(mainMenu)
        mainMenu:create()
    end
    
    package.loaded["tools/hump/gamestate"].old = nil --??
    
    
end

function Underground:log(l)
    log(l)
end

function Underground:dopause()
    if self.score then
        self.score:destroy()
    end
    
    self.score = ScoreDisplay:new({
        map = self,
        data = {3,4},
        h = H()-(H()/4),
        w = W()-(W()/4)
    })
    self.score:create()
end

function Underground:unpause(ignore)
    
    if self.player and self.player.dead and not ignore then
        return
    end
    
    if self.score then
        self.score:destroy()
        self.score = nil
    end
end

function Underground:setMatchTime(time)
    if not time then
        return
    end
    self.matchTime = self.timer:after(time or self.time, function()
        if not self.isClient and not self.training then
            if self.host then
                self.host:sendToAll("do",{"sendAwayStats"})
                self.tryToEnd = true
            else
                self.canEndGame = true
            end
            self:endGame()
        end
    end)
end

function Underground:getScores()
    
        --returns players in order of their scores
        
        local pl,ts
        local score
        local scores = {}
        local tscores = {}
        local randomchoice = lume.randomchoice
        
        for x, i in ipairs(game.players) do
            score = i:getScore()
            table.insert(scores, score)
            tscores[score] = tscores[score] or {}
            table.insert(tscores[score],i)
        end
        
        table.sort(scores)
        
        local players = {}
        for x = 1,#scores do
            ts = tscores[scores[x]]
            
            --so tied players randomly come higher
            pl = randomchoice(ts)
            lume.remove(ts,pl)
            
            table.insert(players,pl)
        end
        return players

end

function Underground:endGame()
    --wait to get achievements
    if self.host and not self.canEndGame then
        return
    end
    
    if self.score then
        self.score:destroy()
    end
    
    local function openAchievements()
        if not self.am then
            self.am = AchievementsMenu:new({
                map=self,
                h = H()-(H()/4),
                w = W()-(W()/4)
             })
            self.am:load()
        end
        self.am:create()
        self.score = self.am
    end
    
    if self.match == "deathmatch" then
        local player = self:getScores()[1]
        gooi.alert({
            text=string.format("%s is the winner!",player.name),
            ok=openAchievements
        })
        --endMenu
    
    elseif self.match == "team deathmatch" then
        
    elseif self.match == "king pin" then
    
    elseif self.match == "capture the flag" then
    
    end
    
    if self.host then
        self.host:sendToAll("do",{"endGame",nil,nil,nil,nil,"map"})
        openAchievements()
        self.am:destroy()
    else
        openAchievements()
    end
end

function Underground:start()
    assert(not self.isOld)
    if self.size then
        self.w = sizes[self.size]
        self.h = self.w
    end
    game.progress.total = self.w*self.h+(((self.w+self.h)*10)+100)
    mainMenu:progress()
        
    self._maxCount = {x=self.w/self.chunk.width,y=self.h/self.chunk.height}
    
    --[[if self.host then
        for x=1,1000 do--waits a 1000 updates or less to start #game.players+1 do
            if game.activate<#game.players then
                game.copaused = 1--game.copaused and game.copaused+1 or 0
                coroutine.yield(game.co)
            else
                break
            end
        end
    end]]
    
    if self.loadAll or self.w*self.h<=2000*2000 then
        self:loadAllChunks()
        self:load()
        self.loadedAll = true
        self.loadAll = true --error(inspect(self.allCurrentChunks))
    else
        self:loadChunks()
        self.loadedAll = fasle
    end
    
    if self.client then
        self.client:send("ready")    
        if self.client then
            self.pause = true
        end
        self.client:update(0)
    end

    self:updateChunks(0)
    self.countG = 0
    

    self.entities[game.player.uuid] = game.player
    
    local maxX = self.maxX-self.minX
    local maxY = self.maxY-self.minY
    for x=1,math.random(1,1) do
        self:spawnDigger(math.random(maxX-10,maxX),math.random(maxY-10,maxY),xx1)
    end
    if not self.isClient then
        self:spawnPlayers()
    end
    
    self.player = game.player or error("No player!")
    
    local pw = W(10)
    local B_W = W()/5.5
    local ww  = B_W/1
    local h   = H()/7
    local three = 2
    local two = 1.5
    local four = 3
    local five = 4
    
    self.cameraButton = gooi.newButton({
        group=self.player.uuid,
        text="",
        icon=game:getSource("camera.png")
    }):onRelease(function()
        local sc = self.cameraMan.scale
        self.cameraMan.scale = (sc==three and four) or (sc==four and five)  or (sc==five and two)
            or 2
        self.player:fixCamera(true)
        end)
        
    self.pauseButton = gooi.newButton({
        group=self.player.uuid,
        text="",
        icon=game:getSource("pause.png"),
        --x=W()/2-pw/2,
        --h=H(50),
        --w=pw
    }):onRelease(function()
        self.player.stats.numberOfTimesPaused = self.player.stats.numberOfTimesPaused+1
        self:dopause()
    end)
    self.pauseMenu = gooi.newPanel({
        group=self.player.uuid,
        layout="grid 1x2",
        x=W()/2-ww/2,
        w=ww,
        h=h
    }):add(
        self.pauseButton,
        self.cameraButton
    ):setStyle({bgColor=getColor(self.player.color,.5)})

    self.message = gooi.newLabel({x=W()*0,y=H()/7,w=W(),group=self.player.uuid}):setText("")

    gooi.setStyleComp(self.message)
    self.message:setStyle({
        bgColor=getColor(self.player.color),
        fgColor=getColor(self.player.color)})
    self.message.w = W()
    self.message:center()

    
    if mainMenu then--.drawProgress then
        mainMenu.drawProgress = false
        if self.host then
            self.host:sendToAll("activate")
        end
        GSS(self)
            
        for x = 1,MAX_PLAYERS do
            mainMenu.waitButtons[x]:setText("...")
        end
        
        log("Starting switch")
        log(tostring(self.client))
    
    end
end

--[[
    y
x       b
    a
]]
function Underground:__init__(kwargs)
    local kwargs = kwargs or {}
    GridMap.__init__(self,kwargs)
    observeGlobals()
    self.seed = kwargs.seied or math.random()
    rot.setSeed(self.seed)
    self.timer = Chrono()
    math.randomseed(self.seed)
    self._tw, self._th = resizeImage(32,32,16,16)
    self._cw, self._ch = resizeImage(512,512,16,16)
    
    local gs = function(s) return game:getSource(s) end
    self.tileSources = {
        ["iron.png"] = gs("iron.png"),
        ["background_dirt.png"] = gs("background_dirt.png"),
        ["dirt.png"] = gs("dirt.png"),
        ["stone.png"] = gs("stone.png"),
        ["whiteDirt.png"] = gs("whiteDirt.png")
    }
    self.tileColors = {
        ["iron.png"] = "silver",
        ["background_dirt.png"] = "dirtbrown",
        ["dirt.png"] = "woodbrown",
        ["stone.png"] = "darkgray",
        ["whiteDirt.png"] = "white"
    }
    
    self.cameraSpeed = 0.02
    self.usePaddy = false
    self.world = bump.newWorld(16)
    self.chunkSize = {
        w=3,
        h=3
    }
    self.chunk = {
        width = 10,
        height = 10,
    }
    self._maxCount = {x=self.w/self.chunk.width,y=self.h/self.chunk.height}
    self.chunkCount = {x=0, y=0}
    
    --self.tile = {w=16,h=16}
    self.allChunks = {}
    self.offset = {x=1,y=1,_y=16,_x=16}
    
    --tiles generated at a time
    self.genRange = kwargs.genRange or 5
    if self.genRange%2 == 0 then
        self.genRange = self.genRange + 1
    end
    
    if type(self.genRange) == "number" then
        self.genRange = {x=self.genRange,y=self.genRange}
    elseif not self.genRange.x then
        self.genRange = {x=self.genRange[1],y=self.genRange[2]}
    end
    
    self.allCurrentChunks = {
        x = {0},
        y = {0}
    }
    self.records = {[1]={[1]=true}}
    
    for t = 1,(self.genRange.x-1)/2 do
        table.insert(self.allCurrentChunks.x,-t)
        table.insert(self.allCurrentChunks.x,t)
    end
    for t = 1,(self.genRange.y-1)/2 do
        table.insert(self.allCurrentChunks.y,-t)
        table.insert(self.allCurrentChunks.y,t)
    end
    
    table.sort(self.allCurrentChunks.x)
    table.sort(self.allCurrentChunks.y)
    self.currentChunks = {}
    self.yChunks = {}
    self.debris = {}
    self.debrisCount = 0
    self.active = true
    
    --self.world:addCollisionClass("Dirt", {ignores={"Dirt"}})
    --self.world:addCollisionClass("Player")
    self.camera.scale=.5
  --  self:loadMap(self.w, self.h)
    self.loadAll = kwargs.loadAll
    
    self.messageLast = 2
    self.tmpMessageLast = self.messageLast
    
    local mw, mh = W()/4,((H()/10)+50+10)
    local font = love.graphics.getFont()
    fh = font:getHeight()
    self.messages = {}
    self.messageLen = {}
    for x = 1, math.floor((H()-mh)/(fh*2)) do
        table.insert(self.messages,{"",0,mh+(x*fh)})
    end
    self.messageLen.h = #self.messages
 --   self.messageLen.w = mw/font:getWidth()
    self._mh = mh
    self._mw = mw
    self.font = font
    self.messages = {}
    
    --for t=1,40 do self:sendMessage("hey"..t) end
  
    
    self.timeSinceLastCrate = 0
    self.createCrateTime = 1
    self.tickRate = 0
    self.updateRate = 1/60
    
    self.diggers = {}
    self.entities = {}
    self.bullets = {}
    self.turrets = {}
    self.crates = {}
    self.toSet = {"x","y","life","vx","vy","source","w","h","angle"}
    
    self.camera.scale=2
    self.cameraMan = {x=0,y=0,scale=2}
    self.vintage = true--cuts some costs and makes game run faster
    self.gore = kwargs.gore or game.settings.gore
    
    self.moving = false
    self.cancelMove = function()
        self.moving = false
        self.camera:setFollowStyle(self.cameraMan.oldStyle)
        self:moveCamera(self.cameraMan.oldXPos, self.cameraMan.oldYPos, self.cameraMan.oldScale, self.cameraMan.oldTime/2, true)
    end
    
    self.cancelMoveLittle = function()
        self.moving = false
        self.camera:setFollowStyle(self.cameraMan.oldStyle)
    end
    
    self._tag = nil
end

function Underground:tag(object)
    self._tag = object
end

function Underground:setSeed(seed)
    self.seed = seed or math.random()
    rot.setSeed(self.seed)
end

function Underground:getItem(uuid)
    return self.entities[uuid]
end

function Underground:updateEntity(data)
    local obj,val = self.entities[data.uuid]
    for x = 1,#self.toSet do
        val = self.toSet[x]
        obj[val] = data[val]
    end
    return obj
end

function Underground:getUpdateData(entity)
    if type(entity) == "string" then
        entity = self.entities[entity]
    end
    
    if entity.getData then
        return entity:getData()
    else
        error("Entity "..tostring(entity.__class).." can not get data")
    end
end

--cuts all connections to entity
function Underground:kill(entity)
    self.entities[entity.uuid] = nil
    --lume.remove(self.bullets,entity)
    self.bullets[entity.uuid] = nil
end

function Underground:storeBullet(ent)
    if self.active then
        --table.insert(self.bullets,ent)
        self.entities[ent.uuid] = nil
        self.bullets[ent.uuid] = ent
        if ent.isBot then log(tostring(ent).."added") end
    end
end

function Underground:sendMessage(mes,color,isPrivate)
    local _color,_mes = color or game.settings.color,mes
    if type(_color) == "string" then
        color = getColor(_color)
    elseif _color then
        error("Color should be text")
    end
    
    mes = lume.wordcut(mes)--self._mw/(self.font:getWidth(mes)*2))
    local n,x, messes
    for x = 1,#mes do
        n = x+#self.messages
        if n>self.messageLen.h then
            messes = lume.copy(self.messages)
            for xx = 1,#messes do
                self.messages[xx] = messes[xx+1]
            end
            self.messages[-1] = nil
            n = self.messageLen.h
        end
        
        self.messages[n] = {mes[x],color}
    end
    
    if not isPrivate and self:hasPeer() then
        if self.client then
            self.client:sendMessage(_mes,_color)
        elseif self.host then
            self.host:sendMessage(_mes,_color)
        end
    end
end

function Underground:hasPeer()
    return self.client or self.host or self.server
end

function Underground:sendPrivateMessage(mes,color)
    return self:sendMessage(mes,color,true)
end

function Underground:drawMessages()
    local mm = #self.messages
    if mm == 0 then
        return
    end
    local m,r,g,b,a-- = table.concat(self.messages)
    for x = 1,self.messageLen.h do
        m = self.messages[x]
        if m then
            if m[2] then
                r,g,b,a = love.graphics.getColor()
                love.graphics.setColor(m[2])
            end
            love.graphics.print(m[1],10,self._mh+(x-1)*self.font:getHeight()*2,0,2,2)
            if m[2] then
                love.graphics.setColor(r,g,b,a)
            end
        end
    end
end
        
function Underground:spawnDigger(w,h,rand)
    local tile = (not rand and self.grid[math.random(self.minX,self.minX+10)]
    [math.random(self.minY,self.minY+10)]) or choice(self.allTiles)
    tile.isBackground = true
    x,y = tile.x,tile.y
    --tile.source = "iron.png"
    self.digger = DungeonDigger:new({w=w,h=h,x=tile._x,y=tile._y,map=self,tile=tile})
    self.diggers[#self.diggers+1] = self.digger --self.camera.scale=1
end

function Underground:getCrate(type)
    --change
    for x, i in pairs(self.bullets) do
        if i.isCrate and i.type == type then
            return i
        end
    end
end

function Underground:spawnCrate(per, amount, tile, crateID, ttype)
    --bullets:60%, tiles:30%, health:10% and ...power ups:10%...?
    
    
    local per = per or math.random(0,101)
    local item = type(per)=="table" and per
    local amount = amount or 1
    
    if not crateID and self.isClient then
        warn("unauthorized spawning of crate from client")
    end
    
    local crateID = crateID or lume.uuid()
    local ttype = ttype
    
    if not item then
        if per > 50 then
            item = choice(bullets)
            amount = amount or math.random(5,50)
            ttype = "bullet"
        elseif per > 10 then --10 then
            item = choice(tiles)
            amount = amount or math.random(5,20)
            ttype = "block"
        else--if per
            item = {heal=math.random(25,50),new=function(self,k) self.map = k.map;return self end}
            ttype = "health"
        end
    end
    
    item = item:new({parent=self,map=self,passive=true})
    item.active = false
    
    local tile = tile or nil --= choice(self.allTiles)
    --log("tile is "..tostring(tile))
    
    if not tile then
        if self.isClient then error("Tile is nil in crate!!"..tostring(tile)) end
        local t,l,i = nil,0
        while (not tile or not tile.isBackground) or l<1 do
            tile = choice(self.allTiles)
            i,l = self.world:querySegment(tile._x,tile._y,self.w,self.h*2,filterBackground)
        end assert(tile.isBackground)
    else
        local _tile = tile--????
        tile = self.entities[tile]
        if not tile then warn("no tile at "..tostring(_tile)) end
    end
    
    --tile.color=getColor("orange")
   -- log("tile is "..tostring(tile))
    local cr = Crate.spawnCrate(item, amount,tile)
    cr.miniID = crateID.."crate"
    cr.x = tile._x
    cr.y = tile._y
    self.crates[crateID] = cr
    cr.crateID = crateID
    cr.type = ttype
    
    self:setMessage("A CRATE HAS BEEN SPAWNED!")
    if self.host then
        self.host:sendToAll("do",{"spawnCrate",per,amount,tile.uuid,crateID,"map"})
    end
    return cr
end

function Underground:destroyItem(uuid,...)
    self.entities[uuid]:destroy(...)
end

function Underground:destroyMini(md)
    self._minis[md]:destroy(1)
end

function Underground:setMessage(mes)
    self.tmpMessageLast = 0
    self.message:setText(mes)
end

function Underground:spawnBot(bot,x,y)
    local tile = self:getTileForSpawn(x,y)
    local bot = bot or Bot:new({map=self})
    bot:setPos(tile._x, tile._y)
    bot.life = 100
    bot.dead = false
    game.players[#game.players+1] = bot
    bot.index = #game.players
    --bot.isCalm = true --??
    
    return bot
end

function Underground:loadAllChunks()
   -- game.progress:setTotal(self.w*self.h)
    --game.progress:setProgress(0)
    
    self:loadChunks({start=-self._maxCount.x/2,en=self._maxCount.x/2},
    {start=-self._maxCount.y/2,en=self._maxCount.y/2})
    --[[for i = -100,self._maxCount.x+1 do
        for t = -100,self._maxCount.y+1 do
            self:recordChunk(i,t)
        end
    end]]
end

function Underground:loadAllTiles()
    for i = -self._maxCount.x/2*self.chunk.width , self._maxCount.x/2*self.chunk.width do
        for iy = -self._maxCount.y/2*self.chunk.height,self._maxCount.y/2*self.chunk.height do
            self:loadTile({},i,iy)
        end
    end
end
       
function Underground:isWalkable(val)
    --log("trying "..tostring(x)..", "..tostring(y))
    --val = self.grid[x] and self.grid[x][y]
    if not val then return end
    if ANY_PATH then return true end
    if cc then return true end
    --max leap can jump across 7 tiles
    --from tile 1 to tile 9 for example.
    
    local x,y = val.x, val.y
    local tmp = {}
    local downTile = self:getTile(x,y+1) or tmp
    local upTile = self:getTile(x,y-1) or tmp
    local leftTile = self:getTile(x-1,y) or tmp
    local rightTile = self:getTile(x+1,y) or tmp
    local downLeftTile = self:getTile(x-1,y+1) or tmp
    local downRightTile = self:getTile(x+1,y+1) or tmp
    
    if val.isBackground and (
       (goalIsDown) or
       (not downTile.isBackground) or 
       (not leftTile.isBackground) or
       (not rightTile.isBackground) or
       (not downLeftTile.isBackground) or
       (not downRightTile.isBackground)
    ) then
        return true
    end 
    
    --shoot through tile
    if  --canDestroyBlocks     and
        not val.isBackground and
        not val.invincible   --and
        --not downTile.isBackground
    then
        return true
    end
    
    if val.isBackground then
        if 1 then return true end
        for i = 1,7 do
            local xx = i*xDirToGoal
            local tile2 = self:getTile(x+xx,y) or tmp
            local midTile = self:getTile(x+xx,y-1) or tmp
            local upTile2 = self:getTile(x+xx,y-2) or tmp
            local downTile2 = self:getTile(x+xx,y+1) or tmp
            
            --can land on midtile
            if upTile2.isBackground and (not midTile.isBackground) then
                return true
            elseif midTile.isBackground and (not tile2.isBackground) then
                return true
            elseif tile2.isBackground and (not downTile2.isBackground) then
                return true
            end
            
            if not upTile2.isBackground then
                if not midTile.isBackground then
                    --jump way is blocked
                    break
                end
            end
            --if not midTile.isBackground or not upTile2.isBackground then
                --return false
            --end
        end
    end
    
    --if not val.isBackground then return true
    --end
    return false
end
Do=4
--returns false if light can't pass through tile
function Underground:lightCallback(fov, x,y)
    local tile = self.grid[x] and self.grid[x][y]
    if not tile then return end
    if tile.parent == self.player then return true end
    if tile.unit then
        if tile.unit~=self.player and not tile.unit.isBullet
        then
            if not tile.unit.active then
                return error(tostring(tile==tile.unit.tile)..","..tostring(tile.unit.tile)..tile.unit.class)--inspect(tile.unit,1))
            end
            return false
        end
    end
    if not tile.isBackground then return false end
    return true
end

function Underground:loadMap(w,h)
    for x = 1,w do
        for y = 1,h do

        end
    end
end

--removes a tile, not adsvised to use
function Underground:removeTile(tile,y)
    if type(tile) == "number" then
        tile = self:getTile(tile,y)
    end
    if tile.collider then
        tile.collider:destroy()
    end
    self.world:remove(tile)
    self.grid[tile.x][tile.y] = nil
end

function Underground:getRandomTileInDungeon(digger)
    local digger = digger or choice(self.diggers)
    return digger:getRandomTile()
end

function Underground:spawnPlayers() --log(inspect(game.players,2))
    local pl
    for x = 1,#game.players do
        pl = game.players[x]
        self:spawnPlayer(pl)
    end
    self.player = game.player
end

local function filterBackground(item)
    return  item.isBackground
end

function Underground:getTileForSpawn(x,y)
    local x = x
    local y = y
    local tile
    
    --local tile = self:getTile(x,y)--,nil,true)
    if not x then
        tile,l,i = nil,0
        while (not tile or not tile.isBackground) or l<1 do
            tile = self:getRandomTileInDungeon()--choice(self.allTiles)
            i,l = self.world:querySegment(tile._x,tile._y,self.w,self.h*2,filterBackground)
        end
    else
        tile = self:getTile(x,y)
    end
    --[[
    local x,y,tile = x,y,l,i
    local maxX = self.maxX-self.minX
    local maxY = self.maxY-self.minY
    if not x then
        tile,l,i = nil,0
        while (not tile or not tile.isBackground) or l<1 do
            tile = self:getRandomTileInDungeon()--choice(self.allTiles)
            i,l = self.world:querySegment(
            self.minX*self.tileW,
            self.minY*self.tileH,
            maxX*self.tileW,
            maxY*self.tileH
            )--,tile._x,tile._y,self.w,self.h*2,filterBackground)
            if l>0 then
                tile = lume.randomchoice(i)
            end
        end
    else
        tile = self:getTile(x,y)
    end]]
    return tile
end

function Underground:spawnBotPlayers(n)
    
    for x = 1,n do
        local bot = Bot:new({map=self})
        self:spawnPlayer(bbot,nil,nil,true)
        bot.index = #game.players+1
        game.players[bot.index] = bot
    end
    
end
        
    
function Underground:spawnPlayer(pl,x,y,isBot)

    local x = x
    local y = y
    local tile = self:getTileForSpawn(x,y)
    
    pl=pl or 1
    if type(pl)=="number" then
        pl = game.players[pl]
    else
    
    end
    
    local player = pl or (isBot and Bot:new({map=self})) or game.player or Player:new({map=self})
    player:setPos(tile._x, tile._y)
    player.life = 100
    player.dead = false
    
    tile.isBackground = true
    x,y = tile.x,tile.y
    self.player = game.player
    
    tile.source = "background_dirt.png"

    
    if player==self.player then
        self.player:fixCamera(true)
        self.player._new = true
    end
    
    if self.host then
        self.host:sendToAll("spawnPlayer",{pl.index or 1,x,y,isBot})
    end
    return player
end

function Underground:playerDied(player,index)
    if (self.slowTarget==player.index) then
        self:endSlowDown()
    end
    
    if player.real or (player.isBot and not self.isClient) then
        self.respawnTime = self.timer:after(10, function()
            self:respawnPlayer(player.index or player)
        end)
        --player:send("do",{"die",index,nil,nil,nil,player.index})
        if player == self.player then
            self:dopause()
        end
    end
end

local round = lume.round
function Underground:getRespawnTime()
    return 10-round(self.respawnTime.time)
end

function Underground:respawnPlayer(playerIndex)
    local player = game.players[playerIndex] or playerIndex
    self.respawnTime = nil
    if self:hasPeer() then
        if self.client then
            self.client:send("requestRespawn",player.index)
        elseif self.host then
            self:doRespawn(player)
        end
    else
        self:doRespawn(player)
    end
end

function Underground:doRespawn(player)
    local player = player or self.player
    player:addToWorld()
    self:spawnPlayer(player)
end

function Underground:maxChunks(dir)
    local dir = dir or "x"
    if dir == "x" then
        if self.chunkCount.x >= self._maxCount.x then
            return true
        else
            return false
        end
    elseif dir == "y" then
        if self.chunkCount.y >= self._maxCount.y then
            return true
        else
           return false
          end
    else
        return self:maxChunks("x") and self:maxChunks("y")
    end
end

--doesn't record chunks exactly, more like the ones loaded at a time
function Underground:recordChunk(ox, oy, skip)
    
    local ax, ay, _ay, _ax, o = 0, 0
    if skip then
        goto skipped
    end
    _ax = self.allCurrentChunks.x
    ax = _ax[#_ax]
    if ox<1 then
        ax = _ax[1]
    end
    
    _ay = self.allCurrentChunks.y
    ay = _ay[#_ay]
    if oy<1 then
        ay = _ay[1]
    end
    o = self.offset
    if not o.newy then
        ay = 0
    end
    if not o.newx then
        ax = 0
    end
    ::skipped::
    self.records[ox+ax] = self.records[ox+ax] or {}
    self.records[ox+ax][oy+ay] = self.records[ox+ax][oy+ay] or true
end
    
    
function Underground:chunkExists(ox, oy, skip)
    local ax, ay, _ay, _ax, o = 0, 0
    if skip then
        goto skipped
    end
    _ax = self.allCurrentChunks.x
    ax = _ax[#_ax]
    if ox<1 then
        ax = _ax[1]
    end
    
    _ay = self.allCurrentChunks.y
    ay = _ay[#_ay]
    if oy<1 then
        ay = _ay[1]
    end
    
    o = self.offset
    if not o.newy then
        ay = 0
    end
    if not o.newx then
        ax = 0
    end
    ::skipped::
    if not (self.records[ox+ax] and self.records[ox+ax][oy+ay]) then log({ox+ax,oy+ay,ax,ay})
        log(inspect(self.records)) end--rror() end
    return (self.records[ox+ax] and self.records[ox+ax][oy+ay]) or false
end

local func = function(x,y)
    local t = game.map:getTile(x,y)
    if t then
        t.pride = RECORDS.updateCount+1
    end
    --log(x..","..y)
end

function Underground:slowDown()
    if (self.slowCount and self.slowCount>0) then
        return
    end
    
    if self.host then
        self.host:sendToAll("do", {"slowDown",nil,nil,nil,nil,nil,"map"})
    end
    
    self.slow = true
    self.slowdt = dddt or 1/650
    self.slowCount = ddff or .7
end

function Underground:endSlowDown()
    self.slow = false
    self.slowdt = nil
    self.slowCount = 0
end

function Underground:moveCamera(x,y,scale,time,skipFunction)
    self.timer:tween(time, self.cameraMan, {
        scale = scale,
        x = x,
        y = y,
    },nil,((not skipFunction and self.cancelMove) or self.cancelMoveLittle) )
    
    self.moving = true
    
    self.cameraMan.oldXPos   = self.cameraMan.x
    self.cameraMan.oldYPos   = self.cameraMan.y
    self.cameraMan.oldScale  = self.cameraMan.scale
    self.cameraMan.oldTime   = time
    self.cameraMan.oldStyle  = self.camera.follow_style
    
    self.camera:setFollowStyle("none")
    
end

function Underground:update(dt)
    
    --self.tickRate = self.tickRate+dt
    --[[if self.tickRate>=self.updateRate then
        self.tickRate = 0
        if self.slow then
            dt = self.updateRate
        end
    elseif not returnTick then 
        --return 
    end]]
    
  --  love.keyboard.setTextInput(true)
    
    self.updated = {}
    
    IM.active = false
    
    local dt = dt or 0
    local peer = self:hasPeer()
    
    if dt > 0.05 then dt = 0.05 end -- smoothen the movement
    
    local truedt = dt
    local fakedt = dt
    local fakePlayerdt = dt
    
    if true then
        fakedt = self.slowdt or dt
        fakePlayerdt = self.slowdt or dt/1.2
    end
    
    if self.slowCount and self.slowCount>0 then
        self.slowCount = self.slowCount-truedt
        if (self.slowCount<=0) then
            self:endSlowDown()
        end
    end
    
    if doDebug then
        hotswap.update(dt)
    end
  
    if self.pause then
        local t = self:hasPeer() and self:hasPeer():update(dt)
        assert(self:hasPeer())
        log(self:hasPeer():getSendMode())
        log("messages to send "..self:hasPeer():getMessageTimeout())
        --log(self:hasPeer():getState())
        --log("Ping(round trip time): "..self:hasPeer():getRoundTripTime())
        
        return
    end
    
    if track then ProFi:start() end
    

    
    media.cleanup()
    RECORDS.updateCount = RECORDS.updateCount+1
    self.timer:update(dt)
    
    
    if self.tmpMessageLast<self.messageLast then
        self.tmpMessageLast = self.tmpMessageLast+dt
        if self.tmpMessageLast>=self.messageLast then
            self.message:setText("")
        end
    end
    
    self.timeSinceLastCrate = self.timeSinceLastCrate+(ooi and dt or 1)
    if not self.isClient and self.timeSinceLastCrate>=self.createCrateTime then
        self.createCrateTime = lume.randomchoice({3*60,5*60,5*60,10*60,10*60,5*60})--80,360,600})
        self.timeSinceLastCrate = 0
        self:spawnCrate()
    end
    
    local trash = self.score and self.score:update(dt)
    
    if not self.camera.shaking then
        self.camera:setFollowStyle("none")
    end
    if self.cameraMan then
        self.camera:follow(self.cameraMan.x,self.cameraMan.y)
        self.camera.scale = self.cameraMan.scale or self.camera.scale
    end
    
    self.camera:update(dt)
    --self.map:update(dt)
    
    if self.cameraMan and self.cameraMan.animation then
        self.cameraMan.animation:update(dt)
        if self.cameraMan.animation.done then
            self.cameraMan.animation = nil
        end
    end
    
    local trash = self.player.paddy and self.player.paddy:update(dt)
    
    --[[if not self.vin9tage and self.player.tile and self.player.tile~= self.oldTile then
        
        local tt = self.player.tile
        self.olTile = tt
        local x, y = tt.chunkx, tt.chunky
        local chunk = self.allChunks[x][y]
        self.allCurrentChunks = {x={x},y={y}}
        for t = 1,(self.genRange.x-1)/2 do
            table.insert(self.allCurrentChunks.x,x-t)
            table.insert(self.allCurrentChunks.x,x+t)
        end
        for t = 1,(self.genRange.y-1)/2 do
            table.insert(self.allCurrentChunks.y,y-t)
            table.insert(self.allCurrentChunks.y,y+t)
        end
        
        table.sort(self.allCurrentChunks.x)
        table.sort(self.allCurrentChunks.y)
    end
    


    local o = self.offset
    local dx, dy = self.player:getDisplacement()
    o._y = dy + o._y
    o._x = dx + o._x
    if 2 then-- gh and (math.abs(dy) > 15 or math.abs(dx) > 15 ) then
   --     log(inspect(o).."\n"..inspect({dx,dy}))
    end
    gh = 4
    o.newy, o.newx = false, false
    ]]
    
    --self:updateChunks(dt,RECORDS.updateCount)

    if not self.vintage then
        local visibleThings, len, obj = self.world:queryRect(self.camera:getVisible())
        --table.sort(visibleThings,sortTiles)


        local up = getUpdateCount()
        for i=1, len do
            obj = visibleThings[i]
            obj._inView = up
            if not obj.isCrate and
                not obj.isBullet and
                    not obj.isPlayer and not obj.isPuff and not obj.isMini then
                obj:update(fakedt,RECORDS.updateCount)
            end
        end
    end--if not self.vintage
    
    --log(string.format("updating objs %s",#self.bullets))
    local obj
    local updated = {}
    
    for x,obj in pairs(self.bullets) do
        if (updated[obj]) then
            goto next
        end
        updated[obj] = true
        obj:update(fakedt,RECORDS.updateCount)
        if self.tickHost then
            if obj.isTurret then
                self.host:sendToAll("updateTurret", {obj.index,obj.life})
            elseif obj.isCrate then
                self.host:sendToAll("updateCrate", {obj.crateID, obj.life, obj.x, obj.y})
            end
        end
        ::next::
    end
    if ww and ww.update==null then log(inspect(self.bullets,1)) end

    local player
    for i = 1,#game.players do
        local player = game.players[i]
        if (not player.isBot) then
            player:update(fakePlayerdt,RECORDS.updateCount)
        end
    end
    
    if self.shangle and game.settings.mouse_shoots then
        self.player.Rjoystick.angle = self.shangle
    end

    if self.player.kickedOut then
        log("woah, out")
        ff()--??
        gooi.alert({text="Kicked out!"})
    end
    
    self.tickHost = false--bool to check whether to send turret and crate updates
    self.tickRate = self.tickRate+dt
    --always will make it run faster?
    --if (always or
    if self.tickRate>=self.updateRate and peer then
        self.tickRate = 0
        local upR = RECORDS.updateCount
        if self.slow then
            --dt = self.updateRate
        end
        peer:update(dt)
        if self.player.kickedOut then
            return
        end
        if self.host then
            local pl
            self.host:sendToAll("activate","")
            self.tickHost = true
            for i = 1,#game.players do
                pl = game.players[i]
                if pl~=self.player then
                    pl._renew_count = pl._renew_count or 0
                    if upR-pl._renew_count>60*60 then
                        self:kickOutPlayer(pl)
                    end
                end
                self.host:sendToAll("do",{"setLife",pl.life,nil,nil,nil,i})
                self.host:sendToAll("do",{"moveTo", pl.x, pl.y, pl.flipX, nil, i})
                self.host:sendToAll("do",{"setRstickAngle",pl.Rjoystick.angle,nil,nil,nil,i})
            end
            --[[for x = 1,#self.turrets do
                pl = self.turrets[x]
            end]]
        elseif self.client then
            if not self.player.send then
                --player has been disconnected and map was destroyed
                return
                --error(inspect(self.player))
            end
            self.player:send("renewClient",self.player.index)
            self.player:send("do",{"setLife",self.player.life})
            self.player:send("do",{"moveTo",self.player.x,self.player.y,self.player.flipX})
            self.player:send("do",{"setRstickAngle",self.player.Rjoystick.angle,nil,nil,nil,self.index})
            self.player:send("do",{"setScore", self.player.kills, self.player.deaths, nil,nil,self.index})
        end
    end
    
    self.cameraMan = self.cameraMan or {x=self.player.x,y=self.player.y}
    if GF then
  --  self.cameraMan = {x=game.players[1].x,y=game.players[1].y}
    end
    
    if self.player.followX and not self.moving then
        self.cameraMan.x = self.player.x
    end
    if self.player.followY and not self.moving then
        self.cameraMan.y = self.player.y
    end
    
    if (self._tag) then
        self.cameraMan.x = self._tag.x
        self.cameraMan.y = self._tag.y
        
        if (not self._tag.active) then
            self._tag = nil
        end
    end
    
    local maxZoom = 10
    local minZoom = 0.05
    if self.camera.scale >  maxZoom then
        self.camera.scale = maxZoom-1
        self.cameraMan.animation= nil
    elseif self.camera.scale < minZoom then
        self.camera.scale = minZoom+0.01
    end
    
    if self.fogOfWar then
        local tile = self.ftile or self.player.tile
        if not self.player.tile then warn("No tile!") else
        self.fov:compute(tile.x,tile.y,20,func) end
    end
    
    if self.tryToEnd then
        if self.canEndGame then
            self.tryToEnd = false
            self:endGame()
        end
    end
    
    --if peer then peer:update(dt) peer.pause = nil end
    
    if track then ProFi:stop() ProFi:writeReport("update.txt") error() end 
end
function ch()    seekNewGlobals(function(n) error(inspect(n)) end) 
end

function Underground:kickOutPlayer(pl,sent)
    if type(pl)=="number" and pl~=self.player.index then
        pl = game.players[pl]
    end
    if pl.kickedOut then return end
    
    if not self.host and not sent then
        return
    elseif self.host then
        self.host:sendToAll("do",{"kickOutPlayer",pl.index,1,nil,nil,"map"})
        if not pl.index then error(inspect(pl,1)) end
        local cl = self.host:getClientByIndex(pl.index-1)
        cl:disconnectNow()
    end
    log("kicking out!")
    pl:destroy()
    pl.kickedOut = 1
    --game.players[pl.index] = {}
end

function Underground:collectGarbage(allow)
    if 1 then--not self.isClient or allow then
        log("trying collection...")
      --  collectgarbage("collect")
        log("garbage collected.")
        if self.host then
       --     self.host:sendToAll("do",{"collectGarbage",true,0,0,0,"map"})
        end
    end
end

function cg(n) collectgarbage("step",n) end
function sw(b) bullets.SmokescreenBullet = b or Bullet end

local function sortTiles(a,b)
    return a.timeMade<b.timeMade
end

function Underground:drawGlow(obj)
    if not obj.tile then return end
    local glow = game:getSource("glow.png")
    local ow, oh = obj.tile.w*2,obj.tile.h*2--obj.w+obj.w/2, obj.h+obj.h/2
    if obj.isMini then
  --      ow,oh = 16,16
    end
    local w,h = resizeImage(glow, ow, oh)
    local x,y = obj.x,obj.y--obj.x-obj.w/2, obj.y-obj.h/2
    love.graphics.draw(glow, x-obj.tile.w/2,y-obj.tile.h/2,0,w,h)
end

local function filter(item)
    return item.isTile
end

function Underground:drawObj(obj)
        obj._inView = self.up
        local r,g,b,a = self.r,self.g,self.b,self.a
        if not obj.isBullet and not obj.isPlayer and not obj.isPuff and not obj.isMini then
            if obj.glow then
                self:drawGlow(obj)
            end --if obj.isFake then obj:draw() end--?!! end
            if self.fogOfWar then
                if obj.isTile and obj.pride and obj.pride>=RECORDS.updateCount then
                    Tile.draw(obj)--visibleThings[i]:draw()
                elseif obj.isTile then
                    love.graphics.setColor(1,1,1,.5)
                    obj.alpha = .5
                    Tile.draw(obj)--visibleThings[i]:draw()
                    obj.alpha = nil
                    love.graphics.setColor(r,g,b,a)

                elseif obj.isDebris or (obj.tile and obj.tile:isVisible()) then
                    obj:draw()
                end
            else
                obj:draw()
            end
        end
end

function Underground:draw(dt)
    if dtrack then ProFi:start() end
    
    local x, y, w, h = self.camera:getVisible()--Window()
    --error(x..", "..y..", "..w..", "..h)
    --x, y = self.camera:toCameraCoords(x,y)
    --w,h = w*self.camera.scale, h*self.camera.scale
    
    local sc = self.camera.scale
    --self.camera.scale = sc*getRatio()

    self.camera:attach()
    --[[for nx,i in pairs(self.colorTiles) do
        self:_colorTile(game:get(nx),i)
    end]]
    --self:drawChunks()
    
    local obj
    local up = getUpdateCount()
    local r,g,b,a = love.graphics.getColor()
    self.r,self.g,self.b,self.a=r,g,b,a
    self.up=up
    local visibleThings, len = self.world:queryRect(--x,y,w,h,nil,self.drawObj,self)--
    self.camera:getVisible())--,filter)
    if 1 then table.sort(visibleThings) end
    
    for i=1, len do
        obj = visibleThings[i]
        if not obj.isBullet and not obj.isPlayer and not obj.isPuff and not obj.isCrate and
        not obj.isDebris and not obj.isMini then --??
            if obj.glow then
                self:drawGlow(obj)
            end --if obj.isFake then obj:draw() end--?!! end
            if self.fogOfWar then
                if obj.isTile and obj.pride and obj.pride>=RECORDS.updateCount then
                    Tile.draw(obj)--visibleThings[i]:draw()
                    obj._inView = nil
                elseif obj.isTile then
                    love.graphics.setColor(1,1,1,.5)
                    obj.alpha = .5
                    Tile.draw(obj)--visibleThings[i]:draw()
                    love.graphics.setColor(r,g,b,a)
                    obj.alpha = nil
                    obj._inView = nil

                elseif obj.isDebris or (obj.tile and obj.tile:isVisible()) then
                    --visibleThings[i]:draw()
                end
            else
                obj._inView = up
                obj:draw()
            end
        else
            obj._inView = up
        end
    end

    local player
    for i = 1,#game.players do
        local player = game.players[i]

        if player.glow then
            self:drawGlow(player)
        end
        if not self.fogOfWar --or not self.ftile
        or player == self.player or (player.tile and
        player.tile:isVisible()) then
            player:draw()
        elseif not Crate.inView(player) then
            Crate.drawPointer(player)
        end
    end

    --?? fix ifs for crate and stuff
    local obj
    for x in pairs(self.bullets) do
        obj = self.bullets[x]
        if obj.glow then
            self:drawGlow(obj)
        end
        if obj.isCrate and obj.tile and (not obj:inView() or not obj.tile:isVisible()) then
            obj:drawPointer()
            goto continue
        end
        if
        --not obj.isDebris and not obj.isCrate and 
        (not self.fogOfWar or not obj.tile or obj.tile:isVisible()) then
            obj:draw()
        end
        ::continue::
    end
    
    self.camera:detach()
    local trash = self.usePaddy and self.paddy:draw()
    local t = self.digger and self.digger:draw()
    
    self.camera:draw()
    --love.graphics.setScissor()

    --love.graphics.rectangle("line",x,y+10,w,h)
    self:drawMessages()
    trash = self.score and self.score:draw()
    if doDebug then
        love.graphics.print(inspect(love.graphics.getStats()),100,H()-450,0,2,2)
        love.graphics.print(string.format("FPS: %s",love.timer.getFPS()),100, H()-150,0,2,2)
        love.graphics.print(collectgarbage("count")/1024,100, H()-100,0,2,2)
        love.graphics.print(love.graphics.getStats().texturememory/(1024*1024),100,H()-50,0,2,2)
    end
    if self.drawProgress then
        game.progress:draw()
    end
    
    self.camera.scale = sc
    
    if dtrack then ProFi:stop() ProFi:writeReport("udraw.txt") error(tt) end

end

function Underground:setProgress(n)
    local p = game.progress
    p:setProgress(n)
end

function Underground:progress(n)
    if not self.drawProgress then
        self.drawProgress = true
    end
    local p = game.progress
    p:progress(n)
end

function Underground:stopProgress()
    self.drawProgress = false
end

function Underground:updateChunks(dt, c)
    local xval, yval
    local _
    for _ = 1, #self.allCurrentChunks.x do
        xval = self.allCurrentChunks.x[_]
        --[[if self.allChunks[xval-5] then
            self.allChunks[xval-5] = nil
        end
        if self.allChunks[xval+5] then
            self.allChunks[xval+5] = nil
        end]] 

        for _y = 1, #self.allCurrentChunks.y do
            yval = self.allCurrentChunks.y[_y]
            if (not self.allChunks[xval] or not self.allChunks[xval][yval]) and 
            not self.loadedAll
            then
                self:loadChunks({start=xval,en=xval},{start=yval,en=yval})
                local chunk = self.allChunks[xval][yval]
                --???self:load()
                chunk:update(dt,c)
            elseif (self.allChunks[xval] and self.allChunks[xval][yval]) then
            
                local chunk = self.allChunks[xval][yval]
                chunk:update(dt,c)
            end
        end
    end
end

function Underground:getTileP(...)
    return self:getTileWithPos(...)
end

function Underground:getTile(_tile,x,y,dontSave,er)
    if type(_tile) == "number" then
        local y = x
        local x = _tile
        return (self.map.grid[x] and self.map.grid[x][y])
    end
    
    local _tile = _tile or {}
    local u = nil
    local tile = (self.map.grid[x] and self.map.grid[x][y]) 
    if tile then return tile end
    if er then error("No tile!") end
    
    tile = Tile:new({_tile=_tile,map=self,x=x,y=y})

    if self.dontSave or dontSave then
        return tile
    end
    
    for _,x in pairs(tile) do
        _tile[_] = _tile[_] or x
        if type(x) == "function" then
            _tile[_] = x
        end
    end
    return _tile
end

function Underground:loadTile(_tile,x,y)
    if not _tile then
        return
    end

    local tile = self:getTile(_tile,x,y,true)
    if tile.loaded then
        return tile
    end

    tile.loaded = true
    for u in pairs(_tile) do
        tile[u] = tile[u] or _tile[u]
    end
    tile.unit = nil
    
    self.map.grid[x] = self.map.grid[x] or {}
    self.map.grid[x][y] = tile
    --self.map.tiles[tile.id] = tile
    tile.gid = tile.id
    table.insert(self.allTiles, tile)
    tile.source = "dirt.png"
    local tx, ty = x*tile.w, y*tile.h
            
    --tile.collider:setObject(tile)
    --tile.collider:setType("static")
    --tile.collider:setCollisionClass("Dirt")

    

    return tile
end

function Underground:loadChunks(x,y,px,py,llog)
    local px = px or 1
    assert(not self.loadedAll)
    local py = py or 1
    local x = x or {start=1,en=self.chunkSize.w}
    local y = y or {start=1,en=self.chunkSize.h}

    for cx = x.start, x.en do
    
        if self.allChunks[cx] then
            self.chunkCount.x = self.chunkCount.x+1
        end
        
        self.allChunks[cx] = self.allChunks[cx] or {}
        
        for cy = y.start, y.en do
            --[[if self.allChunks[cx][cy] then
                goto skipTiles
            end
            log("cy ")
            if not self.yChunks[cy] then
                self.chunkCount.y = self.chunkCount.y+1
                self.yChunks[cy] = true
            end
            --self:recordChunk(cx,cy,true)
            self.allChunks[cx][cy] = self.allChunks[cx][cy] or {
                x = cx,
                y = cy,
                tiles = {},
                allTiles = {},
                width = self.chunk.height,
                height = self.chunk.width,
                parent = self,
                draw = function(self)
                    for x,i in lume.iter(self.allTiles) do
                        i:draw()
                    end
                end,
                update = function(self,dt,c)
                    for x,tile in lume.iter(self.allTiles) do
                        tile:update(dt,c)
                    end
                end
            }
            local chunk = self.allChunks[cx][cy]
            local tiles = chunk.tiles
            
            --Check if it has already been loaded
            if tiles[1] then
                --if so skip this
                goto skipTiles
            end]]
            local tiles = {} 
            for tx = px, self.chunk.width+(px-1) do
                tiles[tx] = tiles[tx] or {}
                for ty = py,self.chunk.height+(py-1) do
                    local ttx,tty = tx + ((cx-1)*self.chunk.width),
                                    ty + ((cy-1)*self.chunk.height)
                    local tile = self:loadTile({},ttx,tty)
                    
                    tiles[tx][ty] = tile
                    
                    --these are children that should be drawn
                   -- tile.children = {}


                --    tile.x = tx
                --    tile.y = ty
                    
           --         tile._x = ttx*tile.w
               --     tile._y = tty*tile.h
                    tile.chunkx = cx
                    tile.chunky = cy

                    local map = self
                    if not self.maxX or ttx>self.maxX then
                        self.maxX = ttx
                    end
                    if not self.minX or ttx<self.minX then
                        self.minX = ttx
                    end
                    if not self.maxY or tty>self.maxY then
                        self.maxY = tty
                    end
                    if not self.minY or tty<self.minY then
                        self.minY = tty
                    end
                    
                    tile.getAdjacentTiles = function(self)
                        return map:getAdjacentTiles(self.ttx,self.tty)
                    end
                  --  if tile.isRiver then tile.ttx=100000 end
                    --table.insert(chunk.allTiles,tile)
                    
                end
            
            end
            
            if mainMenu.drawProgress then
                mainMenu:setProgress(#self.allTiles)
                game.copaused = 1--game.copaused and game.copaused+1 or 0
                        
                coroutine.yield(game.co)
            end
            
            --tiles[1][1].source="red.png"
            ::skipTiles::
        end
    end
    
    
    local t
    local am = 2

    for x = self.minX-am, self.maxX+am do
        for y = -am, am do
            t = self:getTile(x, self.minY+y)
            t = t and t:setInvincible(true)
            t = self:getTile(x, self.maxY+y)
            t = t and t:setInvincible(true)
        end
    end
        
    for y = self.minY-am, self.maxY+am do
        for x = -am,am do
            t = self:getTile(self.minX+x,y)
            t = t and t:setInvincible(true)
            t = self:getTile(self.maxX+x,y)
            t = t and t:setInvincible(true)
        end
    end


    
  --  self.allChunks[1][1].tiles[1][1].source = "player_down.png"
end

function Underground:getAdjacentTiles(x,y)
    local tiles = {}
    for _,x in lume.iter(getRound(1)) do
        local tile = self.tiles[x+c[1]] and self.tiles[x+c[1]][y+c[2]]
        if tile then
            table.insert(tiles,tile)
        end
    end
    assert(#tiles~=0,"Couldn't get adjacent tiles.")
    return tiles
end

function Underground:drawChunks()
    local xval, yval
    for _ = 1, #self.allCurrentChunks.x do
        xval = self.allCurrentChunks.x[_]
        for _y = 1, #self.allCurrentChunks.y do
            yval = self.allCurrentChunks.y[_y]
            if not self.allChunks[xval] or not self.allChunks[xval][yval] then
            else
                local chunk = self.allChunks[xval][yval]
                chunk:draw()
            end
        end
    end
end

function Underground:reloadAllCurrentChunks(x,y)
    local x,y = x or 0, y or 0
    self.allCurrentChunks = {
        x = {x},
        y = {y}
    }
    for t = 1,(self.genRange.x-1)/2 do
        table.insert(self.allCurrentChunks.x,-t+x)
        table.insert(self.allCurrentChunks.x,t+x)
    end
    for t = 1,(self.genRange.y-1)/2 do
        table.insert(self.allCurrentChunks.y,-t+y)
        table.insert(self.allCurrentChunks.y,t+y)
    end
    
    table.sort(self.allCurrentChunks.x)
    table.sort(self.allCurrentChunks.y)
end

function Underground:findPathTo(obj,x,y)

    local obj = obj
    local x = x
    local y = y
    if type(obj) ~= "table" then
        x = obj
        y = x
        obj = y
    end
    
    local floor = math.floor
    local _x,_y = (obj.getPos and obj:getPos()) or obj.x, obj.y
    
    --start = {x=floor(_x),y=floor(_y)}
    
    if not self.grid[floor(x)] then log("no tile,returning") return end
    if not self.grid[floor(x)][floor(y)] then log("no tile, returning") return end
    if self.grid[floor(x)][floor(y)].collidable == true then
        --log("collidable")
        return
    end
    obj._goto = self.grid[floor(x)][floor(y)]
    self.c_unit = obj


    local function check(x,y)
        return self:isWalkable(x,y)
    end
    
    local path = {}
    local function buildPath(x,y)
        table.insert(path,_Path:new(x,y))
    end
    
    goal = {x=x,y=y}
    start = {x=_x,y=_y}
    local pth = rot.Path.AStar(goal.x,goal.y,check)
    pth:compute(start.x,start.y,buildPath)
    return path
end

function Underground:load()
    local fov = function(fov,x,y) return self:lightCallback(fov,x,y) end
    self.fov = rot.FOV.Precise:new(fov, {topology=4})
    
    local tfov = function(fov,x,y) return self:tLightCallback(fov,x,y) end
    self.terrainFov = rot.FOV.Precise:new(tfov)
    
    self:loadPathfinder()
    self.pathfinder:setHeuristic(botHeuristic)

end

function Underground:mousereleased(x, y, button)
    local trash = self.score and self.score:mousereleased(x,y,button)
    if trash then
        self.buttonPressed = true
    end
    --self.player:released()
    gooi.released()
end

function Underground:mousepressed(x, y, button)
    local trash = self.score and self.score:mousepressed(x,y,button)
    if trash then
        self.buttonPressed = true
    elseif game.settings.mouse_shoots then
        local x,y = self.camera:toWorldCoords(x,y)
        local angle = -math.atan2(
        y-self.player.y,
        x-self.player.x)
        self.shangle = angle
        if (game.settings.mouse_shoots) then
            self.player.toShoot = {x=x,y=y}
        end
    end
   -- self.player:pressed()
    gooi.pressed()
end

function Underground:touchpressed(id, x, y, dx, dy, pressure)
    if DESKTOP then
        return
    end
    self.super.touchpressed(self,id,x,y,dx,dy,pressure)
    
    local trash = self.player.pressed and self.player:pressed(id,x,y,dx,dy,pressure)
    local trash = (self.score and self.score:touchpressed(id,x,y,dx,dy,pressure)) or trash
    if trash then
        self.buttonPressed = true
    end
    local x,y = self.camera:toWorldCoords(x,y)
   
    --local tile = self:getTileWithPos(x,y)
    if tile then
        log(inspect(tile.color)..'\n'..inspect(tile.source))
        --self.player:placeBlock(0,0,nil,tile)
    end
end

function Underground:touchmoved(id,x,y,dx,dy,pressure)
    if DESKTOP then
        return
    end
    
    local t = self.player.touchmoved and self.player:touchmoved(id,x,y,dx,dy,pressure)
    local x,y = self.camera:toWorldCoords(x,y)
   
    --local tile = self:getTileWithPos(x,y)
    if tile then
        log(inspect(tile.color)..'\n'..inspect(tile.source))
        --self.player:placeBlock(0,0,nil,tile)
    end
end

function Underground:touchreleased(id, x, y, dx, dy, pressure)
    if DESKTOP then
        return
    end
    self.super.touchreleased(self,id,x,y,dx,dy,pressure)
    
    local trash = self.player.released and self.player:released(id,x,y,dx,dy,pressure)
    
    trash = (self.score and self.score:touchreleased(id,x,y,dx,dy,pressure)) or trash
    if trash then
        self.buttonPressed = true
    end
    if not self.buttonPressed then
    else
        self.buttonPressed = false
        self.ftile = false
        return
    end
    
    local x,y = self.camera:toWorldCoords(x,y)
   
    --local tile = self:getTileWithPos(x,y)
    if tile then
        log(inspect(tile.color)..'\n'..inspect(tile.source))
        --self.player:placeBlock(0,0,nil,tile)
    end
    if self.stealthInspection and tile then
        self.ftile = tile
        log(tile.life)
    end
    
    local sx,sy = self.map:convertPixelToTile(x,y)
    sx = math.floor(sx)
    sy = math.floor(sy)
end

function Underground:textinput(text)
  gooi.textinput(text)
end

function Underground:keypressed(key, scancode, isrepeat)
  IM:keyPressed(key)
  
  gooi.keypressed(key, scancode, isrepeat)
  --[[if key == "escape" then
    p:close()
    love.event.quit()
  end]]
end

function Underground:keyreleased(key, scancode)
  IM:keyReleased(key)
  gooi.keyreleased(key, scancode)
  if key=="escape" then
      if self.exiting then
        self.score:destroy()
      else
        self:dopause()
        self.score.exit.events.r()
      end
  end
  if key=="enter" and self.exiting then
    self:exit()
  end
end

function Underground:exit()
    self:unpause()
    if self.map:hasPeer() then
        if game.host then
            game.host:destroy()
        end
    end
        
    self:destroy()
    mainMenu:create("menu")
    GSS(mainMenu)
end

return Underground