--[[
-- media file
-- This file loads and controls all the sounds of the game.
-- * media.load() reads the sounds from the disk. It must be called before
--   the sounds or music are used
-- * media.music contains a source with the music
-- * media.sfx.* contains multisources (see lib/multisource.lua)
-- * media.cleanup liberates unused sounds.
-- * media.countInstances counts how many sound instances are there in the
--   system. This is used for debugging
]]

local media = {}

function media.newSource(name,_name)
  local source = media.am:getSound(_name)
  local pl = {source=source}
  
  function pl:play()
    return media.am:playSound(source)
  end
  
  return pl
end

local function getMusic()
    return "tmp.ogg"
end

media.load = function()
    media.am = AudioManager:new()
  
    local names = [[
    debris
    entity_destroyed
    explosion
    bullet_shot
    laser
    player_landed
    player_jumped
    ]]
    
    media.sfx = {}
    
    local items = love.filesystem.getDirectoryItems("assets/sfx/")
    for _, _name in ipairs(items) do
        local name = _name:sub(1,-5)
        media.sfx[name] = media.newSource(name,_name)--,_name:sub(-4,-1))
    end

  media.loadMusic()

end

media.loadMusic = function(source, notPlay)
  media.music = media.am:playMusic(source or getMusic(),true)
  return media.music
end

media.cleanup = function(dt)
  media.am:update(dt)
  --[[for _,sfx in pairs(media.sfx) do
    sfx:cleanup()
  end]]
end

media.setSFXVolume = function(v)
  --[[v = v/10
  for _,sfx in pairs(media.sfx) do --error(inspect(sfx))
    sfx.source:setVolume(v)
  end]]
  
end

media.setVolume = function(v)
    v = v/10
    love.audio.setVolume(v)
end

return media
