local LifeBar = class:extend("LifeBar")

function LifeBar:__init__(kwargs)
    self.parent = kwargs.parent or kwargs
    self.totalHealth = kwargs.totalHealth or self.parent.totalHealth
    self.x = kwargs.x or 0
    self.y = kwargs.y or 0
    self.w = kwargs.w or 40
    self.h = kwargs.h or 4
    
    self.damagedColor = getColor(kwargs.damagedColor or "red")
    self.healthColor = getColor(kwargs.healthColor or "green")
    
    self.alwaysDraw = kwargs.alwaysDisplay or kwargs.alwaysDraw
end

function LifeBar:getHealth()
    local h = self.parent.life or self.parent.health or 100
    if h<=0 then
        return 0
    end
    return h
end

function LifeBar:getTotalHealth()
    return self.parent.totalHealth or self.parent.totalLife or self.totalHealth or 100
end

function LifeBar:draw()
    self.x = self.parent:getCenter()
    self.x = self.x-self.w/2--self.x-self.lifeBar.w/4
    self.y = self.parent.y-self.h-self.parent.h/5
    local h = self:getHealth()
    local t = self:getTotalHealth()
    if h/t >= 1 and not self.alwaysDraw then
        return
    end
    local r,g,b, a = love.graphics.getColor()
    
    love.graphics.setColor(self.damagedColor)
    love.graphics.rectangle('fill', self.x, self.y, self.w, self.h)
    love.graphics.setColor(self.healthColor)
    love.graphics.rectangle('fill', self.x, self.y, h / t * self.w, self.h)
        
    love.graphics.setColor(r,g,b,a)
end

local ProgressionBar = class:extend("ProgressionBar")

function ProgressionBar:__init__(kwargs)
    self.loadedColor = getColor(kwargs.loadedColor or "white")
    self.totalColor = getColor(kwargs.totalColor or "red")
    self.progression = kwargs.progression or 0
    self.total = kwargs.total or 100
    
    self.x = kwargs.x or 0
    self.y = kwargs.y or 0
    self.h = kwargs.h or 10
    self.w = kwargs.w or 10

end

function ProgressionBar:progress(n)
    self.progression = self.progression+(n or 0)
end

function ProgressionBar:setProgress(n)
    self.progression = n or 0
end

function ProgressionBar:setTotal(n)
    self.total = n or 100
end

function ProgressionBar:draw()
    local p = self.progression
    local t = self.total
    
    local r,g,b, a = love.graphics.getColor()
    
    love.graphics.setColor(self.totalColor)
    love.graphics.rectangle("fill", self.x, self.y, self.w, self.h)
    love.graphics.setColor(self.loadedColor)
    love.graphics.rectangle("fill", self.x, self.y, p / t * self.w, self.h)
    
    love.graphics.setColor(r,g,b,a)
end

return {LifeBar, ProgressionBar}