
local dir = love.filesystem.enumerate or love.filesystem.getDirectoryItems
local time = love.timer.getTime or os.time
local _lastscan = 0
local _reloadinterval = .5
local _files = {}
  

local updated = {}
local function updateTable(old,new)
    if updated[old] then
        return old
    end
    
    updated[old] = true
    
    local oldmt, newmt = getmetatable(old), getmetatable(new)
    if oldmt and newmt then updateTable(oldmt, newmt) end
    
    for index, newValue in pairs(new) do
        if type(newValue)=="table" and type(old[index])=="table" then
            updateTable(old[index],newValue)
        else
            old[index] = newValue
        end
    end
end

function reloadClass(class)
    log("reloading class "..tostring((class.class or "")))
    ---if type(class.class)=="function" then error(inspect(class),2) end
    local path = _can_reload[class]
    if not path then 
        error("No path for class at "..(class.class or "Unknown").." at id "..tostring(Tile))
        return
    end
    
    updated = {}
    package.loaded[path] = nil
    
    local ok,new = pcall(require,path)
    
    if ok then
        updateTable(class,new)
    else
        log(new)
        return
    end
    updated = {}
    log("reloaded")
    return class
end


local function isdir(path)
    local info = love.filesystem.getInfo(path)
    return info.type == "directory"
end

local function lastmodified(path)
    local info = love.filesystem.getInfo(path, "file")
    return info.modtime
end


local function listdir(path, recursive, skipdotfiles)
  path = path or "entities"--(path == ".") and "" or path or ""
  local function fullpath(x) return path .. "/" .. x end
  local t = {}
  for _, f in pairs(lume.map(dir(path), fullpath)) do
    if not skipdotfiles or not f:match("/%.[^/]*$") then
      if recursive and isdir(f) then
        t = lume.concat(t, listdir(f, true, true))
      else
        table.insert(t, lume.trim(f, "/"))
      end
    end
  end

  return t
end

local function getchanged(path)
  local function fn(f)
    return f:match("%.lua$") and _files[f] ~= lastmodified(f)
  end
  return lume.filter(listdir(path, true, true), fn)
end

local function resetfile(f)
  _files[f] = lastmodified(f)
end

function reloadFile(_file)
    local mod
    local file = _file:sub(1,-5)
    for x,i in pairs(_can_reload) do
        if i==file then
            mod = x
            break
        end
    end
    if not mod then
        warn("no module "..file)
        return
    end
    resetfile(_file)
    return reloadClass(mod)
end


local paths = {"entities","maps","bases","tools"}
local function dohotswap()
  local changed
  for i,x in ipairs(paths) do
      changed = getchanged(x)
      lume.each(changed, reloadFile)
      if #changed>0 then
          --break
      end
  end
  return changed or {}
end

return
{
class = "hotswap",
init = function()
  if 4 then return end
  lume.each(getchanged(), resetfile)
end,
update = function(dt)
  if 3 then return end
  local diff = time() - _lastscan
  if diff > _reloadinterval then
    _lastscan = _lastscan + diff
    local changed = dohotswap()
    if #changed > 0 then
        for x,i in ipairs(changed) do
            log("changed "..tostring(i))
        end
    end
  end
end
}