local Tutorial = class:extend("Tutorial")

function Tutorial:__init__(kwargs)

    self.currentMap = Underground:new({w = 50,h = 50, seed = 0.10996465317326 })
    self.map = self.currentMap
    game.map = self.map

    if not game.player then
        game.player = Player:new({map=self.currentMap})
        game.players[1] = game.player
        game.player.index = 1
    end

    if game.player then
        game.player.real = true
        gooi.addComponent(game.player.displayBlocks)
        gooi.addComponent(game.player.displayBullets)
    end
    
    game.players[2] = Bot:new({map=self.map,index=2,isCalm=4,team=game.player.team})
    
    self.player = game.player
    
    --??
    self.map.match = "deathmatch"
    self.map:setMatchTime()
    self.map.fogOfWar = kwargs.fogOfWar
    self.map.size = small
    self.map.stealthInspection = mainMenu.stealthInspection
    self.map.mapType = "cave"
    self.map.training = true
    self.map.player = self.player
    self.map.tutorial = self
    
    mainMenu:destroy()
    
    local start = function()
        self.map:start()
        if self.host then
            self.map.pause = 1
        end
        self.bot = Bot:new({
            map = self.map,
            aggressive = false,
            team = self.player.team,
            isCalm = 4
        })
        self.bot:reload(
            "Sh@dow",
            "black",
            self.player._hat2,
            self.player._hat,
            "black",
            "black"
        )
        tut = self
        vv = self.bot
        self.bot.isCalm = true
        self.map:spawnBot(self.bot,self.player.tile.x+1,self.player.tile.y)
        game.players[3] = self.bot
        --self.map:spawnBotPlayers(1)
        --self.map.timer:after(10,function() self:say(nil,"Yo bro!") end)
        --self.map.timer:after(15,function() self:say(self.bot,"Ha ha , nice")
        --self:say(self.bot,"To be clear, I'm here to teach you...") end)
        GSS(self)
        
        self:startTutorial()
    end
    
    if not doDebug then
        self.co = coroutine.create(start)
    else
        start()
    end
    
    game.co = self.co
    if self.co then
        coroutine.resume(self.co)
    end

end

function Tutorial:update(dt)
    Talkies.update(dt)
    self.map.timeSinceLastCrate = -1
    self.map:update(dt)
    if self.act then
        if self.act[3] then
            self.act[3](self.act,self.map,dt)
        end
    end
end

function Tutorial:draw()
    self.map:draw()
    Talkies.draw()
end

function Tutorial:startTutorial()
    self.tutorial = lume.deepcopy(_G.tutorial)
    self.currentState = self.tutorial.start
    self:doAction()
end

function Tutorial:doAction()
    local act = self.tutorial[self.currentState]
    if not act then
        assert(not self.currentState, self.currentState)
        return
    end
    
    if type(act[1]) == "string" then
        self:say(self.bot, act[1])
    else
        act[1](act, self.map)
    end
    
    self.act = act
end

function Tutorial:doNextAction(canDo)
    local act = tutorial[self.currentState]
    
    if not act then
        return
    end
    
    if not act[3] or canDo then
        self.currentState = act[2]
        self:doAction()
    end
end

function Tutorial:say(object, message, now)
    object = object or self.player
    local d = self.dialog
    local color = getColor(object.color or "black")
    if color[1]==0 and color[2]==0 and color[3]==0 then
        color = getColor("white")
    end
    --assert(not color[4] or color[4]>=1)
    self.dialog = Talkies.say(object.name, message, {
        backgroundColor = color,
        object = object,
        --imageColor = object.color,
        messageColor = colors.black,--(color~=colors.black  and colors.black) or colors.white,
        imageOpacity = 1,
        titleColor = colors.black,
        textSpeed = sch or 0.05
    })
    if d and now then
        Talkies.onAction()
    end
end

function Tutorial:mousereleased(x, y, button)
    self.map:mousereleased(x,y,button)
end

function Tutorial:mousepressed(x, y, button)
    self.map:mousepressed(x,y,button)
    Talkies.onAction()
    if not Talkies.hasDialog() then
        if self.currentState then
            self:doNextAction()
        end
    end
end

function Tutorial:touchpressed(id, x, y, dx, dy, pressure)
    self.map:touchpressed(id,x,y,dx,dy,pressure)

end

function Tutorial:touchmoved(id,x,y,dx,dy,pressure)
    self.map:touchmoved(id,x,y,dx,dy,pressure)

end

function Tutorial:touchreleased(id, x, y, dx, dy, pressure)
    self.map:touchreleased(id,x,y,dx,dy,pressure)
end

function Tutorial:textinput(text)
  gooi.textinput(text)
end

function Tutorial:keypressed(key, scancode, isrepeat)
    self.map:keypressed(key, scancode, isrepeat)
end

function Tutorial:keyreleased(key, scancode)
    self.map:keyreleased(key, scancode)
end

return Tutorial